/* bidiCommon.c -- Common functions for Bidi
 * Copyright (C) 2003 EPSON KOWA Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: bidiCommon.c,v 1.8 2004/02/16 11:21:30 toko Exp $ */

/**
 * @file  bidiCommon.c
 * @brief bidi ϢǻѤ붦̴ؿ
 *
 * @author toko (EPSON KOWA Co.) <toko@epkowa.co.jp>
 * @date 2004-01-16
 * @version $Id: bidiCommon.c,v 1.8 2004/02/16 11:21:30 toko Exp $
 */


#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

#include <netinet/in.h>

#if defined(DEBUG)
#include <stdarg.h>
#endif /* DEBUG */


/** Debugѥåϴؿ
 * @param *fmt,... ɸؿǤ printf ؿƱΰȤ
 * @return ̵
 */
static void DebugPrint(const char *fmt, ...)
{
#ifdef DEBUG
	va_list ap;

	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	fflush(stderr);
#endif
	return;
}


/** ꥵʬΥǡɤ߹
 * @param fd ɤ߹ѥեǥץ
 * @param *buf ɤ߹ǡǼΰؤΥݥ
 * @param size ɤ߹٤Хȿ
 * @retval  ºݤɤ߹᤿Хȿ
 * @retval Ի 
 */
static int fullRead(int fd, char *buf, int size)
{
	int len = 0;
	int p = 0;
	while (p < size) {
		len = read(fd, buf + p, size - p);
		if (len == 0 || len == -1) {
			DebugPrint("%s(%d) %s: %s (%d).\n", __FILE__, __LINE__, __FUNCTION__, strerror(errno), errno);
			return len;
		}
		p += len;
	}
	return size;
}

/** ꥵʬΥǡ񤭹
 * @param fd 񤭹ѥեǥץ
 * @param *buf 񤭹٤ǡǼΰؤΥݥ
 * @param size 񤭹٤Хȿ
 * @retval  ºݤ˽񤭹᤿Хȿ
 * @retval Ի 
 */
static int fullWrite(int fd, char *buf, int size)
{
	int len = 0;
	int p = 0;
	while (p < size) {
		len = write(fd, buf + p, size - p);
		if (len == -1) {
			DebugPrint("%s(%d) %s: %s (%d).\n", __FILE__, __LINE__, __FUNCTION__, strerror(errno), errno);
			return -1;
		}
		p += len;
	}
	return size;
}

/** fd 8Хɤ߹ߡbidi ̿ΥإåȤƽ
 * @param fd ɤ߹ѥեǥץ
 * @param *cmdID ɤ߹ command ID
 * @retval  data length
 * @retval Ի 
 */
static int readHeader(int fd, int *cmdID)
{
	int len;
	int header[2];					/* ɤ߹إåǼΰ衣int ǤʤФʤʤ */

	len = fullRead(fd, (char *)header, 8);
	if (len <= 0) {
		return(-1);
	}
	*cmdID = ntohl(header[0]);
	return(ntohl(header[1]));

}

/** fd Ф bidi ̿ΥѥåȤȤƽ񤭹
 * @param fd 񤭹ѥեǥץ
 * @param cmdID command ID
 * @param cnt data length (񤭹Ĺ)
 * @param args[] data ()
 * @retval  ºݤ˽񤭹᤿Хȿ
 * @retval Ի 
 */
int writePacket(int fd, int cmdID, int cnt, int args[])
{
	int len, i;
	int packet[1 + 1 + cnt];		/* 񤭹ߥǡǼΰ衣int ǤʤФʤʤ */

	packet[0] = htonl(cmdID);
	packet[1] = htonl(cnt * 4);
	for (i = 0; i < cnt; i++) {
		packet[i + 2] = htonl(args[i]);
	}
	len = fullWrite(fd, (char *)packet, 4 + 4 + cnt * 4);
	if (len == -1) {
		return(-1);
	}
	return(4 + 4 + cnt * 4);

}



/** fd  bidi ̿ΥѥåȤȤɤ߹
 * @param fd ɤ߹ѥեǥץ
 * @param *cmdID ɤ߹ޥɤID
 * @param *buf ɤ߹ǡǼΰؤΥݥ
 * @param size ΰ礭 (int Ǥο)
 * @retval  ºݤɤ߹᤿Хȿ
 * @retval Ի 
 */
int cmdRead(int fd, int *cmdID, int *buf, int size)
{
	int	len;
	int	i;

	len = readHeader(fd, cmdID);

	memset((char *)buf, 0, size);				/* ʸɤ߹ơ0 ꥢ */

	if (len > size) {
		DebugPrint("%s(%d) %s: read buffer too small, len = %d, size = %d\n", __FILE__, __LINE__, __FUNCTION__, len, size);
		return(-1);
	}

	if ((len > 0) && (len <= size)) {
		len = fullRead(fd, (char *)buf, len);
		for (i = 0; i < ((len + 3) / 4); i++) {
			/* ХȥĴ */
			buf[i] = ntohl(buf[i]);
		}
	}

	return(len);
}

/* End of bidiCommon.c */
