/*  bidiEpson.c -- EPSON Credit-Based Transport
 *  Copyright (C) 2003  EPSON KOWA Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: bidiEpson.c,v 1.17 2004/02/23 01:29:31 kitayama Exp $ */

/**
 * @file bidiEpson.c
 * @brief EPSON Credit-Based Transportμ
 * @author EPSON KOWA Corporation.
 * @date Fri Feb 23 2004
 *
 * @todo ͥ¤ϡꥹȹ¤ˤ٤
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <sys/time.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <signal.h>

#if defined(DEBUG)
#include <stdarg.h>
#endif /* DEBUG */

#include "bidiEpson.h"

#ifndef FALSE
/**  */
#define FALSE 0
#endif

#ifndef TRUE
/**  */
#define TRUE (!FALSE)
#endif

#ifndef NULL
#ifdef __cplusplus
/** NULL */
#define NULL 0
#else
/** NULL */
#define NULL (void *)0
#endif
#endif


/** read/writeԻԤ(usec) */
#define IO_WAITTIME (5000)
/** I/OΥॢȻ(sec) */
#define ECBT_IO_TIMEOUT (5)
/** ޥȯԤ顢ץ饤ޤǤwait(micro sec) */
#define ECBT_IO_WAITTIME (100000)

/** replyǼХåեκ祵 */
#define REPLY_BUFFER_MAX (4096)
/** replyѤbuffer size */
#define ECBT_REPLY_MAX (512)

/** ۥȤݻǽCredit(ץ󥿤󶡲ǽCredit) */
#define ECBT_HOST_CREDIT_MAX (0x0080)
/** ץ󥿤ݻǽCredit(ۥȤ󶡤Credit) */
#define ECBT_PRINTER_CREDIT_MAX (0x0008)
/** ץ󥿤󶡤륯쥸åȿ */
#define ECBT_QUANTITY_PROVIDING_CREDIT (1)

/** ȥ󥶥ѥåȤκ祵 */
#define ECBT_TRANSACTION_FORWARD_MAX (512)
/** ȥ󥶥ѥåȤκ祵 */
#define ECBT_TRANSACTION_REVERSE_MAX ECBT_REPLY_MAX
/** 楳ޥѥåȤκ祵 */
#define ECBT_COMMAND_FORWARD_MAX (512)
/** 楳ޥɼѥåȤκ祵 */
#define ECBT_COMMAND_REVERSE_MAX ECBT_REPLY_MAX
/** ǡѥåȤκ祵 */
#define ECBT_DATA_FORWARD_MAX (2048)
/** ǡѥåȤκ祵 */
#define ECBT_DATA_REVERSE_MAX ECBT_REPLY_MAX

/** IEEE1284.4κѥåȥ */
#define CBT_PACKET_SIZE_MAX (0xFFFF)
/** IEEE1284.4Υإå */
#define CBT_HEADER_SIZE (0x06)
/** IEEE1284.4κǡ */
#define CBT_DATA_SIZE_MAX (CBT_PACKET_SIZE_MAX - CBT_HEADER_SIZE)

/** @defgroup cbt_header_offset IEEE1284.4Υѥåȥإåγƥեå
    @{*/
/** PSID */
#define CBT_HEADER_PSID(header)     ((unsigned char)(header[0]))
/** SSID */
#define CBT_HEADER_SSID(header)     ((unsigned char)(header[1]))
/** ѥåĹ */
#define CBT_HEADER_LENGTH(header)   ((((unsigned int)((unsigned char)(header[2]))) << 8) | (unsigned char)(header[3]))
/** ԥХå쥸å */
#define CBT_HEADER_CREDIT(header)   ((unsigned char)(header[4]))
/**  */
#define CBT_HEADER_CONTROL(header)  ((unsigned char)(header[5]))
/** (reply) Reply ID */
#define CBT_HEADER_TYPECODE(header) ((unsigned char)(header[6]))
/** (reply) ơ */
#define CBT_HEADER_RESULT(header)   ((unsigned char)(header[7]))
/** ǡƬ */
#define CBT_HEADER_DATA(header)     (header + 6)
/*@}*/

/** @defgroup command_id CBT Packet ID
    @{*/
typedef enum ECBT_COMMAND_ID {
	ECBT_CMD_INIT            = 0x00, /**< Ԥ */
	ECBT_CMD_OPENCHANNEL     = 0x01, /**< ͥ򥪡ץ󤹤 */
	ECBT_CMD_CLOSECHANNEL    = 0x02, /**< ͥ򥯥 */
	ECBT_CMD_CREDIT          = 0x03, /**< 쥸åȤȯԤ */
	ECBT_CMD_CREDITREQUEST   = 0x04, /**< 쥸åȤ׵᤹ */
	ECBT_CMD_DEBIT           = 0x05, /**<  */
	ECBT_CMD_DEBITREQUEST    = 0x06, /**<  */
	ECBT_CMD_CONFIGSOCKET    = 0x07, /**<  */
	ECBT_CMD_EXIT            = 0x08, /**< CBT̿λ */
	ECBT_CMD_GETSOCKETID     = 0x09, /**<  */
	ECBT_CMD_GETSERVICENAME  = 0x0A, /**<  */
	ECBT_CMD_EPSONPACKING    = 0x45, /**< CBT̿򳫻Ϥ */
	ECBT_CMD_ERROR           = 0x7F, /**< 顼ѥå */
	ECBT_CMD_NONE            = 0xFF /**< ѤΥߡID */
} EcbtCommandId;		/**< ECBT_COMMAND_ID̾ */
/*@}*/

/** ޥɾ
 * command ID, reply ID, command size, reply size
 */
static unsigned char ecbtCommandTable[] = {
	ECBT_CMD_INIT,		 0x80, 0x08, 0x09,
	ECBT_CMD_OPENCHANNEL,	 0x81, 0x11, 0x10,
	ECBT_CMD_CLOSECHANNEL,	 0x82, 0x0A, 0x0A,
	ECBT_CMD_CREDIT,	 0x83, 0x0B, 0x0A,
	ECBT_CMD_CREDITREQUEST,	 0x84, 0x0D, 0x0C,
	ECBT_CMD_DEBIT,		 0x85, 0xFF, 0xFF,
	ECBT_CMD_DEBITREQUEST,	 0x86, 0xFF, 0xFF,
	ECBT_CMD_CONFIGSOCKET,	 0x87, 0xFF, 0xFF,
	ECBT_CMD_EXIT,		 0x88, 0x07, 0x08,
	ECBT_CMD_GETSOCKETID,	 0x89, 0x07, 0x09,
	ECBT_CMD_GETSERVICENAME, 0x8A, 0x08, 0x09,
	ECBT_CMD_EPSONPACKING,	 0xC5, 0x0A, 0x08,
	ECBT_CMD_ERROR,          0xFF, 0x1B, 0xFF,
	ECBT_CMD_NONE,           0xFF, 0xFF, 0xFF
};

/** @defgroup command_info_offset ޥɾơ֥Υեå
    @{*/
#define TABLE_COMMAND_OFFSET      0 /**< commmand ID */
#define TABLE_REPLY_OFFSET        1 /**< reply ID */
#define TABLE_COMMAND_SIZE_OFFSET 2 /**< command size */
#define TABLE_REPLY_SIZE_OFFSET   3 /**< reply size */
#define TABLE_PITCH               4 /**< ξؤεΥ */
/*@}*/

/** @defgroup command_results CBT Packet Results
    @{*/
#define CBT_RESULT_SUCCESS	0x00 /**< ޥ */
#define CBT_RESULT_BUSY 	0x01 /**< CBT̿򳫻ϤǤʤ(⤷
					ϡѥåȥ⡼ɤˤǤʤ)
					ޥԤϡ³׵
					Ǥ */
#define CBT_RESULT_BADVERSION	0x02 /**< ׵ӥϥݡȤ
					ʤӥѹơ
					³׵Ǥ */
#define CBT_RESULT_CHNOCLOSE	0x03 /**< CloseChannelޥɤϵݤ
					줿ͥ뤬CBTȥ
					ͥä᥯
					ʤ */
#define CBT_RESULT_NORESOURCE	0x04 /**< ׵ᤵ줿³ΤΥ꥽
					­Ƥ뤿ᡢ³
					ʤ */
#define CBT_RESULT_NOCONNECT	0x05 /**< ³ϵݤ줿 */
#define CBT_RESULT_CHREOPEN	0x06 /**< ˥ͥ뤬ץ󤵤
					뤿ᡢͥ륪ץ󤬵
					ݤ줿ͥ륪ץϷ
					³쥸åȤѲϤʤ
					 */
#define CBT_RESULT_CREDITOVER	0x07 /**< 쥸åȥСե
					쥸åȤ0xFFFF򥪡С
					쥸åȤϡξ֤
					롣 */
#define CBT_RESULT_CHNOTOPEN	0x08 /**< ޥɤϵݤ줿
					뤬ץ󤷤Ƥʤ */
#define CBT_RESULT_CHNOSUPPORT	0x09 /**< ꤵ줿̤ͥΤ̤
					ݡȤǤ롣 */
#define CBT_RESULT_NOSERVICE	0x0A /**< åIDؤΥӥ̾Υޥ
					ԥ󥰼ԡ(ꥵӥ̵
					) */
#define CBT_RESULT_NOINIT	0x0B /**< InitޥɤԤŬ
					ʻԤ򤷤ơInit
					ɬפ */
#define CBT_RESULT_SIZEOVER	0x0C /**< ѥåȥ̵Ǥ
					롣(0x0001-0x0005Ϥ) */
#define CBT_RESULT_SIZEZERO	0x0D /**< ѥåȥH/P
					0x0000Ǥä̿
					
					 */
/*@}*/


#if (defined ENABLE_SEND_ERROR_PACKET)
/** 顼ѥååȤȯԤɬפʥ顼 */
#define	CBT_ERR_INDEFINITE	0x80 /**< ΥѥåȤä
					 */
#define	CBT_ERR_UEPACKET	0x81 /**< 쥸åȤȯԤƤʤ
					ѥåȤä */
#define	CBT_ERR_UEREPLY		0x82 /**< ̤׵᤬̵ˡ
					ɥץ饤ä */
#define	CBT_ERR_BIGGER		0x83 /**< ꤷåȤФƼ
					᤿祵ۤǡ
					ΥѥåȤä */
#define	CBT_ERR_UEDATA		0x84 /**< Ƥʤͥ˥ǡ
					ѥåȤä */
#define	CBT_ERR_UERESULT	0x85 /**< ǽResultͤä
					ץ饤ѥåȤä */
#define	CBT_ERR_PIGGYCREDIT	0x86 /**< ǡѥå
					PiggyBackCreditˤäơ
					CreditΥСեȯ
					 */
#define	CBT_ERR_UNKNOWNREPLY	0x87 /**< Բǽ"CBT"ץ饤
					ä */
#endif /* (defined ENABLE_SEND_ERROR_PACKET) */

/** Ring BufferΥϥɥ */
typedef void * RingHandle;

/** Ring Buffer */
typedef struct {
	char *in;		/**< ǡ񤭹ߥɥ쥹 */
	char *out;		/**< ǡɤ߹ߥɥ쥹 */
	char *top;		/**< Ƭɥ쥹 */
	char *bottom;		/**< üɥ쥹 */
	size_t size;		/**< Хåե */
	int ioStat;		/**< 0=ľ 1=ľ */
} RingInfo;

/** Channel */
typedef struct  CH_INFO {
	int active;		/**< ͥOpen/Closeե饰 */
	int printerCredit;	/**< ץ󥿤Υ쥸åȻ */
	int hostCredit;		/**< ۥȤΥ쥸åȻ */
	int forwardSize;	/**< ѥåȤκ祵 */
	int reverseSize;	/**< ѥåȤκ祵 */
	RingHandle reply;	/**< ץ饤Ǽѥ󥰥Хåեϥɥ */
} ChInfo;

/** μ  */
typedef enum ESCAPE_OPERATION {
	ESCAPE_NOTOPERATION = 0,
	ESCAPE_PACKETWRITE,	/**< write */
	ESCAPE_PACKETREAD,	/**< read */
} EscapeOperation;

/**  */
typedef struct ECBT_ESCAPE_INFO
{
	EscapeOperation operation; /**< μ */
	ChID chId;		/**< channel ID */
	int command;		/**< ޥID */
	size_t length;		/**< ǡ  */
	char *buffer;		/**< ǡ */
} EcbtEscapeInfo;

/** ECBT */
typedef struct ECBT_INFO
{
	int fd;			/**< ץ󥿥ݡȤΥեǥץ */
	ChInfo control;		/**< ȥ󥶥ͥ */
	ChInfo command;		/**< ޥɥͥ */
	ChInfo data;		/**< ǡͥ */
	/* escape buffer */
	EcbtEscapeInfo *escapeInfo; /**< signalȯ */
} EcbtInfo;


#ifdef __cplusplu
extern "C" {
#endif /* __cplusplus */
	static void AlarmHandler(int);
	static void MsgDebug(const char *, ...);
	static int SafeUsleep(unsigned long);
	static RingHandle RingNew(size_t);
	static void RingDelete(RingHandle);
	static ssize_t RingWrite(RingHandle, char *, size_t);
	static ssize_t RingRead(RingHandle, char *, size_t);
	static ssize_t RingReadNotMove(RingHandle, char *, size_t);
	static ssize_t RingGetDataSize(RingHandle);
	static void RingClear(RingHandle);
	static int GetCommandSpec(int, int);
	static ChInfo *FindChInfo(EcbtInfo *, int);
	static int DummyRead(EcbtInfo *);
	static int SendMagic(EcbtInfo *);
	static int SendInit(EcbtInfo *);
	static int SendExit(EcbtInfo *);
#if (defined ENABLE_SEND_ERROR_PACKET)
	static int SendError(EcbtInfo *, int, int);
#endif /* (defined ENABLE_SEND_ERROR_PACKET) */
	static int SendOpenChannel(EcbtInfo *, int);
	static int SendCloseChannel(EcbtInfo *, int);
	static int SendCreditProvide(EcbtInfo *, int, int);
	static int SendCreditRequest(EcbtInfo *, int, int);
	static int SendData(EcbtInfo *, int, char *, size_t, ssize_t *);
	static int RecvData(EcbtInfo *, int, char *, size_t, ssize_t *);
	static int ReplyCheck(char *, int, unsigned char);
	static int WriteDoit(EcbtInfo *, ChID, char *, size_t, ssize_t *);
	static int ReadDoit(EcbtInfo *, ChID, char *, size_t, ssize_t *);
	static int CheckErrorPacket(char *);
	static int Begin(EcbtInfo *);
	static int Terminate(EcbtInfo *, EcbtErrCode);
	static int RecoverCommand(EcbtInfo *, EcbtEscapeInfo *);
	static int RecoverData(EcbtInfo *, EcbtEscapeInfo *);
	static int CreateEscapeBuffer(EcbtInfo *, EscapeOperation, char *, size_t);
	static int SetEscapeOperation(EcbtInfo *, EcbtCommandId, ChID);
	static int DeleteEscapeBuffer(EcbtInfo *);
	static int RealPacketRead(EcbtInfo *, char *, size_t, size_t *);
	static int ReadLoop(int, void *, size_t);
	static int WriteLoop(int, void *, size_t);
#ifdef __cplusplus
}
#endif /* __cplusplus */



/***************************************************************************
 * Signal Functions
 ***************************************************************************/

/** SIGALRMΥϥɥ
    @param sig	ʥID
    @return	̵
 */
static void
AlarmHandler(int sig)
{
  return;
}

/***************************************************************************
 * Debug Functions
 ***************************************************************************/

/** DebugstderrؽϤ
    @param fmt vfprintfϤեޥåʸ
    @param ... fmtѿ
    @return ̵
 */
static void
MsgDebug(const char *fmt, ...)
{
#ifdef DEBUG
	va_list ap;

	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	fflush(stderr);
#endif
	return;
}

/***************************************************************************
 * securer sleep
 ***************************************************************************/

/** Micro seconds Ԥİsleep
    @param usec sleep֡ñ̤ϥޥ
    @return Ǥ0򡢥顼ʤ0ʳ֤
*/
static int
SafeUsleep(unsigned long usec)
{
	int retval;
	struct timeval tv;

	/* Ԥ֤ */
	tv.tv_sec = usec / 1000000;
	tv.tv_usec = usec % 1000000;

	retval = select(1, NULL, NULL, NULL, &tv);

	return(retval);
}

/***************************************************************************
 * Ring Buffer Functions
 ***************************************************************************/

/** Ring BufferΥݥ󥿴֤εΥ */
#define	RING_SUB(ring, from, to) \
(((to) - (from) > 0) ? (to) - (from) : ((ring)->size) - ((from) - (to)))
/** Ring BufferΥݥ󥿤βû */
#define	RING_INC(ring, p1, val) \
(((p1) + (val) < (ring)->bottom) ? (p1) + (val) : ((p1) + ((val) - (ring)->size)))
/** Ring BufferΥݥ󥿤θ */
#define	RING_DEC(ring, p1, val) \
(((p1) - (val) >= (ring)->top) ? (p1) - (val) : (((p1) - (val)) - (ring)->size))
/** Ring BufferΥХåեǤTRUE֤ */
#define RING_EMPTY(ring) \
((ring)->in == (ring)->out && (ring)->ioStat == 1) ? TRUE : FALSE
/** Ring BufferΥХåեϰդǤTRUE֤ */
#define RING_FULL(ring) \
((ring)->in == (ring)->out && (ring)->ioStat == 0) ? TRUE : FALSE

/** Ring Buffer
    @param size Ring BufferΥ
    @return ǤкRing BufferΥϥɥ򡢥顼Ǥ
            NULL֤
 */
RingHandle
RingNew(size_t size)
{
	RingInfo *ring;

	ring = (RingInfo *)calloc(1, sizeof(RingInfo));
	if (ring == NULL) {
		return (NULL);
	}

	ring->top = (char *)calloc(size, sizeof(char));
	if (ring->top == NULL) {
		free(ring);
		return (NULL);
	}

	ring->size = size;
	ring->in = ring->out = ring->top;
	ring->bottom = ring->top + ring->size;
	ring->ioStat = 1; /* iostat==1  ǡХåեǤ򼨤 */

    return ((RingHandle)ring);
}


/** Ring Buffer
    @param hl Ring BufferΥϥɥ
    @return ̵
 */
void
RingDelete(RingHandle hl)
{
	RingInfo *ring = (RingInfo *)hl;

	if (ring) {
		if (ring->top) {
			free(ring->top);
		}
		free(ring);
	}
	return;
}


/** Ring Bufferؽ񤭹ߤԤ
    @param hl Ring BufferΥϥɥ
    @param buffer 񤭹ǡؤΥݥ
    @param size 񤭹ǡΥ
    @return ºݤ˳ǼǡΥ֤
 */
ssize_t
RingWrite(RingHandle hl, char *buffer, size_t size)
{
	RingInfo *ring =(RingInfo *)hl;

	if (size == 0 || RING_FULL(ring)) {
		return 0;
	}
	/* 񤭹ॵ */
	if (RING_SUB(ring, ring->in, ring->out) < size
	    && !RING_EMPTY(ring)) {
		/* ׵ᤵ줿ꡢ꡼ʤ */
		size = RING_SUB(ring, ring->in, ring->out);
	}

	if (RING_SUB(ring, ring->in, ring->bottom) >= size) {
		/* ǡϢ³꡼˵Ͽǽ */
		memcpy(ring->in, buffer, size);
	} else {
		/* ǡʬƵϿ */
		ssize_t tmpSize;

		tmpSize = RING_SUB(ring, ring->in, ring->bottom);
		memcpy(ring->in, buffer, tmpSize);
		memcpy(ring->top, buffer + tmpSize, size - tmpSize);
	}

	ring->in = RING_INC(ring, ring->in, size);
	ring->ioStat = 0;

	return size;
}


/** Ring Bufferɤ߹ߤԤɤǡRing Buffer
    @param hl Ring BufferΥϥɥ
    @param buffer ɤ߹ǡǼݥ
    @param size ɤ߹ǡΥ
    @return ºݤɤ߹ǡ֤
 */
ssize_t
RingRead(RingHandle hl, char *buffer, size_t size)
{
	RingInfo *ring =(RingInfo *)hl;

	size = RingReadNotMove(hl, buffer, size);
	if (size) {
		ring->out = RING_INC(ring, ring->out, size);
		ring->ioStat = 1;
	}

	return size;
}


/** Ring Buffer,ǡɤ߹ߤԤ
    @param hl Ring BufferΥϥɥ
    @param buffer ɤ߹ǡǼݥ
    @param size ɤ߹ǡΥ
    @return ºݤɤ߹ǡ֤
 */
ssize_t
RingReadNotMove(RingHandle hl, char *buffer, size_t size)
{
	RingInfo *ring =(RingInfo *)hl;

	if (size == 0 || RING_EMPTY(ring)) {
		return 0;
	}
	/* ɤ߹ॵ */
	if (RING_SUB(ring, ring->out, ring->in) < size
	    || RING_FULL(ring)) {
		/* ׵ᤵ줿ǡʤ */
		size = RING_SUB(ring, ring->out, ring->in);
	}

	if (RING_SUB(ring, ring->out, ring->bottom) >= size) {
		/* ǡϢ³꡼¸ߤ */
		memcpy(buffer, ring->out, size);
	} else {
		/* ǡΥ¸ߤ */
		ssize_t tmpSize;
		
		tmpSize = RING_SUB(ring, ring->out, ring->bottom);
		memcpy(buffer, ring->out, tmpSize);
		memcpy(buffer + tmpSize, ring->top, size - tmpSize);
	}
	
	return size;
}

/** Ring Bufferˡ߳ǼƤǡΥ
    @param hl Ring BufferΥϥɥ
    @return ߳ǼƤǡΥ֤
 */
ssize_t
RingGetDataSize(RingHandle hl)
{
	RingInfo *ring =(RingInfo *)hl;

	if (RING_EMPTY(ring)) {
		return 0;
	}

	return RING_SUB(ring, ring->out, ring->in);
}


/** Ring BufferΥǡ򥯥ꥢ
    @param hl Ring BufferΥϥɥ
    @return ̵
 */
void
RingClear(RingHandle hl)
{
	RingInfo *ring =(RingInfo *)hl;
	
	if (ring) {
		ring->in = ring->out = ring->top;
		ring->ioStat = 1;
	}

	return;
}

/** ޥID饳ޥɥ */
#define GET_COMMAND_SIZE(command) (GetCommandSpec(command, TABLE_COMMAND_SIZE_OFFSET))
/** ޥIDץ饤ID */
#define GET_REPLY(command) (GetCommandSpec(command, TABLE_REPLY_OFFSET))
/** ޥIDץ饤 */
#define GET_REPLY_SIZE(command) (GetCommandSpec(command, TABLE_REPLY_SIZE_OFFSET))

/** ޥID顢ޥɥ,ץ饤ID,ץ饤
    @param command ޥID
    @param offset ͤ򤹤
                  - TABLE_COMMAND_SIZE_OFFSET ޥɥ
                  - TABLE_REPLY_OFFSET ץ饤ID
                  - TABLE_REPLY_SIZE_OFFSET ץ饤
    @return ޥɥ,ץ饤ID,ץ饤֤0xFFϡǡ
            ¸ߤʤ򼨤
 */
static int
GetCommandSpec(int command, int offset)
{
	int i;

	for (i = 0; ecbtCommandTable[i * TABLE_PITCH] != ECBT_CMD_NONE; i++) {
		if ((int)ecbtCommandTable[i * TABLE_PITCH] == command) {
			break;
		}
	}
	
	return (int)ecbtCommandTable[i * TABLE_PITCH + offset];
}


/***************************************************************************
 * ECBT API
 ***************************************************************************/

/** ǥХȤ̿򳫻Ϥ
    @param *hl	handle֤Υݥ
    @param fd	ץ󥿤³ƤݡȤΥեǥץ
    @return	Error Code
*/
EcbtErrCode
EcbtOpen(EcbtHandle *hl, int fd)
{
	EcbtInfo *ecbtInfo;	
	int fileFlag;
	int result = ECBT_ERR_NOERR;

	if (hl == NULL || fd < 0)
		return ECBT_ERR_PARAM;

	ecbtInfo = (EcbtInfo *)calloc(1, sizeof(EcbtInfo));
	if (ecbtInfo == NULL) {
		return ECBT_ERR_NOMEM;
	}

	ecbtInfo->fd = fd;
	ecbtInfo->command.active = FALSE;
	ecbtInfo->data.active = FALSE;
	ecbtInfo->control.active = TRUE;
	ecbtInfo->control.reply = RingNew(REPLY_BUFFER_MAX);
	ecbtInfo->control.forwardSize = ECBT_TRANSACTION_FORWARD_MAX;
	ecbtInfo->control.reverseSize = ECBT_TRANSACTION_REVERSE_MAX;

	if (ecbtInfo->control.reply == NULL) {
		free(ecbtInfo);
		return ECBT_ERR_NOMEM;
	}

	/* Ūno-block-modeذܹԤ */
	fileFlag = fcntl(ecbtInfo->fd, F_GETFL);
	fcntl(ecbtInfo->fd, F_SETFL, fileFlag & O_NONBLOCK);

	DummyRead(ecbtInfo);

	/* ǥХǧ */
	result = SendMagic(ecbtInfo);
	if (result == ECBT_ERR_NOERR) {
		/* Initޥ */
		result = SendInit(ecbtInfo);
	}

	if (result != ECBT_ERR_NOERR) {
		RingDelete(ecbtInfo->control.reply);
		free(ecbtInfo);
		return result;
	}

	*hl = (EcbtHandle)ecbtInfo;
	return Terminate(ecbtInfo, result);
}

/** ǥХȤ̿λ
    @param hl	ECBT Handle
    @return	Error Code
*/
EcbtErrCode
EcbtClose(EcbtHandle hl)
{
	EcbtInfo *ecbtInfo = (EcbtInfo *)(hl);
	int result = ECBT_ERR_NOERR;

	if (hl == NULL) {
		return ECBT_ERR_PARAM;
	}

	result = Begin(ecbtInfo);
	if (result != ECBT_ERR_NOERR) {
		return Terminate(ecbtInfo, result);
	}

	/* ʲνǤϡ顼äƤϥɥβϹԤ */

	/* DATAͥ򥯥(顼̵) */
	result = EcbtChannelClose(hl, CHID_DATA);

	if (result != ECBT_ERR_NOERR) {
		/* COMMANDͥ򥯥(顼̵) */
		result = EcbtChannelClose(hl, CHID_COMMAND);
	}

	if (result != ECBT_ERR_NOERR) {
		/* Exitޥ */
		result = SendExit(ecbtInfo);
	}

	RingDelete(ecbtInfo->control.reply);
	free(ecbtInfo);
	
	return result;
}


/** ͥ򳫤
  @param hl	ECBT Handle
  @param chId	Channel ID
  @return	Error Code
 */
EcbtErrCode
EcbtChannelOpen(EcbtHandle hl, int chId)
{
	EcbtInfo *ecbtInfo =(EcbtInfo *)(hl);
	ChInfo *chInfo;
	int result = ECBT_ERR_NOERR;

	if (hl == NULL) {
		return ECBT_ERR_PARAM;
	}
	
	result = Begin(ecbtInfo);
	if (result != ECBT_ERR_NOERR) {
		return Terminate(ecbtInfo, result);
	}

	chInfo = FindChInfo(ecbtInfo, chId);
	if (chInfo == NULL) {
		return ECBT_ERR_PARAM;
	} else if (chInfo->active == TRUE) {
		return ECBT_ERR_CHREOPN;
	}

	if (chId == CHID_COMMAND) {
		chInfo->forwardSize = ECBT_COMMAND_FORWARD_MAX;
		chInfo->reverseSize = ECBT_COMMAND_REVERSE_MAX;
	} else if (chId == CHID_DATA) {
		chInfo->forwardSize = ECBT_DATA_FORWARD_MAX;
		chInfo->reverseSize = ECBT_DATA_REVERSE_MAX;
	} else {
		return ECBT_ERR_FATAL;
	}
	chInfo->reply = RingNew(REPLY_BUFFER_MAX);
	if (chInfo->reply == NULL) {
		return ECBT_ERR_NOMEM;
	}

	/* ChannelOpenޥ */
	result = SendOpenChannel(ecbtInfo, chId);
	if (result == ECBT_ERR_NOERR) {
		chInfo->active = TRUE;
	}
	return Terminate(ecbtInfo, result);
}

/** ͥĤ
  @param hl	ECBT Handle
  @param chId	Channel ID
  @return	Error Code
*/
EcbtErrCode
EcbtChannelClose(EcbtHandle hl, int chId)
{
	EcbtInfo *ecbtInfo =(EcbtInfo *)(hl);
	ChInfo *chInfo;
	int result = ECBT_ERR_NOERR;

	if (hl == NULL) {
		return ECBT_ERR_PARAM;
	}

	result = Begin(ecbtInfo);
	if (result != ECBT_ERR_NOERR) {
		return Terminate(ecbtInfo, result);
	}

	chInfo = FindChInfo(ecbtInfo, chId);
	if (chInfo == NULL) {
		return ECBT_ERR_PARAM;
	} if (chInfo->active == FALSE) {
		return ECBT_ERR_CHNOPN;
	}
	
	/* ChannelCloseޥ */
	result = SendCloseChannel(ecbtInfo, chId);
	if (result == ECBT_ERR_NOERR) {
		RingClear(chInfo->reply);
		chInfo->active = FALSE;
	}
	return Terminate(ecbtInfo, result);
}


/** ꤷͥ˥ǡ
  @param hl	ECBT Handle
  @param chId	Channel ID
  @param *wBuf	񤭹ǡǼХåե
  @param count	񤭹ॵ
  @param *wCount	ºݤ˽񤭹
  @return	Error Code
*/
EcbtErrCode
EcbtWrite(EcbtHandle hl, int chId, void *wBuf, size_t count, ssize_t *wCount)
{
	EcbtInfo *ecbtInfo =(EcbtInfo *)(hl);
	ChInfo *chInfo;
	int result = ECBT_ERR_NOERR;

	if (hl == NULL || wBuf == NULL || wCount == NULL) {
		return ECBT_ERR_PARAM;
	}

	*wCount = 0;

	result = Begin(ecbtInfo);
	if (result != ECBT_ERR_NOERR) {
		return Terminate(ecbtInfo, result);
	}

	chInfo = FindChInfo(ecbtInfo, chId);
	if (chInfo == NULL) {
		return ECBT_ERR_PARAM;
	} else if (chInfo->active != TRUE) {
		return ECBT_ERR_CHNOPN;
	}

	while (count > 0 && result == ECBT_ERR_NOERR) {
		int packCount;
		int packWCount;

		/* 쥸åȤ̵ͭå */
		if (chInfo->printerCredit <= 0)
		{
			/* CreditRequestޥ(쥸åȼ) */
			result = SendCreditRequest(ecbtInfo, chId, ECBT_PRINTER_CREDIT_MAX);
			if (result != ECBT_ERR_NOERR) {
				break;
			}
		}
		
		/* packetΥǡ */
		packCount = chInfo->forwardSize - CBT_HEADER_SIZE;
		packCount = (count < packCount) ? count : packCount;

		/* ǡ */
		result = SendData(ecbtInfo, chId, (char *)wBuf, packCount, &packWCount);
		wBuf += packWCount;
		count -= packWCount;
		*wCount += packWCount;
		if (result == ECBT_ERR_NOERR) {
			/* ۥȤΥ쥸åȤ */
			chInfo->printerCredit --;
		}
	}

	return Terminate(ecbtInfo, result);
}


/** ꤷͥ뤫ǡ
  @param hl	ECBT Handle
  @param chId	Channel ID
  @param *rBuf	ɤ߹ǡǼХåե
  @param count	ɤ߹ǡΥ
  @param *rCount	ºݤɤ߹ǡΥ
  @return	Error Code
*/
EcbtErrCode
EcbtRead(EcbtHandle hl, int chId, void *rBuf, size_t count, ssize_t *rCount)
{
	EcbtInfo *ecbtInfo =(EcbtInfo *)(hl);
	ChInfo *chInfo;
	int result = ECBT_ERR_NOERR;

	*rCount = 0;

	result = Begin(ecbtInfo);
	if (result != ECBT_ERR_NOERR) {
		return Terminate(ecbtInfo, result);
	}

	if (hl == NULL || rCount == NULL) {
		return ECBT_ERR_PARAM;
	}
	
	chInfo = FindChInfo(ecbtInfo, chId);
	if (chInfo == NULL) {
		return ECBT_ERR_FATAL;
	} else if (chInfo->active != TRUE) {
		return ECBT_ERR_CHNOPN;
	}
	
	/* 쥸åȤ̵ͭå */
	if (chInfo->hostCredit <= 0) {
		/* Creditޥ(쥸åȯ) */
		result = SendCreditProvide(ecbtInfo, chId, ECBT_QUANTITY_PROVIDING_CREDIT);
		
		if (result != ECBT_ERR_NOERR) {
			return Terminate(ecbtInfo, result);
		} else if (chInfo->hostCredit <= 0) {
			return ECBT_ERR_FATAL;
		}
	}
	/* ǡ */
	result = RecvData(ecbtInfo, chId, (char *)rBuf, count, (int *)rCount);
	return Terminate(ecbtInfo, result);
}


/***************************************************************************
 * ECBT Static Functions
 ***************************************************************************/

/** ͥIDͥ
 * @param *ecbtInfo	ECBT
 * @param chId		Channel ID
 * @return	顼ʤNULL򡢤ǤʤХͥؤΥݥ󥿤֤
 */
static ChInfo *
FindChInfo(EcbtInfo *ecbtInfo, int chId)
{
	ChInfo *chInfo;

	if (ecbtInfo == NULL) {
		return NULL;
	}

	if (chId == CHID_TRANSACTION) {
		chInfo = &(ecbtInfo->control);
	} else if (chId == CHID_COMMAND) {
		chInfo = &(ecbtInfo->command);
	} else if (chId == CHID_DATA) {
		chInfo = &(ecbtInfo->data);
	} else {
		chInfo = NULL;
	}

	return chInfo;
}

/** ǥХΥ꡼˻Ĥ줿ǡõ(ɤ)
 * @param *ecbtInfo	ECBT
 * @return		0
 */
static int
DummyRead(EcbtInfo *ecbtInfo)
{
	char buf[128];

	MsgDebug("in %s\n", __FUNCTION__);

	while (1) {
		int rc;
		void* sigfunc;

		sigfunc = signal(SIGALRM, AlarmHandler);
		alarm(ECBT_IO_TIMEOUT);

		rc = read(ecbtInfo->fd, buf, 128);

		alarm(0);
		signal(SIGALRM, sigfunc);
		
		if (rc < 128) {
			break;
		}
	}
	
	return 0;
}


/** ̿ϥޥɤ
 * @param *ecbtInfo	ECBT
 * @return		Error Code
 */
static int
SendMagic(EcbtInfo *ecbtInfo)
{
	int result = ECBT_ERR_NOERR;
	int wSize = 0;
	char buf[REPLY_BUFFER_MAX];
	int rSize = 0;

	/*
	  char buf[REPLY_BUFFER_MAX];
	  int rSize = 0;
	*/
	char data[] = {
		/* header */
		0x00, 0x00, 0x00, 0x1b, 0x01, 0x40,
		/* data */
		0x45, 0x4a, 0x4c, 0x20, 0x31, 0x32, 0x38, 0x34,
		0x2e, 0x34, 0x0a, 0x40, 0x45, 0x4a, 0x4c, 0x0a,
		0x40, 0x45, 0x4a, 0x4c, 0x0a
	};

	MsgDebug("in %s\n", __FUNCTION__);

	result = WriteDoit(ecbtInfo, CHID_TRANSACTION, data, sizeof(data), &wSize);
	if (result == ECBT_ERR_NOERR) {
		SafeUsleep(ECBT_IO_WAITTIME);
		result = ReadDoit(ecbtInfo, CHID_TRANSACTION, buf, GET_REPLY_SIZE(ECBT_CMD_EPSONPACKING), &rSize);
	}

	if (result == ECBT_ERR_NOERR) {
		result = ReplyCheck(buf, rSize, ECBT_CMD_EPSONPACKING);
	}

	if (result == ECBT_ERR_SIGNAL) {
		result = SetEscapeOperation(ecbtInfo, ECBT_CMD_EPSONPACKING, CHID_TRANSACTION);
	}
	return result;
}


/** ̿ޥɤ
 * @param *ecbtInfo	ECBT
 * @return		Error Code
 */
static int
SendInit(EcbtInfo *ecbtInfo)
{
	int result = ECBT_ERR_NOERR;
	int wSize = 0;
	char buf[REPLY_BUFFER_MAX];
	int rSize = 0;

	char data[] = {
		/* header */
		0x00, 0x00, 0x00, 0x08, 0x01, 0x00,
		/* data */
		0x00, 0x10
	};

	MsgDebug("in %s\n", __FUNCTION__);

	result = WriteDoit(ecbtInfo, CHID_TRANSACTION, data, sizeof(data), &wSize);

	if (result == ECBT_ERR_NOERR) {
		SafeUsleep(ECBT_IO_WAITTIME);
		result = ReadDoit(ecbtInfo, CHID_TRANSACTION, buf, GET_REPLY_SIZE(ECBT_CMD_INIT), &rSize);
	}

	if (result == ECBT_ERR_NOERR) {
		result = ReplyCheck(buf, rSize, ECBT_CMD_INIT);
	}

	if (result == ECBT_ERR_SIGNAL) {
		result = SetEscapeOperation(ecbtInfo, ECBT_CMD_INIT, CHID_TRANSACTION);
	}
	return result;
}

/** ̿λޥɤ
 * @param *ecbtInfo	ECBT
 * @return		Error Code
 */
static int
SendExit(EcbtInfo *ecbtInfo)
{
	int result = ECBT_ERR_NOERR;
	int wSize = 0;
	char buf[REPLY_BUFFER_MAX];
	int rSize = 0;

	char data[] = {
		/* header */
                0x00, 0x00, 0x00, 0x07, 0x01, 0x00,
		/* data */
		0x08
	};

	MsgDebug("in %s\n", __FUNCTION__);

	result = WriteDoit(ecbtInfo, CHID_TRANSACTION, data, sizeof(data), &wSize);
	if (result == ECBT_ERR_NOERR) {
		SafeUsleep(ECBT_IO_WAITTIME);
		result = ReadDoit(ecbtInfo, CHID_TRANSACTION, buf, GET_REPLY_SIZE(ECBT_CMD_EXIT), &rSize);
	}

	if (result == ECBT_ERR_NOERR) {
		result = ReplyCheck(buf, rSize, ECBT_CMD_EXIT);
	}

	if (result == ECBT_ERR_SIGNAL) {
		result = SetEscapeOperation(ecbtInfo, ECBT_CMD_EXIT, CHID_TRANSACTION);
	}

	return result;
}

#if (defined ENABLE_SEND_ERROR_PACKET)
/** 顼ѥåȤ
 * @param *ecbtInfo	ECBT
 * @param chId		Channel ID
 * @param code		Packet Error Code
 * @return		Error Code
 * @todo 顼ѥåȤȯ\n
 *     顼ѥåȤϡȯ顼ɤߤ,Terminateˤ
 *     ȯԤ롣顼ȯͥIDɬ\n
 *     \n
 *     顼ѥåȤ顢Ƥ̿򥯥ʤФʤʤ
 */
static int
SendError(EcbtInfo *ecbtInfo, int chId, int code)
{
	int result;
	int wSize = 0;

	char data[] = {
		/* header */
                0x00, 0x00, 0x00, 0x0a, 0x01, 0x00,
		/* data */
		0x7f, 0x00, 0x00, 0x00
	};

	MsgDebug("in %s\n", __FUNCTION__);


	data[7] = (char)(chId);
	data[8] = (char)(chId);
	data[9] = (char)(code);

	result = WriteDoit(ecbtInfo, CHID_TRANSACTION, data, sizeof(data), &wSize);
	if (result == ECBT_ERR_SIGNAL) {
		result = SetEscapeOperation(ecbtInfo, ECBT_CMD_ERROR, CHID_TRANSACTION);
	}
	return result;
}
#endif /* (defined ENABLE_SEND_ERROR_PACKET) */

/** ͥ򳫤ޥɤ
  @param *ecbtInfo	ECBT
  @param chId		Channel ID
  @return		Error Code
*/
static int
SendOpenChannel(EcbtInfo *ecbtInfo, int chId)
{
	int result = ECBT_ERR_NOERR;
	int wSize = 0;
	char buf[REPLY_BUFFER_MAX];
	int rSize = 0;
	ChInfo *chInfo;

	char data[] = {
		/* header */
                0x00, 0x00, 0x00, 0x11, 0x01, 0x00,
		/* data */
		0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00
	};


	MsgDebug("in %s\n", __FUNCTION__);

	/* ͥΥåϾ̤ǹԤäƤϤʤΤǡԤ
	   FatalError */
	chInfo = FindChInfo(ecbtInfo, chId);
	if (!chInfo) {
		return ECBT_ERR_FATAL;
	}

	data[7]  = (char)(chId);
	data[8]  = (char)(chId);
	data[9]  = (char)((chInfo->forwardSize & 0xff00) >> 8);
	data[10] = (char)(chInfo->forwardSize & 0x00ff);
	data[11] = (char)((chInfo->reverseSize & 0xff00) >> 8);
	data[12] = (char)(chInfo->reverseSize & 0x00ff);

	result = WriteDoit(ecbtInfo, CHID_TRANSACTION, data, sizeof(data), &wSize);
	if (result == ECBT_ERR_NOERR) {
		SafeUsleep(ECBT_IO_WAITTIME);
		result = ReadDoit(ecbtInfo, CHID_TRANSACTION, buf, GET_REPLY_SIZE(ECBT_CMD_OPENCHANNEL), &rSize);
	}

	if (result == ECBT_ERR_NOERR) {
		result = ReplyCheck(buf, rSize, ECBT_CMD_OPENCHANNEL);
	}

	if (result == ECBT_ERR_NOERR) {
		chInfo->forwardSize = ((int)(buf[10]) << 8) + buf[11];
		chInfo->reverseSize = ((int)(buf[12]) << 8) + buf[13];
		chInfo->hostCredit = ((int)(buf[14]) << 8) + buf[15];
	}
	
	if (result == ECBT_ERR_SIGNAL) {
		result = SetEscapeOperation(ecbtInfo, ECBT_CMD_OPENCHANNEL, CHID_TRANSACTION);
	}
	return  result;
}


/** ͥĤ륳ޥɤ
  @param *ecbtInfo	ECBT
  @param chId		Channel ID
  @return		Error Code
*/
static int
SendCloseChannel(EcbtInfo *ecbtInfo, int chId)
{
	int result = ECBT_ERR_NOERR;
	int wSize = 0;
	char buf[REPLY_BUFFER_MAX];
	int rSize = 0;

	char data[] = {
		/* header */
                0x00, 0x00, 0x00, 0x0a, 0x01, 0x00,
		/* data */
		0x02, 0x00, 0x00, 0x00
	};

	data[7]  = (char)(chId);
	data[8]  = (char)(chId);


	MsgDebug("in %s\n", __FUNCTION__);

	result = WriteDoit(ecbtInfo, CHID_TRANSACTION, data, sizeof(data), &wSize);
	if (result == ECBT_ERR_NOERR) {
		SafeUsleep(ECBT_IO_WAITTIME);
		result = ReadDoit(ecbtInfo, CHID_TRANSACTION, buf, GET_REPLY_SIZE(ECBT_CMD_CLOSECHANNEL), &rSize);
	}

	if (result == ECBT_ERR_NOERR) {
		result = ReplyCheck(buf, rSize, ECBT_CMD_CLOSECHANNEL);
	}

	if (result == ECBT_ERR_SIGNAL) {
		result = SetEscapeOperation(ecbtInfo, ECBT_CMD_CLOSECHANNEL, CHID_TRANSACTION);
	}
	return  result;
}


/** ץ󥿤Ф쥸åȤȯԤ륳ޥɤ
  @param *ecbtInfo	ECBT
  @param chId		Channel ID
  @param credit		ȯԤ륯쥸åȿ
  @return		Error Code
*/
static int
SendCreditProvide(EcbtInfo *ecbtInfo, int chId, int credit)
{
	int result = ECBT_ERR_NOERR;
	int wSize = 0;
	char buf[REPLY_BUFFER_MAX];
	int rSize = 0;
	ChInfo *chInfo;

	char data[] = {
		/* header */
                0x00, 0x00, 0x00, 0x0b, 0x01, 0x00,
		/* data */
		0x03, 0x00, 0x00, 0x00, 0x01
	};

	MsgDebug("in %s\n", __FUNCTION__);

	data[7]  = (char)(chId);
	data[8]  = (char)(chId);
	data[9]  = (char)((credit & 0xff00) >> 8);
	data[10] = (char)(credit & 0x00ff);

	/* ͥΥåϾ̤ǹԤäƤϤʤΤǡԤ
	   FatalError */
	chInfo = FindChInfo(ecbtInfo, chId);
	if (!chInfo) {
		return ECBT_ERR_FATAL;
	}

	result = WriteDoit(ecbtInfo, CHID_TRANSACTION, data, sizeof(data), &wSize);
	if (result == ECBT_ERR_NOERR) {
		SafeUsleep(ECBT_IO_WAITTIME);
		result = ReadDoit(ecbtInfo, CHID_TRANSACTION, buf, GET_REPLY_SIZE(ECBT_CMD_CREDIT), &rSize);
		MsgDebug("%s(%d) %s: after ReadDoit(): result = %d\n",
				 __FILE__, __LINE__, __FUNCTION__, result);
	}

	if (result == ECBT_ERR_NOERR) {
		result = ReplyCheck(buf, rSize, ECBT_CMD_CREDIT);
	}

	if (result == ECBT_ERR_NOERR) {
		/* 쥸åȤȯԤΤǡ󥿤ɲä */
		chInfo->hostCredit += credit;
	}

	if (result == ECBT_ERR_SIGNAL) {
		result = SetEscapeOperation(ecbtInfo, ECBT_CMD_CREDIT, CHID_TRANSACTION);
	}
	return  result;
}


/** ץ󥿤ء쥸åȤ׵᤹륳ޥɤ
  @param *ecbtInfo	ECBT
  @param chId		Channel ID
  @param credit		׵᤹륯쥸åȿ
  @return		Error Code
*/
static int
SendCreditRequest(EcbtInfo *ecbtInfo, int chId, int credit)
{
	int result = ECBT_ERR_NOERR;
	int wSize = 0;
	char buf[REPLY_BUFFER_MAX];
	int rSize = 0;
	ChInfo *chInfo;

	char data[] = {
		/* header */
		0x00, 0x00, 0x00, 0x0d, 0x01, 0x00,
		/* data */
		0x04, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff
	};

	MsgDebug("in %s\n", __FUNCTION__);

	data[7]  = (char)(chId);
	data[8]  = (char)(chId);
	data[9]  = (char)((ECBT_HOST_CREDIT_MAX & 0xff00) >> 8);
	data[10] = (char)(ECBT_HOST_CREDIT_MAX & 0x00ff);

	/* ͥΥåϾ̤ǹԤäƤϤʤΤǡԤ
	   FatalError */
	chInfo = FindChInfo(ecbtInfo, chId);
	if (!chInfo) {
		return ECBT_ERR_FATAL;
	}

	result = WriteDoit(ecbtInfo, CHID_TRANSACTION, data, sizeof(data), &wSize);
	if (result == ECBT_ERR_NOERR) {
		SafeUsleep(ECBT_IO_WAITTIME);
		result = ReadDoit(ecbtInfo, CHID_TRANSACTION, buf, GET_REPLY_SIZE(ECBT_CMD_CREDITREQUEST), &rSize);
		MsgDebug("%s(%d) %s: after ReadDoit(): result = %d\n",
				 __FILE__, __LINE__, __FUNCTION__, result);
	}

	if (result == ECBT_ERR_NOERR) {
		result = ReplyCheck(buf, rSize, ECBT_CMD_CREDITREQUEST);
	}

	if (result == ECBT_ERR_NOERR) {
		MsgDebug("Add credit +%d\n", (int)(buf[10] << 8) + buf[11]);
		/* 쥸åȤ򥫥󥿤ɲä */
		chInfo->printerCredit += (int)(buf[10] << 8) + buf[11];

		/* 쥸å0ϥǥХӥ֤ˤ򼨤 */
		if (chInfo->printerCredit == 0) {
			return ECBT_ERR_BUSY;
		}
	}

	if (result == ECBT_ERR_SIGNAL) {
		result = SetEscapeOperation(ecbtInfo, ECBT_CMD_CREDITREQUEST, CHID_TRANSACTION);
	}

	MsgDebug("%s(%d) %s: end result = %d\n",
			 __FILE__, __LINE__, __FUNCTION__, result);
	return  result;
}

/** ꤷͥФǡ
  @param *ecbtInfo	ECBT
  @param chId		Channel ID
  @param *buffer	ǡǼХåեؤΥɥ쥹
  @param count		ǡ
  @param *wCount	ºݤǡǼ륢ɥ쥹
  @return		Error Code
 */
static int
SendData(EcbtInfo *ecbtInfo, int chId, char *buffer, size_t count, ssize_t *wCount)
{
	int result = ECBT_ERR_NOERR;
	int packetSize;
	char tmpBuffer[ECBT_DATA_FORWARD_MAX];
	int i;

	static int debugCounter = 0;

	MsgDebug("in %s: %d\n", __FUNCTION__, debugCounter);
	debugCounter++;


	packetSize = count + CBT_HEADER_SIZE;
	
	tmpBuffer[0] =(char)(chId);
	tmpBuffer[1] =(char)(chId);
	tmpBuffer[2] =(char)((packetSize & 0xff00) >> 8);
	tmpBuffer[3] =(char)(packetSize & 0x00ff);
	tmpBuffer[4] = 0;
	/* üǡ
	   ܥ⥸塼ˤ̿ǤϡĤΥѥåȤ٤write
	   (ʣʬʤ) */
	tmpBuffer[5] = 0x01;

	for(i = 0; i < count; i++) {
		tmpBuffer[CBT_HEADER_SIZE + i] = buffer[i];
	}

	result = WriteDoit(ecbtInfo, CHID_TRANSACTION, tmpBuffer, packetSize, wCount);
	*wCount -= CBT_HEADER_SIZE;

	if (result == ECBT_ERR_SIGNAL) {
		result = SetEscapeOperation(ecbtInfo, ECBT_CMD_NONE, chId);
	}
	return result;
}

/** ꤷͥ뤫顢ǡ
  @param *ecbtInfo	ECBT
  @param chId		Channel ID
  @param *buffer	ǡǼХåեؤΥɥ쥹
  @param count		ǡ
  @param *rCount	ºݤ˼ǡǼ륢ɥ쥹
  @return		Error Code
  */
static int
RecvData(EcbtInfo *ecbtInfo, int chId, char *buffer, size_t count, ssize_t *rCount)
{
	int result = ECBT_ERR_NOERR;
	char *tmpBuf;

	MsgDebug("in %s\n", __FUNCTION__);

	/* ѥåȤΥإåʬΰɲ */
	if (buffer == NULL) {
		tmpBuf = NULL;
	} else {
		tmpBuf = malloc(count + CBT_HEADER_SIZE);
		if (tmpBuf == NULL) {
			return ECBT_ERR_NOMEM;
		}
	}

	result = ReadDoit(ecbtInfo, chId, tmpBuf, count + CBT_HEADER_SIZE, rCount);
	MsgDebug("%s(%d) %s: after ReadDoit(): result = %d\n",
			 __FILE__, __LINE__, __FUNCTION__, result);
	*rCount = (*rCount < CBT_HEADER_SIZE) ? 0 : *rCount - CBT_HEADER_SIZE;
	if (result != ECBT_ERR_NOERR) {
		if (tmpBuf != NULL) {
			free(tmpBuf);
		}

		if (result == ECBT_ERR_SIGNAL) {
			result = SetEscapeOperation(ecbtInfo, ECBT_CMD_NONE, chId);
		}
		return result;
	}

	/* (TBD)إåγǧ */
	
	/* ǡ */
	if (tmpBuf != NULL) {
		memcpy(buffer, tmpBuf + CBT_HEADER_SIZE, count);
		free(tmpBuf);
	}
	return result;
}


/** ץ󥿤Υץ饤Ƥå(ȥ󥶥ͥΤ)
  @param *buffer	ץ饤ǼХåե
  @param bufferSize	ХåեΥ
  @param command	command ID
  @return		Error Code
 */
static int
ReplyCheck(char *buffer, int bufferSize, unsigned char command)
{
	int replySize;

	/*
	 * NOTE: ȥ󥶥ͥΥ쥸åȤϡۤΤ
	 * 礬ƤΤȤPiggyback creditγǧϹԤʤ
	 */
	
	/* ѥ᡼å */
	if (bufferSize < CBT_HEADER_SIZE
	    || CBT_HEADER_PSID(buffer) != CHID_TRANSACTION
	    || CBT_HEADER_SSID(buffer) != CHID_TRANSACTION) {
		return ECBT_ERR_UNCONF;
	}

	/* ޥɤμǧ */
	if(CBT_HEADER_TYPECODE(buffer) != GET_REPLY(command)) {
		return ECBT_ERR_UNCONF;
	}

	/* ǧ */
	replySize = GET_REPLY_SIZE(command);
	if (replySize != 0xff && replySize != CBT_HEADER_LENGTH(buffer)) {
		return ECBT_ERR_UNCONF;
	}

	/* resultǧ */
#if (defined DEBUG)
	if (CBT_HEADER_RESULT(buffer) != CBT_RESULT_SUCCESS) {
		MsgDebug("ECBT: Error Packet (%d)\n", (int)buffer[7]);
	}
#endif /* DEBUG */

	switch(CBT_HEADER_RESULT(buffer)) {
	case CBT_RESULT_SUCCESS:
		/* ޥɤ */
		break;

	case CBT_RESULT_CHREOPEN:
	case CBT_RESULT_CHNOTOPEN:
		/* signalˤꡢܥ顼֤ǽ롣³Ԥϲ
		   ǽ */
		MsgDebug("%s:%d:*** DEBUG IT ***", __FUNCTION__, __LINE__);
		break;

	case CBT_RESULT_BUSY:
	case CBT_RESULT_NORESOURCE:
	case CBT_RESULT_NOINIT:
		/* ƻԲǽ */
		return ECBT_ERR_BUSY;
		break;

	case CBT_RESULT_BADVERSION:
	case CBT_RESULT_CHNOSUPPORT:
	case CBT_RESULT_NOSERVICE:
		/* ݡȳνԤ줿 */
		return ECBT_ERR_PARAM;
		break;

	case CBT_RESULT_CHNOCLOSE:
	case CBT_RESULT_SIZEOVER:
	case CBT_RESULT_SIZEZERO:
		/* ȯƤϤʤʤ顼 */
		MsgDebug("%s:%d:*** DEBUG IT ***", __FUNCTION__, __LINE__);
		return ECBT_ERR_FATAL;
		break;

	case CBT_RESULT_NOCONNECT:
		/* ³ */
		return ECBT_ERR_NOCONNECT;
		break;

	default:
		return ECBT_ERR_UNCONF;
	}

	return  ECBT_ERR_NOERR;
}


/** ǡν񤭹ߤԤ
 * @param ecbtInfo	ECBT
 * @param chId		Channel ID
 * @param buffer	񤭹ǡǼХåե
 * @param length	񤭹ǡΥ
 * @param count		ºݤ˽񤭹ǡΥ
 * @return ErrorCode
*/
static int
WriteDoit(EcbtInfo *ecbtInfo, ChID chId, char *buffer, size_t length, ssize_t *count)
{
	/* signalǧ */
	if (signalCatchedFlag != 0) {
		/* signalȯ */
		*count = length;
		return CreateEscapeBuffer(ecbtInfo, ESCAPE_PACKETREAD, buffer, length);
	}

	
	*count = WriteLoop(ecbtInfo->fd, buffer, length);
	if (*count < 0) {
#if (defined DEBUG)
		MsgDebug("write error: %s(%d)\n", strerror(errno), errno);
#endif
		*count = 0;
		/* 顼³顼 */
		return ECBT_ERR_NOCONNECT;
	}
	if (*count < length) {
		/* signalȯ */
		return CreateEscapeBuffer(ecbtInfo, ESCAPE_PACKETWRITE, buffer + *count, length - *count);
	}
	
	return ECBT_ERR_NOERR;
}


/** ǡɹԤ
 * @param ecbtInfo	ECBT
 * @param chId		Channel ID
 * @param buffer	ɤ߹ǡǼХåե
 * @param length	ɤ߹ǡΥ
 * @param count		ºݤɤ߹ǡΥ
 * @return ErrorCode
*/
static int
ReadDoit(EcbtInfo *ecbtInfo, ChID chId, char *buffer, size_t length, ssize_t *count)
{
	ChInfo *chInfo;

	MsgDebug("%s(%d) %s: length = %d\n",
			   __FILE__, __LINE__, __FUNCTION__, length);

	*count = 0;

	chInfo = FindChInfo(ecbtInfo, chId);
	/* ͥIDΥåϾ̤ǹԤäƤϤʤΤǡԤ
	   FatalError */
	if (chInfo == NULL) {
		return ECBT_ERR_FATAL;
	}

	for (;;) {
		int packetReadCount;
		int result;

		/*  ꤵ줿ͥ˴˾󤬤ϡ鸡Ф  */
		if (RingGetDataSize(chInfo->reply) > 0) {
			int packetSize; /* size of packet */
			int size = 0;
			char header[CBT_HEADER_SIZE];

			size = RingReadNotMove(chInfo->reply, header, CBT_HEADER_SIZE);
			if (size == CBT_HEADER_SIZE) {
				packetSize = CBT_HEADER_LENGTH(header);
				
				if (length < packetSize) {
					/* ɤ߹ߥѥåȤ礭 */
					return ECBT_ERR_UNCONF;
				}
				
				/* bufferNULLξϡΤߤ֤(إåդ) */
				if (buffer == NULL) {
					*count = packetSize;
					break; /* readλ */
				}
				
				/* ɤ߹ǡˤϡإåޤޤƤ */
				*count = RingRead(chInfo->reply, buffer, packetSize);
				if (*count != packetSize) {

					return ECBT_ERR_FATAL;
				}
				break;	/* readλ */
			}
		}

		/* ץ󥿤read */
		result = RealPacketRead(ecbtInfo, NULL, 0, &packetReadCount);
		MsgDebug("%s(%d) %s: after RealPacketRead(): result = %d, packetReadCount = %d\n",
				   __FILE__, __LINE__, __FUNCTION__, result, packetReadCount);
		if (result != ECBT_ERR_NOERR) {
			return result;
		}
	}
	
	return ECBT_ERR_NOERR;
}

/** ѥåȤ顼ѥåȤǤ뤫ȽǤ
 * @param buf	ѥåȤǼХåեؤΥݥ
 * @return	ErrorCode
*/
static int
CheckErrorPacket(char *buf)
{
	if (CBT_HEADER_PSID(buf) != CHID_TRANSACTION
	    || CBT_HEADER_SSID(buf) != CHID_TRANSACTION
	    || CBT_HEADER_LENGTH(buf) != GET_COMMAND_SIZE(ECBT_CMD_ERROR)
	    || CBT_HEADER_TYPECODE(buf) != ECBT_CMD_ERROR) {
		/* 顼ѥåȤǤϤʤ */
		
		return ECBT_ERR_NOERR;
	}

	/* 顼ѥåȤ */
	MsgDebug("Received error packet: PSID=%d, SSID=%d ERROR=%d\n",
		 (int)buf[7], (int)buf[8], (int)buf[9]);
	return ECBT_ERR_UNCONF;
}

/** Ǥ줿ޥɤν³
 * @param ecbtInfo	ECBT
 * @param escapeInfo	
 */
static int
RecoverCommand(EcbtInfo *ecbtInfo, EcbtEscapeInfo *escapeInfo)
{
	char reply[ECBT_REPLY_MAX];
	int length = 0;
	int count;
	ChInfo *chInfo;
	int result = ECBT_ERR_NOERR;
	int command;
	int operation;
	

	command = escapeInfo->command;
	operation = escapeInfo->operation;

	/* ͥΥå */
	chInfo = FindChInfo(ecbtInfo, CHID_TRANSACTION);
	if (!chInfo) {
		return ECBT_ERR_CHNOPN;
	}

	/* write */
	if (operation == ESCAPE_PACKETWRITE) {
		/* writeʤΤǡޤϽ */
		result = WriteDoit(ecbtInfo, CHID_TRANSACTION, escapeInfo->buffer, escapeInfo->length, &count);
		if (result != ECBT_ERR_NOERR) {
			if (result == ECBT_ERR_SIGNAL) {
				/* signalȯ */
				result = CreateEscapeBuffer(ecbtInfo, operation, escapeInfo->buffer + count, escapeInfo->length - count);
				if (result == ECBT_ERR_SIGNAL) {
					result = SetEscapeOperation(ecbtInfo, command, CHID_TRANSACTION);
				}
			}
			return result;
		}
		length = 0;
	} else if (operation == ESCAPE_PACKETREAD) {
		/* writeϽλƤ */
		memcpy(reply, escapeInfo->buffer, escapeInfo->length);
		length = escapeInfo->length;
	}
	
	
	/* Error Packetξϡǽλ */
	if (command == ECBT_CMD_ERROR) {
		return ECBT_ERR_NOERR;
	}

	/* read */
	result = RealPacketRead(ecbtInfo, reply, length, &count);
	if (result == ECBT_ERR_NOERR) {
		result = ReadDoit(ecbtInfo, CHID_TRANSACTION, reply, GET_COMMAND_SIZE(command), &count);
		MsgDebug("%s(%d) %s: after ReadDoit(): result = %d\n",
				 __FILE__, __LINE__, __FUNCTION__, result);
	}

	if (result != ECBT_ERR_NOERR) {
		if (result == ECBT_ERR_SIGNAL) {
			/* signalȯ */
			result = SetEscapeOperation(ecbtInfo, command, CHID_TRANSACTION);
		}
		return result;
	}
	
	result = ReplyCheck(reply, count, command);
	if (result) {
		return result;
	}
	
	if (command == ECBT_CMD_CREDIT) {
		if (reply[7] != 0x00) {
			/* 쥸åȤȯԤPSIDۤʤ */
			MsgDebug("Reply of send_credit_provide is unjust. (0x%02x)\n", reply[7]);
			return ECBT_ERR_UNCONF;
		}
		
		/* 쥸åȤȯԤΤǡ󥿤ɲä */
		/* (TBD) ǤSendCreditProvideǻꤷ쥸åȿ
		   ݻƤ٤ܥ⥸塼ǰ٤󶡤륯쥸
		   ȿɬECBT_QUANTITY_PROVIDING_CREDITʤΤǡǤľ
		   ͤû롣 */
		chInfo->hostCredit += ECBT_QUANTITY_PROVIDING_CREDIT;

	} else if (command == ECBT_CMD_CREDITREQUEST) {
		if (reply[7] != 0x00) {
			/* 쥸åȤ׵ᤷPSIDۤʤ */
			MsgDebug("Reply of send_credit_request is unjust. (0x%02x)\n", reply[7]);
			return ECBT_ERR_UNCONF;
		}
		
		MsgDebug("Add credit +%d\n", (int)(reply[10] << 8) + reply[11]);
		/* 쥸åȤ򥫥󥿤ɲä */
		chInfo->printerCredit += (int)(reply[10] << 8) + reply[11];

	} else if (command == ECBT_CMD_EXIT) {
		/* ̵ */
	}
	return result;
}

/** Ǥ줿ǡ³
 * @param ecbtInfo	ECBT
 * @param escapeInfo	
 */
static int
RecoverData(EcbtInfo *ecbtInfo, EcbtEscapeInfo *escapeInfo)
{
	ChInfo *chInfo;
	int count;
	int result = ECBT_ERR_NOERR;
	int operation = escapeInfo->operation;
	int length = escapeInfo->length;
	int chId = escapeInfo->chId;

	/* ͥΥå */
	chInfo = FindChInfo(ecbtInfo, chId);
	if (!chInfo) {
		return ECBT_ERR_CHNOPN;
	}

	if (operation == ESCAPE_PACKETWRITE) {
		/* writeξϽ³롣 */
		result = WriteDoit(ecbtInfo, chId, escapeInfo->buffer, length, &count);
		if (result == ECBT_ERR_NOERR) {
			/* ۥȤΥ쥸åȤ */
			chInfo->printerCredit --;
		}
	} else if (operation == ESCAPE_PACKETREAD) {
		char reply[ECBT_REPLY_MAX];
		int count = 0;

		result = RealPacketRead(ecbtInfo, escapeInfo->buffer, length, &count);
		if (result == ECBT_ERR_NOERR) {
			length = count;
			result = ReadDoit(ecbtInfo, CHID_TRANSACTION, reply, length, &count);
			MsgDebug("%s(%d) %s: after ReadDoit(): result = %d\n",
					 __FILE__, __LINE__, __FUNCTION__, result);
		}
		if (result == ECBT_ERR_SIGNAL) {
			result = SetEscapeOperation(ecbtInfo, ECBT_CMD_NONE, chId);
		}
	}

	return result;
}

/** ׵ᤵ줿ˡԤ
 *  - Ԥ⤷ʤread/write
 * @param ecbtInfo	ECBT
 * @note ܴؿϡEcbtOpen˥뤹ɬפϤʤ
 */
static int
Begin(EcbtInfo *ecbtInfo)
{
	int result = ECBT_ERR_NOERR;

	/* 󤬤С³ */
	if (ecbtInfo->escapeInfo != NULL) {
		int command;
		EcbtEscapeInfo escapeInfo = *(ecbtInfo->escapeInfo);
		
		if (escapeInfo.length > 0) {
			escapeInfo.buffer = (char *)malloc(escapeInfo.length);
			if (escapeInfo.buffer == NULL) {
				return ECBT_ERR_NOMEM;
			}
			memcpy(escapeInfo.buffer, ecbtInfo->escapeInfo, escapeInfo.length);
		}
		
		/* ͽᡢΥǡϺ */
		DeleteEscapeBuffer(ecbtInfo);

		switch (command) {
		case ECBT_CMD_CREDIT:
		case ECBT_CMD_CREDITREQUEST:
		case ECBT_CMD_EXIT:
		case ECBT_CMD_ERROR:
			result = RecoverCommand(ecbtInfo, &escapeInfo);
			break;

		case ECBT_CMD_NONE:
			result = RecoverData(ecbtInfo, &escapeInfo);

		case ECBT_CMD_INIT:
		case ECBT_CMD_OPENCHANNEL:
		case ECBT_CMD_CLOSECHANNEL:
		case ECBT_CMD_DEBIT:
		case ECBT_CMD_DEBITREQUEST:
		case ECBT_CMD_CONFIGSOCKET:
		case ECBT_CMD_GETSOCKETID:
		case ECBT_CMD_GETSERVICENAME:
		case ECBT_CMD_EPSONPACKING:
		default:
			/* ΥޥɤʤΥ
			   ޥɤȯԤAPIϺƳƤϤʤʤ
			   ǤƼ̵顼֤ */
			result = ECBT_ERR_UNCONF;
			break;
		}

		if (escapeInfo.length > 0) {
			free(escapeInfo.buffer);
		}
	}
	
	return result;
}

/** 顼ɤ򸵤ˡԤ
 *   - 顼ѥåȤ
 *   - ǽǤС顼
 * @param ecbtInfo	ECBT
 * @param code		Error Code
 * @return		Error Code
 *
 */
static int
Terminate(EcbtInfo *ecbtInfo, EcbtErrCode code)
{
	int result = code;
#if (defined DEBUG)
	if (result != ECBT_ERR_NOERR) {
		MsgDebug("ERROR:result=%d\n", result);
	}
#endif
	return result;
}


/** signalȯΥХåեξ֤¸
 * @param *ecbtInfo	ECBT
 * @param operation	μ
 * @param buffer	writeξˤϡ񤭹٤ĤΥǡ
 * 			readξˤϡɤ߹ߺѤߤΥǡǼ롣
 * @param length	writeξˤϡ񤭹٤ĤΥǡ
 * 			readξϡɤ߹ߺѤΥǡǼ롣
 * @return		ǤECBT_ERR_SIGNAL򡢤ǤʤФʳΥ
 *          		顼֤
 *
 * @note ʥб\n
 *     read/write˥ʥǳޤ줿ɬ
 *     פ롣ǤСʣΥϥɥ뤬Ʊ˺Ƥ
 *     ꤷơƤΥϥɥˤƥե饰򤿤Ƥ٤\n
 *     \n
 *     readξϡΤΤ򥭥󥻥뤹롣ɤʬϼ
 *     ƤƤޤĤɤ߼ΤƤ롣\n
 *     \n
 *     writeξϡǤơAPIƤӽФκݤ˽
 *     ³롣\n
 *     \n
 *     ǡ֤ͥ̿ϡƥꥫХԤޥɥ
 *     ξ硢ĤΥޥɤǥꥫХԤƤϹԤʤ
 *     б륳ޥɤϰʲ̤ꡣ\n
 *     \n
 *     ECBT_CMD_CREDIT\n
 *     ECBT_CMD_CREDITREQUEST\n
 *     ECBT_CMD_EXIT\n
 *     ECBT_CMD_ERROR
 */
static int
CreateEscapeBuffer(EcbtInfo *ecbtInfo, EscapeOperation operation, char *buffer, size_t length)
{
	int result = ECBT_ERR_SIGNAL;

	/* εϿ */
	if (length > 0) {
		EcbtEscapeInfo *escapeInfo = NULL;
		
		/* ǰΰ٤clearƤ뤫ǧ */
		if (ecbtInfo->escapeInfo != NULL) {
			DeleteEscapeBuffer(ecbtInfo);
		}
		
		escapeInfo = (EcbtEscapeInfo *)malloc(sizeof(EcbtEscapeInfo));
		if (escapeInfo == NULL) {
			result = ECBT_ERR_NOMEM;
		} else {
			
			escapeInfo->buffer = (char *)malloc(sizeof(char *) * length);
			if (escapeInfo->buffer == NULL) {
				result = ECBT_ERR_NOMEM;
			} else {
				memcpy(escapeInfo->buffer, buffer, length);
				escapeInfo->length = length;
				escapeInfo->operation = operation;
			}
		}
		ecbtInfo->escapeInfo = escapeInfo;
	}
	
	return result;
}

/** signalȯΥޥɽ֤¸
 * @param ecbtInfo	ECBT
 * @param command	command ID
 * @param chId		channel ID
 * @return		Error Code
 */
static int
SetEscapeOperation(EcbtInfo *ecbtInfo, EcbtCommandId command, ChID chId)
{
	EcbtEscapeInfo *escapeInfo = ecbtInfo->escapeInfo;

	/* ꥫХꤹ٤ǡ뤫? */
	if (ecbtInfo->escapeInfo != NULL) {
		escapeInfo->chId = chId;
		escapeInfo->command = command;
	}

	return ECBT_ERR_SIGNAL;
}

/** ǡ򥯥ꥢ
 * @param ecbtInfo	ECBT
 * @return		Error Code
 */
static int
DeleteEscapeBuffer(EcbtInfo *ecbtInfo)
{
	int result = ECBT_ERR_NOERR;
	EcbtEscapeInfo *escapeInfo = ecbtInfo->escapeInfo;

	if (escapeInfo != NULL) {
		if (escapeInfo->buffer != NULL) {
			free(escapeInfo->buffer);
		}
		free(escapeInfo);
	}

	ecbtInfo->escapeInfo = NULL;
	return result;
}


/** ºݤ˥ݡȤФƥѥåȤɹԤ
 * @param ecbtInfo	ECBT
 * @param buffer	ɤ߹ѥåȤǼХåեؤΥݥ
 * @param offset	ɤ߹ޤƤǡ
 * @param rcount	ºݤɤ߹ǡǼݥ
 * @return		Error Code
 */
static int
RealPacketRead(EcbtInfo *ecbtInfo, char *buffer, size_t offset, size_t *rcount)
{
	char packetBuffer[ECBT_REPLY_MAX];
	int packetSize; /* size of packet */
	int dataSize;
	ChInfo *packetChInfo; /* ѥåȤΥͥ */
	int readCount;

	MsgDebug("%s(%d) %s: offset = %d\n",
			   __FILE__, __LINE__, __FUNCTION__, offset);

	/* sizeǧ */
	if (buffer != NULL && offset > 0) {
		memcpy(packetBuffer, buffer, offset);
	}

	/* signalǧ */
	if (signalCatchedFlag != 0) {
		/* signalȯ */
		*rcount = offset;
		return CreateEscapeBuffer(ecbtInfo, ESCAPE_PACKETREAD, packetBuffer, *rcount);
	}

	/* إåޤɤ߹ޤƤʤϡ˥إåɤ߹ */
	if (offset >= 0 && offset < CBT_HEADER_SIZE) {
		readCount = ReadLoop(ecbtInfo->fd, packetBuffer + offset, CBT_HEADER_SIZE - offset);
		if (readCount < 0) {
			readCount = 0;
			/* 顼³顼 */
			return ECBT_ERR_NOCONNECT;
		}
		*rcount = offset + readCount;
		offset = CBT_HEADER_SIZE;

		if (readCount < (CBT_HEADER_SIZE - offset)) {
			/* signalȯ */
			return CreateEscapeBuffer(ecbtInfo, ESCAPE_PACKETREAD, packetBuffer, *rcount);
		}
	}
	
	/* ѥåȥ */
	packetSize = CBT_HEADER_LENGTH(packetBuffer);
	packetChInfo = FindChInfo(ecbtInfo, CBT_HEADER_PSID(packetBuffer));
	dataSize = packetSize - offset;

	/* إåγǧ */
	if (packetChInfo == NULL || packetSize > packetChInfo->reverseSize) {
		return ECBT_ERR_UNCONF;
	}

	/* ѥåȥǡɤ߹ */
	if (packetSize > offset) {
		readCount = ReadLoop(ecbtInfo->fd, packetBuffer + offset, dataSize);
		if (readCount < 0) {
			readCount = 0;
			/* 顼³顼 */
			return ECBT_ERR_NOCONNECT;
		}

		*rcount = offset + readCount;

		if (readCount < dataSize) {
			/* signalȯ */
			return CreateEscapeBuffer(ecbtInfo, ESCAPE_PACKETREAD, packetBuffer, *rcount);
		}
		
		/* 顼ѥåȤγǧ */
		if (CheckErrorPacket(packetBuffer) != ECBT_ERR_NOERR){
			return ECBT_ERR_UNCONF;
		}
		
		/* 쥸åȤ */
		if (CBT_HEADER_PSID(packetBuffer) != CHID_TRANSACTION) {
			packetChInfo->hostCredit --;
		}
		
		/* ǡΥå */
		if (RingWrite(packetChInfo->reply, packetBuffer, packetSize) != packetSize) {
			return ECBT_ERR_FATAL;
		}
	}

	return ECBT_ERR_NOERR;
}



/** ƤΥǡɤफ顼ȯޤǤloopread
 * @param fd	եǥץ
 * @param buf	ɤ߹ǡǼХåեؤΥݥ
 * @param count	ɤ߹ǡΥ
 * @return	ɤ߹ǡΥ򡢥顼ʤ֤
 */
static int
ReadLoop(int fd, void *buf, size_t count)
{
	int result = 0;

	while (count > 0) {
		ssize_t rc;
		void* sigfunc;
		int en;

		/* ݡȤΥ֥åɤˡॢȤꤹ */
		sigfunc = signal(SIGALRM, AlarmHandler);
		alarm(ECBT_IO_TIMEOUT);

		rc = read(fd, buf, count);
		en = errno;

		alarm(0);
		signal(SIGALRM, sigfunc);
		/* signalȯ? */
		if (signalCatchedFlag != 0) {
			if (rc >= 0) {
				result += rc;
			}
			break;
		}

		if (rc < 0) {
#if (defined DEBUG)
			MsgDebug("read error: errno=%d\n", en);
#endif
			if ((en != EAGAIN) /*&& (en != EIO)*/) {
				result = -1;
				break;
			}
		} else if (rc == 0) {
			SafeUsleep(ECBT_IO_WAITTIME);
		} else {
			count -= rc;
			buf += rc;
			result += rc;
		}
	}

	return result;
}


/** ƤΥǡ񤯤顼ȯޤǤloopwrite
 * @param fd	եǥץ
 * @param buf	񤭹ǡǼХåեؤΥݥ
 * @param count	񤭹ǡΥ
 * @return	ʤ񤭹ǡΥ򡢥顼ʤ֤
 */
static int
WriteLoop(int fd, void *buf, size_t count)
{
	int result = 0;

	while (count > 0) {
		ssize_t wc;
		void* sigfunc;
		int en;

		/* ݡȤΥ֥åɤˡॢȤꤹ */
		sigfunc = signal(SIGALRM, AlarmHandler);
		alarm(ECBT_IO_TIMEOUT);

		wc = write(fd, buf, count);
		en = errno;

		alarm(0);
		signal(SIGALRM, sigfunc);
		/* signalȯ? */
		if (signalCatchedFlag != 0) {
			if (wc >= 0) {
				result += wc;
			}
			break;
		}

		if (wc < 0) {
#if (defined DEBUG)
			MsgDebug("write error: errno=%d\n", en);
#endif
			if ((en != EAGAIN) /*&& (en != EIO)*/) {
				result = -1;
				break;
			}
		} else if (wc == 0) {
			SafeUsleep(ECBT_IO_WAITTIME);
		} else {
			count -= wc;
			buf += wc;
			result += wc;
		}
	}

	return result;
}
