/* bidiEpsonStatus.c -- for Printer Status
 * Copyright (C) 2003 EPSON KOWA Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: bidiEpsonStatus.c,v 1.24 2004/02/20 17:34:27 ishii Exp $ */

/**
 * @file  bidiEpsonStatus.c
 * @brief EPSON 󥯥åȥץ󥿤Υơؿ
 *			¸ʤ褦˽񤤤PX-V700PM-870C 2ǤƥȤƤʤ
 * @author toko (EPSON KOWA Co.) <toko@epkowa.co.jp>
 * @date 2004-01-16
 * @version $Id: bidiEpsonStatus.c,v 1.24 2004/02/20 17:34:27 ishii Exp $
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <locale.h>
#include <langinfo.h>
#include <libintl.h>
#include <iconv.h>
#include <errno.h>

#if defined(DEBUG)
#include <stdarg.h>
#endif /* DEBUG */

#include "bidiEpson.h"
#include "bidiEpsonStatus.h"


#define _(String)	gettext(String)				/**< gettext  */
//#define N_(String)  gettext_noop(String)
#define N_(String)	(String)					/**< gettext  */

#define PACKAGE "bidiEpsonStatus"				/**< gettext  */
#define VERSION "1.0.0"							/**< gettext  */


/* XML Ϣʸ */
/** XML_xxxx  XML Ǥ */
#define XML_HEADER	"<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"

#define XML_PMIBS	"<Printer>\n"
#define XML_PMIBE	"</Printer>\n"

#define XML_MIBS	"<Printer-MIB>\n"
#define XML_MIBE	"</Printer-MIB>\n"



#define STATUS_START	"@BDC ST\r\n"			/**< ץ󥿤Υơγʸ */



#define INK_COLOR_MAX		7					/**< 7ޤб */
#define MICROWEAVE_ID_MAX	3					/**< microweave ID θĿ */
#define IG_INFO_MAX			10					/**< ֤οʬɬ */





/* ---------------------------------------------------------------- */
/** "ST" ѥơ֥ */
struct StConstTable_t {
	int		mib;								/**< MIB  */
	int		status;								/**< ץ󥿤֤äƤ륹ơ (ֹ) */
	char	*str;								/**< ơɤɽ (ʸ) */
} const StConstTable[] = {
	{ 1, STATUS_ERROR, N_("Error") },
	{ 4, STATUS_PRINTING, N_("Printing") },
	{ 1, STATUS_BUSY, N_("Busy") },
//	{ 1, STATUS_WAIT, N_("Wait") },				/* ͽǤεҤϤʤäƤ */
	{ 1, STATUS_WAIT, N_("Printing") },
	{ 3, STATUS_IDLE, N_("Idle") },
	{ 1, STATUS_IMPOSSIBLE, N_("Print impossible") },
	{ 1, STATUS_INK_DRY, N_("Ink under dryness") },
	{ 1, STATUS_CLEANING, N_("Cleaning") },
	{ 1, STATUS_FACTORY_DEFAULT, N_("Factory shipment state") },
	{ 1, STATUS_MOTOR_OFF, N_("Motor OFF") },
	{ 1, STATUS_SHUTTINGDOWN, N_("Shutting down") },
	{ 1, STATUS_WAIT_INIT, N_("Wait paper initialize") },
	{ 1, STATUS_INITIALIZING, N_("Paper initializing") },
};

/** StConstTable ΤɤǤͭݤɽ */
struct StTable_t {
	bool	valid;								/**< TRUE: ͭFALSE:̵ */
};

/** "ER" ѥơ֥ */
/* TBD #### RFC §ä codegrouplevel 򵭽Ҥ */
struct ErConstTable_t {
	int		code;								/**< AlertCode */
	int		group;								/**< AlertGroup */
	int		level;								/**< SeverityLevel */
	char	*key;								/**< ץ󥿤֤äƤ륳 */
	char	*str;								/**< ɤɽ (ʸ) */
} ErConstTable[] = {
	{ 1, 1, 3, "00", N_("General error") },
	{ 1, 1, 3, "01", N_("Printing data from another port") },
	{ 3, 6, 3, "02", N_("Printer cover open") },
	{ 1, 1, 3, "03", N_("Incorrect release lever position") },
	{ 8, 8, 3, "04", N_("Paper jam") },
	{ 1102, 1, 3, "05", N_("Ink out") },
	{ 1, 1, 3, "06", N_("Paper not loaded correctly") },
	{ 1, 1, 3, "07", N_("Initialize") },
	{ 1, 1, 3, "08", N_("Unknown") },
	{ 1, 1, 3, "09", N_("Paper changing") },
	{ 1, 1, 3, "0A", N_("Paper size (SIDM)") },
	{ 1, 1, 3, "0B", N_("Ribbon jam") },
	{ 1, 1, 3, "0C", N_("Paper out or incorrect output tray position") },
	{ 1, 1, 3, "0D", N_("Incorrect lever position") },
	{ 1, 1, 3, "0E", N_("Paper unload error") },
	{ 1, 1, 3, "0F", N_("SIMM copy error") },
	{ 1, 1, 3, "10", N_("Ink over flow") },
	{ 1, 1, 3, "11", N_("Wait tear off") },
	{ 1, 1, 3, "12", N_("Trans multiple papers") },
	{ 1, 1, 3, "13", N_("Head hot") },
	{ 1, 1, 3, "14", N_("Paper cut error") },
	{ 1, 1, 3, "15", N_("Release lever error") },
	{ 1, 1, 3, "16", N_("Can't cleaning") },
	{ 1, 1, 3, "17", N_("Paper error") },
	{ 1, 1, 3, "18", N_("Paper slant") },
	{ 1, 1, 3, "19", N_("Exceed cleaning times") },
	{ 1, 1, 3, "1A", N_("Ink cover open") },
	{ 1, 1, 3, "1B", N_("Ink cartridge") },
	{ 1, 1, 3, "1C", N_("Cutter error") },
	{ 1, 1, 3, "1D", N_("Cutter jam") },
	{ 1, 1, 3, "1E", N_("Ink color") },
	{ 1, 1, 3, "1F", N_("Cutter cover open") },
	{ 1, 1, 3, "20", N_("Ink lever released") },
	{ 1, 1, 3, "22", N_("No maintenance tank") },
	{ 1, 1, 3, "23", N_("Ink cartridge combination") },
	{ 1, 1, 3, "24", N_("Command error") },
	{ 1, 1, 3, "25", N_("Rear cover open") },
	{ 1, 1, 3, "26", N_("Multi censor gain") },
	{ 1, 1, 3, "27", N_("Can't auto adjust") },
	{ 1, 1, 3, "28", N_("Cleaning fail") },
	{ 1, 1, 3, "29", N_("No paper tray") },
};

/** ErConstTable ΤɤǤͭݤɽ */
struct ErTable_t {
	bool	valid;								/**< TRUE: ͭFALSE:̵ */
};

/** "WR" ѥơ֥ */
/* TBD #### RFC §ä codegrouplevel 򵭽Ҥ */
struct WrConstTable_t {
	int		code;								/**< AlertCode */
	int		group;								/**< AlertGroup */
	int		level;								/**< SeverityLevel */
	char	*key;								/**< ץ󥿤֤äƤ륳 */
	char	*str;								/**< ɤɽ (ʸ) */
} WrConstTable[] = {
	{ 1, 11, 4, "00", N_("Near end bin 1") },
	{ 1, 11, 4, "01", N_("Near end bin 2") },
	{ 1, 11, 4, "10", N_("Ink low Black") },
	{ 1, 11, 4, "11", N_("Ink low Cyan") },
	{ 1, 11, 4, "12", N_("Ink low Magenta") },
	{ 1, 11, 4, "13", N_("Ink low Yellow") },
	{ 1, 11, 4, "14", N_("Ink low Light Cyan") },
	{ 1, 11, 4, "15", N_("Ink low Light Magenta") },
	{ 1, 11, 4, "16", N_("Ink low Dark Yellow") },
	{ 1, 11, 4, "17", N_("Ink low Black 2") },
	{ 1, 11, 4, "18", N_("Ink low Black 3") },
	{ 1, 11, 4, "19", N_("Ink low Black 4") },
	{ 1, 11, 4, "20", N_("Maintenance request will be occurred as soon") },
	{ 1, 11, 4, "21", N_("Maintenance request") },
	{ 1, 11, 4, "30", N_("Main tank ink end Black") },
	{ 1, 11, 4, "31", N_("Main tank ink end Cyan") },
	{ 1, 11, 4, "32", N_("Main tank ink end Magenta") },
	{ 1, 11, 4, "33", N_("Main tank ink end Yellow") },
	{ 1, 11, 4, "34", N_("Main tank ink end Light Cyan") },
	{ 1, 11, 4, "35", N_("Main tank ink end Light Magenta") },
	{ 1, 11, 4, "36", N_("Main tank ink end Dark Yellow") },
	{ 1, 11, 4, "37", N_("Main tank ink end Black 2") },
	{ 1, 11, 4, "38", N_("Main tank ink end Black 3") },
	{ 1, 11, 4, "39", N_("Main tank ink end Black 4") },
	{ 1, 11, 4, "40", N_("Ink cover open") },
};

/** WrConstTable ΤɤǤͭݤɽ */
struct WrTable_t {
	bool	valid;								/**< TRUE: ͭFALSE:̵ */
};

/** "IQ" ѥơ֥ */
struct IqConstTable_t {
	char	*color;								/**< ̾ */
} IqConstTable[] = {
	{ N_("Black") },
	{ N_("Cyan") },
	{ N_("Magenta") },
	{ N_("Yellow") },
	{ N_("Light Cyan") },
	{ N_("Light Magenta") },
	{ N_("Dark Yellow") },
};

/** "IQ" ˥ץ󥿤֤ä뤫⤷ʤ warning ʸѥơ֥ */
const char *IqWrList[] = {
	"NA",
	"RE",
	"WE",
	"CI",
};

/** ץ󥿤֤äƤ "IQ" ǡǼ뤿Υơ֥ */
struct IqTable_t {
	bool	valid;								/**< TRUE: ͭFALSE:̵ */
	bool	wearing;							/**<  */
	int		quantity;							/**<  */
};

/** "AI:CW:<nn><KK><CC><MM><YY><cc><mm><dydy><kk>" 顼 ID (nn ʬ) ѥơ֥ */
struct AiCwNumberConstTable_t {
	int		number;							/**< 󥯤ο */
	char	*nn;							/**< 󥯤ο */
	char	*str;							/**< ɤɽ󥯤ο (ʸ) */
} AiCwNumberConstTable[] = {
	{ 3, "01", N_("3 color") },
	{ 4, "02", N_("4 color") },
	{ 6, "03", N_("6 color") },
	{ 7, "04", N_("7 color") },
	{ 6, "05", N_("6 color") },
};

/** AiCwNumberConstTable ΤɤǤͭݤɽ */
struct AiCwNumberTable_t {
	bool	valid;								/**< TRUE: ͭFALSE:̵ */
};

/** "AI:CW:<nn><KK><CC><MM><YY><cc><mm><dydy><kk>" 󥯽 ID (nn μʹ) ѥơ֥ */
struct AiCwWeightTable_t {
	bool	valid;								/**< TRUE: ͭFALSE:̵ */
	int		weight;								/**< 󥯽 ID */
} AiCwWeightTable[] = {
	{ FALSE, 0 },
	{ FALSE, 0 },
	{ FALSE, 0 },
	{ FALSE, 0 },
	{ FALSE, 0 },
	{ FALSE, 0 },
	{ FALSE, 0 },
	{ FALSE, 0 },
};

/** "AI:CW:<nn><KK><CC><MM><YY><cc><mm><dydy><kk>,MI:<mid>" microweave ID (mid ʬ) ѥơ֥ */
struct AiMiTable_t {
	bool	valid;								/**< TRUE: ͭFALSE:̵ */
	char	id[MICROWEAVE_ID_MAX];				/**< ץ󥿤֤äƤ microweave ID */
};


/** "TC" ѥơ֥ */
struct TcTable_t {
	bool	valid;								/**< TRUE: ͭFALSE:̵ */
	int		time;								/**< ꡼˥󥰻 ޤ 󥯽Ŷ */
};


/** "INK:<m1><m2><m3>[,<m1><m2><m3>...]"  <m1> νѥơ֥ */
struct InkM1ConstTable_t {
	char	key;								/**< ץ󥿤֤äƤ륳 */
	char	*desc;								/**< ɤɽ̣ (ʸ) */
} InkM1ConstTable[] = {
	{ '0', "Unknown" },
	{ '1', "Black Ink Cartridge" },
	{ '2', "Color Ink Cartridge" },
	{ '3', "Cyan Ink Cartridge"},
	{ '4', "Magenta Ink Cartridge"},
	{ '5', "Yellow Ink Cartridge"},
	{ '6', "Light Cyan Ink Cartridge"},
	{ '7', "Light Magenta Ink Cartridge"},
	{ '8', "Light Yellow Ink Cartridge"},
	{ '9', "Dark Yellow Ink Cartridge"},
	{ 'A', "Extra 0A Ink Cartridge"},
	{ 'B', "Extra 0B Ink Cartridge"},
};

/** "INK:<m1><m2><m3>[,<m1><m2><m3>...]"  <m2> νѥơ֥ */
struct InkM2ConstTable_t {
	char	key;								/**< ץ󥿤֤äƤ륳 */
	char	*desc;								/**< ɤɽ̣ (ʸ) */
} InkM2ConstTable[] = {
	{ '0', N_("Unknown") },
	{ '1', N_("Black") },
	{ '2', N_("Cyan (including Light Cyan)") },
	{ '3', N_("Magenta (including Light Magenta)") },
	{ '4', N_("Yellow (including Light Yellow and Dark Yellow)") },
};

/** "INK:<m1><m2><m3>[,<m1><m2><m3>...]"  <m3> νѥơ֥ */
struct InkM3ConstTable_t {
	int		key;								/**< ץ󥿤֤äƤ륳 */
	int		index;								/**< "IqConstTable" ơ֥Υǥå */
	char	*desc;								/**< ɤɽ̣ (ʸ) */
} InkM3ConstTable[] = {
	{ 0x01, 0, N_("1st data of IQ (Black)") },
	{ 0x02, 1, N_("2nd data of IQ (Cyan)") },
	{ 0x04, 2, N_("3rd data of IQ (Magenta)") },
	{ 0x08, 3, N_("4th data of IQ (Yellow)") },
	{ 0x10, 4, N_("5th data of IQ (Light Cyan)") },
	{ 0x20, 5, N_("6th data of IQ (Light Magenta)") },
	{ 0x40, 6, N_("7th data of IQ") },
};

/** ץ󥿤֤äƤ "INK" ǡǼ뤿Υơ֥ */
struct InkTable_t {
	bool	valid;								/**< TRUE: ͭFALSE:̵ */
	int		indexM1;							/**< InkM1ConstTable ơ֥Ǥ index  */
	int		indexM2;							/**< InkM2ConstTable ơ֥Ǥ index  */
	int		indexM3;							/**< InkM3ConstTable ơ֥Ǥ index  */
};

/** "IG" () ѥơ֥ */
struct IgNn1ConstTable_t {
	char	*key;								/**< ץ󥿤֤äƤ륳 */
	char	*str;								/**< ɤɽ̣ (ʸ) */
} IgNn1ConstTable[] = {
	{ "00", N_("Unknown") },
	{ "01", N_("Auto sheet feeder") },
	{ "02", N_("Unchangeable Sheet feeder bin 1") },
	{ "03", N_("Sheet feeder bin 1") },
	{ "04", N_("Unchangeable Sheet feeder bin 2") },
	{ "05", N_("Sheet feeder bin 2") },
	{ "06", N_("Unchangeable Sheet feeder bin 3") },
	{ "07", N_("Sheet feeder bin 3") },
	{ "08", N_("Manual feed front") },
	{ "09", N_("Manual feed rear") },
	{ "0A", N_("Manual") },
	{ "0B", N_("Tractor front") },
	{ "0C", N_("Tractor rear") },
	{ "0D", N_("Tractor") },
	{ "0E", N_("Roll") },
};

/** ץ󥿤֤äƤ "IG" ǡǼ뤿Υơ֥ */
struct IgTable_t {
	bool	valid;								/**< TRUE: ͭFALSE:̵ */
	int		indexNn1;							/**< IgNn1ConstTable ơ֥Ǥ index */
	int		capacity;							/**< ֤ */
	int		quantity;							/**< ̡-1  NA Ǥ */
};

/** "CA" ѥơ֥ */
struct CaConstTable_t {
	char	*key;								/**< ץ󥿤֤äƤ륳 */
	char	*str;								/**< ɤɽ̣ (ʸ) */
} CaConstTable[] = {
	{ "01", N_("No request") },
	{ "A1", N_("Initializing") },
	{ "81", N_("Exist request") },
};

/** CaConstTable ΤɤǤͭݤɽ */
struct CaTable_t {
	bool	valid;								/**< TRUE: ͭFALSE:̵ */
};


/** 嵭 ¤Τδѥơ֥ */
typedef struct StatusList_t {
	struct StTable_t	StTable[sizeof(StConstTable) / sizeof(struct StConstTable_t)];
	struct ErTable_t	ErTable[sizeof(ErConstTable) / sizeof(struct ErConstTable_t)];
	struct WrTable_t	WrTable[sizeof(WrConstTable) / sizeof(struct WrConstTable_t)];
	struct IqTable_t	IqTable[sizeof(IqConstTable) / sizeof(struct IqConstTable_t)];
	struct AiCwNumberTable_t	AiCwNumberTable[sizeof(AiCwNumberConstTable) / sizeof(struct AiCwNumberConstTable_t)];
	struct AiCwWeightTable_t	AiCwWeightTable[INK_COLOR_MAX];
	struct AiMiTable_t	AiMiTable;
	struct TcTable_t	TcTable;
	struct InkTable_t	InkTable[INK_COLOR_MAX];
	struct IgTable_t	IgTable[IG_INFO_MAX];
	struct CaTable_t	CaTable[sizeof(CaConstTable) / sizeof(struct CaConstTable_t)];
} StatusList;




/** ɤ߹ơνѥơ֥ */
typedef struct StatusTable_t {
	char	*keyStr;							/**< ơʸ */
	void	(*init)(StatusList *sl);			/**< ؿ */
	int		(*set)(StatusList *sl, char *, int, struct StatusTable_t *);	/**< ơʸåؿ
																			 * ˤäƤϻѤʤ⤢ */
} StatusTable;


/* ---------------------------------------------------------------- */
/** XML ѥơ֥ */
typedef struct MakeXmlTable_t {
	char	*name;										/**< TAG ̾ */
	char	*text;										/**< TAG ˰Ϥޤʸ */
	char	*(*make)(StatusList *sl, int, int, int);	/**< ߤ TAG νؿ
														 * ˤäƤϻѤʤ⤢ */

	struct MakeXmlTable_t	*child;						/**<  ؿ */
	struct MakeXmlTable_t	*next;						/**<  ؿ */
} MakeXmlTable;

/** '<' '>' ʸ뤿 */
typedef enum makeXmlContentFlag_t {
	MAKE_XML_CONTENT_BRACE_NON,					/**< '<' '>' ղäʤ */
	MAKE_XML_CONTENT_BRACE_START,				/**< '<' '>' ղä */
	MAKE_XML_CONTENT_BRACE_END					/**< "</" ">\n" ղä */
} makeXmlContentFlag;


/** ʸ뤫ݤΥե饰 */
#define MAKE_XML_CONTENT_RELEASE_NON	0x00	/**< ʤ */
#define MAKE_XML_CONTENT_RELEASE_1		0x01	/**< ܤΰ */
#define MAKE_XML_CONTENT_RELEASE_2		0x02	/**< ܤΰ */





/** DebugѤ˥åϤ
 * @param printf ؿƱΰȤ
 * @return ̵
 */
static void DebugPrint(const char *fmt, ...)
{
#ifdef DEBUG
	va_list ap;

	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	fflush(stderr);
#endif
	return;
}

/* ---------------------------------------------------------------- */
/* ʹߤ ֤ǤѤؿ */

/** ܸ쥳ɤUTF-8 Ѵ
 * @param *fromStr ѴʸǼƤХåեؤΥݥ
 * @retval  ѴʸǼХåեؤΥݥ
 * @retval Ի NULL
 */
static char *Utf8Conv(char *fromStr)
{
	char	*code;
	iconv_t	cd;
	size_t	nconv;
	char	*toStr;
	char	*toPtr;
	int		fromLen, toLen;
	int		result;

    code = nl_langinfo(CODESET);
	cd = iconv_open("UTF-8", code); /* EUC-JP -> UTF-8 (may be) */
	if (cd == (iconv_t)-1) {
		DebugPrint("%s(%d) %s: code \"%s\", %d,%s\n", __FILE__, __LINE__, __FUNCTION__, code, errno, strerror(errno));
		return(NULL);
	}

	fromLen = strlen(fromStr);
	toLen = fromLen * 6;			/* Τ餤פʡ */
	toStr = malloc(toLen + 1);
	if (toStr == NULL) {
		DebugPrint("%s(%d) %s: Not enough core\n", __FILE__, __LINE__, __FUNCTION__);
		iconv_close(cd);
		return(NULL);
	}
	memset(toStr, 0, toLen);
	toPtr = toStr;
	nconv = iconv(cd, &fromStr, &fromLen, &toPtr, &toLen);
	if (nconv == -1) {
		DebugPrint("%s(%d) %s: %d,%s\n", __FILE__, __LINE__, __FUNCTION__, errno, strerror(errno));
		iconv_close(cd);
		free(toStr);
		return(NULL);
	}

	result = iconv_close(cd);
	if (result == -1) {
		DebugPrint("%s(%d) %s: %d,%s\n", __FILE__, __LINE__, __FUNCTION__, errno, strerror(errno));
		free(toStr);
		return(NULL);
	}
	return(toStr);
}


/* ---------------------------------------------------------------- */

/** 10ʸѴ
 * @param *str 10ʸΥݥ
 * @param length ѴʸθĿ
 * @param *result Ѵ̤ǼХåեؤΥݥ
 * @retval TRUE 
 * @retval FALSE 
 */
bool str2Int(char *str, int length, int *result)
{
	int	i;

	*result = 0;
	for (i = 0; i < length; i++) {
		if ((str[i] >= '0') && (str[i] <= '9')) {
			*result = *result * 10 + str[i] - '0';
		} else {
			return(FALSE);
		}
	}
	return(TRUE);
}

/** 16ʸѴ
 * @param *str 16ʸؤΥݥ
 * @param length ѴʸθĿ
 * @param *result Ѵ̤ǼХåեؤΥݥ
 * @retval TRUE 
 * @retval FALSE 
 */
bool strX2Int(char *str, int length, int *result)
{
	int	i;

	*result = 0;
	for (i = 0; i < length; i++) {
		if ((str[i] >= '0') && (str[i] <= '9')) {
			*result = *result * 16 + str[i] - '0';
		} else if ((str[i] >= 'A') && (str[i] <= 'F')) {
			*result = *result * 16 + str[i] - 'A' + 10;
		} else if ((str[i] >= 'a') && (str[i] <= 'f')) {
			*result = *result * 16 + str[i] - 'a' + 10;
		} else {
			return(FALSE);
		}
	}
	return(TRUE);
}


/* ----------------------------------------------------------------
 * for ST
 */

/** "ST" ѥơ֥ν
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @return ʤ
 */
static void initST(StatusList *sl)
{
	int		i;
	struct StTable_t	*StTable = sl->StTable;

	for (i = 0; i < (sizeof(StConstTable) / sizeof(struct StConstTable_t)); i++) {
		StTable[i].valid = FALSE;
	}
}

/** ɤ߹ơǡ "ST" ѥơ֥ؤ
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @param *status ơǡʸؤΥݥ
 * @param strLength ̤
 * @param *tbl ơǡѥơ֥ؤΥݥ
 * @retval TRUE 
 * @retval FALSE 
 */
static bool setST(StatusList *sl, char *status, int strLength, StatusTable *tbl)
{
	char	*statusP;
	int		result;
	int		value;
	int		statusL;
	struct StTable_t	*StTable = sl->StTable;

	statusP = status + strlen(tbl->keyStr);

	statusL = 3;
	if (statusP[statusL - 1] != STATUS_RS) {
		return(FALSE);
	}
	result = strX2Int(statusP, 2, &value);
	if (result == FALSE) {
		return(FALSE);
	}
	if ((sizeof(StConstTable) / sizeof(struct StConstTable_t)) <= value) {
		return(FALSE);
	}

	StTable[value].valid = TRUE;

	return(TRUE);
}


/* ----------------------------------------------------------------
 * for ER
 */

/** "ER" ѥơ֥ν
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @return ʤ
 */
static void initER(StatusList *sl)
{
	int		i;
	struct ErTable_t	*ErTable = sl->ErTable;

	for (i = 0; i < (sizeof(ErConstTable) / sizeof(struct ErConstTable_t)); i++) {
		ErTable[i].valid = FALSE;
	}
}

/** ɤ߹ơǡ "ER" ѥơ֥ؤ
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @param *status ơǡʸؤΥݥ
 * @param strLength ̤
 * @param *tbl ơǡѥơ֥ؤΥݥ
 * @retval TRUE 
 * @retval FALSE 
 */
static bool setER(StatusList *sl, char *status, int strLength, StatusTable *tbl)
{
	int		i;
	char	*statusP;
	struct ErTable_t	*ErTable = sl->ErTable;

	statusP = status + strlen(tbl->keyStr);

	while (*statusP != STATUS_RS) {
		for (i = 0; i < (sizeof(ErConstTable) / sizeof(struct ErConstTable_t)); i++) {
			if (strncmp(statusP, ErConstTable[i].key, strlen(ErConstTable[i].key)) == 0) {
				ErTable[i].valid = TRUE;
				break;
			}
		}

		statusP += strlen(ErConstTable[i].key);
		if (*statusP == STATUS_FS) {
			statusP++;
		}
	}

	return(TRUE);
}


/* ----------------------------------------------------------------
 * for WR
 */

/** "WR" ѥơ֥ν
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @return ʤ
 */
static void initWR(StatusList *sl)
{
	int		i;
	struct WrTable_t	*WrTable = sl->WrTable;

	for (i = 0; i < (sizeof(WrConstTable) / sizeof(struct WrConstTable_t)); i++) {
		WrTable[i].valid = FALSE;
	}
}

/** ɤ߹ơǡ "WR" ѥơ֥ؤ
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @param *status ơǡʸؤΥݥ
 * @param strLength ̤
 * @param *tbl ơǡѥơ֥ؤΥݥ
 * @retval TRUE 
 * @retval FALSE 
 */
static bool setWR(StatusList *sl, char *status, int strLength, StatusTable *tbl)
{
	int		i;
	char	*statusP;
	struct WrTable_t	*WrTable = sl->WrTable;

	statusP = status + strlen(tbl->keyStr);

	while (*statusP != STATUS_RS) {
		for (i = 0; i < (sizeof(WrConstTable) / sizeof(struct WrConstTable_t)); i++) {
			if (strncmp(statusP, WrConstTable[i].key, strlen(WrConstTable[i].key)) == 0) {
				WrTable[i].valid = TRUE;
				break;
			}
		}

		statusP += strlen(WrConstTable[i].key);
		if (*statusP == STATUS_FS) {
			statusP++;
		}
	}

	return(TRUE);
}


/* ----------------------------------------------------------------
 * for IQ
 */

/** "IQ" ѥơ֥ν
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @return ʤ
 */
static void initIQ(StatusList *sl)
{
	int		i;
	struct IqTable_t	*IqTable = sl->IqTable;

	for (i = 0; i < (sizeof(IqConstTable) / sizeof(struct IqConstTable_t)); i++) {
		IqTable[i].valid = FALSE;
	}
}

/** ɤ߹ơǡ "IQ" ѥơ֥ؤ
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @param *status ơǡʸؤΥݥ
 * @param strLength ̤
 * @param *tbl ơǡѥơ֥ؤΥݥ
 * @retval TRUE 
 * @retval FALSE 
 */
static bool setIQ(StatusList *sl, char *status, int strLength, StatusTable *tbl)
{
	char	*statusP;
	int		result;
	int		value;
	int		i, j;
	struct IqTable_t	*IqTable = sl->IqTable;

	statusP = status + strlen(tbl->keyStr);

	/* TBD ####
	 * ɡ
	 * ʸåˤäơ
	 * ϿƤʤäͤȤưŪˤ */
	for (i = 0; (i < (sizeof(IqConstTable) / sizeof(struct IqConstTable_t))) && (*statusP != STATUS_RS); i++) {
		value = 0;
		result = strX2Int(statusP, 2, &value);
		if (result == FALSE) {
			/* ͤǤϤʤʸ󤫤⤷ʤ */
			for (j = 0; j < (sizeof(IqWrList) / sizeof(char *)); j++) {
				if (strncmp(statusP, IqWrList[i], strlen(IqWrList[i])) == 0) {
				/* ޤɤʤäߤ */
					IqTable[i].wearing = FALSE;
					IqTable[i].quantity = 0;
					break;
				}
			}
			if (j == (sizeof(IqWrList) / sizeof(char *))) {
				return(FALSE);
			}
		} else {
			IqTable[i].wearing = TRUE;
			IqTable[i].quantity = value;
		}
		IqTable[i].valid = TRUE;
		statusP += 2;
	}

	return(TRUE);
}


/* ----------------------------------------------------------------
 * for AI
 */

/** "AI:CW" ѥơ֥ν
 * @param *sl "AI" ץ󥿥ơѹ¤ΤؤΥݥ
 * @return ʤ
 */
static void initAICW(StatusList *sl)
{
	int		i;
	struct AiCwNumberTable_t	*AiCwNumberTable = sl->AiCwNumberTable;
	struct AiCwWeightTable_t	*AiCwWeightTable = sl->AiCwWeightTable;

	for (i = 0; i < (sizeof(AiCwNumberConstTable) / sizeof(struct AiCwNumberConstTable_t)); i++) {
		AiCwNumberTable[i].valid = FALSE;
	}
	for (i = 0; i < INK_COLOR_MAX; i++) {
		AiCwWeightTable[i].valid = FALSE;
	}
}

/** ɤ߹ơǡ "AI:CW" ѥơ֥ؤ
 * @param *sl "AI" ץ󥿥ơѹ¤ΤؤΥݥ
 * @param *status ơǡʸؤΥݥ
 * @param strLength ̤
 * @param *tbl ơǡѥơ֥ؤΥݥ
 * @retval TRUE 
 * @retval FALSE 
 */
static bool setAICW(StatusList *sl, char *status, int strLength, StatusTable *tbl)
{
	int		i, j;
	char	*statusP;
	int		result;
	int		size;
	struct AiCwNumberTable_t	*AiCwNumberTable = sl->AiCwNumberTable;
	struct AiCwWeightTable_t	*AiCwWeightTable = sl->AiCwWeightTable;

	statusP = status + strlen(tbl->keyStr);

	size = sizeof(AiCwNumberConstTable) / sizeof(struct AiCwNumberConstTable_t);
	for (i = 0; i < size; i++) {
		if (strncmp(statusP, AiCwNumberConstTable[i].nn, strlen(AiCwNumberConstTable[i].nn)) == 0) {
			AiCwNumberTable[i].valid = TRUE;
			statusP += strlen(AiCwNumberConstTable[i].nn);
			break;
		}
	}
	if (i == size) {
		return(FALSE);
	}

	size = AiCwNumberConstTable[i].number;
	for (j = 0; j < size; j++) {
		result = strX2Int(statusP, 2, &AiCwWeightTable[j].weight);
		if (result == FALSE) {
			return(FALSE);
		}
		AiCwWeightTable[j].valid = TRUE;
		statusP += 2;
	}

	return(TRUE);
}


/** "AI:,,,MI" ѥơ֥ν
 * @param *sl "AI" ץ󥿥ơѹ¤ΤؤΥݥ
 * @return ʤ
 */
static void initAIMI(StatusList *sl)
{
	int		i;
	struct AiMiTable_t	*AiMiTable = &(sl->AiMiTable);

	AiMiTable->valid = FALSE;

	for (i = 0; i < (sizeof(AiMiTable->id) / sizeof(char)); i++) {
		AiMiTable->id[i] = '\0';
	}
}

/** ɤ߹ơǡ "AI:,,,MI" ѥơ֥ؤ
 * @param *sl "AI" ץ󥿥ơѹ¤ΤؤΥݥ
 * @param *status ơǡʸؤΥݥ
 * @param strLength ơǡʸĹ
 * @param *tbl ơǡѥơ֥ؤΥݥ
 * @retval TRUE 
 * @retval FALSE 
 */
static bool setAIMI(StatusList *sl, char *status, int strLength, StatusTable *tbl)
{
	int		i;
	char	*statusP;
	int		statusL;
	struct AiMiTable_t	*AiMiTable = &(sl->AiMiTable);

	statusP = status + strlen(tbl->keyStr);

	statusL = sizeof(AiMiTable->id) / sizeof(char);
	if (strLength < statusL) {
		return(FALSE);
	}
	if (statusP[statusL - 1] != STATUS_RS) {
		return(FALSE);
	}

	for (i = 0; i < (statusL - 1); i++) {
		AiMiTable->id[i] = statusP[i];
	}
	AiMiTable->valid = TRUE;

	return(TRUE);
}


/** "AI" Ϣѥơ֥ */
StatusTable statusAITbl[] = {
	{ "CW:", initAICW, setAICW },
	{ "MI:", initAIMI, setAIMI },
};

/** "AI" ѥơ֥ν
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @return ʤ
 */
static void initAI(StatusList *sl)
{
	int		i;

	for (i = 0; i < (sizeof(statusAITbl) / sizeof(struct StatusTable_t)); i++) {
		statusAITbl[i].init(sl);
	}
}

/** ɤ߹ơǡ "AI" ѴؿƤӽФ
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @param *status ơǡʸؤΥݥ
 * @param strLength ơǡʸĹ
 * @param *tbl ơǡѥơ֥ؤΥݥ
 * @retval TRUE 
 * @retval FALSE 
 */
static bool setAI(StatusList *sl, char *status, int strLength, StatusTable *tbl)
{
	int		i;
	int		result;
	int		statusL;

	statusL = strlen(tbl->keyStr);
	strLength -= statusL;
	status += statusL;

	result = FALSE;
	while ((strLength > 0) && (*status != STATUS_EOR) && (*status != STATUS_RS)) {
		for (i = 0; i < (sizeof(statusAITbl) / sizeof(StatusTable)); i++) {
			if (strncmp(status, statusAITbl[i].keyStr, strlen(statusAITbl[i].keyStr)) == 0) {
				result = statusAITbl[i].set(sl, status, strLength, &statusAITbl[i]);
				/* 顼̵뤷Ʋ³ */
/*
				if (result == FALSE) {
					return(result);
				}
*/
				/* ֤˰¸ʤᡢƬõ */
				break;
			}
		}
		while ((strLength > 0) &&
			   (*status != STATUS_EOR) &&
			   (*status != STATUS_RS) &&
			   (*status != STATUS_FS)) {
			strLength--;
			status++;
		}
		if (*status == STATUS_FS) {
			strLength--;
			status++;
		}
	}

	return(result);
}


/* ----------------------------------------------------------------
 * for TC
 */

/** "TC" ѥơ֥ν
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @return ʤ
 */
static void initTC(StatusList *sl)
{
	struct TcTable_t	*TcTable = &(sl->TcTable);

	TcTable->valid = FALSE;
}

/** ɤ߹ơǡ "TC" ѥơ֥ؤ
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @param *status ơǡʸؤΥݥ
 * @param strLength ̤
 * @param *tbl ơǡѥơ֥ؤΥݥ
 * @retval TRUE 
 * @retval FALSE 
 */
static bool setTC(StatusList *sl, char *status, int strLength, StatusTable *tbl)
{
	char	*statusP;
	int		result;
	int		value;
	struct TcTable_t	*TcTable = &(sl->TcTable);

	statusP = status + strlen(tbl->keyStr);

	result = strX2Int(statusP, 4, &value);
	if (result == FALSE) {
		return(FALSE);
	}
	TcTable->valid = TRUE;
	TcTable->time = value;

	return(TRUE);
}


/* ----------------------------------------------------------------
 * for INK
 */

/** "INK" ѥơ֥ν
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @return ʤ
 */
static void initINK(StatusList *sl)
{
	int		i;
	struct InkTable_t	*InkTable = sl->InkTable;

	for (i = 0; i < INK_COLOR_MAX; i++) {
		InkTable[i].valid = FALSE;
	}
}

/** ɤ߹ơǡ "INK" ѥơ֥ؤ
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @param *status ơǡʸؤΥݥ
 * @param strLength ̤
 * @param *tbl ơǡѥơ֥ؤΥݥ
 * @retval TRUE 
 * @retval FALSE 
 */
static bool setINK(StatusList *sl, char *status, int strLength, StatusTable *tbl)
{
	int		i, j;
	char	*statusP;
	int		result;
	int		value;
	int		size;
	struct AiCwNumberTable_t	*AiCwNumberTable = sl->AiCwNumberTable;
	struct InkTable_t	*InkTable = sl->InkTable;

	statusP = status + strlen(tbl->keyStr);

	/* 󥯤ܿå
	 * "INK"  "AI:CW" ˥åƤȦ */
	for (i = 0; i < (sizeof(AiCwNumberConstTable) / sizeof(struct AiCwNumberConstTable_t)); i++) {
		if (AiCwNumberTable[i].valid != FALSE) {
			size = AiCwNumberConstTable[i].number;
			break;
		}
	}
	if (i == (sizeof(AiCwNumberConstTable) / sizeof(struct AiCwNumberConstTable_t))) {
		/* 뤫ʬʤΤ7ʬå롣
		 * STATUS_RS ǤåƤΤǥХȯʤȦ */
		size = 7;
	}

	for (i = 0; (*statusP != STATUS_RS) && (i < size); i++) {
		for (j = 0; j < (sizeof(InkM1ConstTable) / sizeof(struct InkM1ConstTable_t)); j++) {
			if (InkM1ConstTable[j].key == *statusP) {
				InkTable[i].indexM1 = j;
				break;
			}
		}
		if (j == (sizeof(InkM1ConstTable) / sizeof(struct InkM1ConstTable_t))) {
			return(FALSE);
		}

		for (j = 0; j < (sizeof(InkM2ConstTable) / sizeof(struct InkM2ConstTable_t)); j++) {
			if (InkM2ConstTable[j].key == *(statusP + 1)) {
				InkTable[i].indexM2 = j;
				break;
			}
		}
		if (j == (sizeof(InkM2ConstTable) / sizeof(struct InkM2ConstTable_t))) {
			return(FALSE);
		}

		result = strX2Int(statusP + 2, 2, &value);
		if (result == FALSE) {
			return(FALSE);
		}
		for (j = 0; j < (sizeof(InkM3ConstTable) / sizeof(struct InkM3ConstTable_t)); j++) {
			/* TBD #### ¤ϥݡоݳ */
			if (InkM3ConstTable[j].key == value) {
				InkTable[i].indexM3 = j;
				break;
			}
		}
		if (j == (sizeof(InkM3ConstTable) / sizeof(struct InkM3ConstTable_t))) {
			return(FALSE);
		}
		InkTable[i].valid = TRUE;
		statusP += 5;
	}

	return(TRUE);
}


/* ----------------------------------------------------------------
 * for IG
 */

/** "IG" ѥơ֥ν
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @return ʤ
 */
static void initIG(StatusList *sl)
{
	int		i;
	struct IgTable_t	*IgTable = sl->IgTable;

	for (i = 0; i < IG_INFO_MAX; i++) {
		IgTable[i].valid = FALSE;
	}
}

/** ɤ߹ơǡ "IG" ѥơ֥ؤ
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @param *status ơǡʸؤΥݥ
 * @param strLength ̤
 * @param *tbl ơǡѥơ֥ؤΥݥ
 * @retval TRUE 
 * @retval FALSE 
 */
static bool setIG(StatusList *sl, char *status, int strLength, StatusTable *tbl)
{
	int		i, j;
	char	*statusP;
	int		result;
	int		value;
	struct IgTable_t	*IgTable = sl->IgTable;

	statusP = status + strlen(tbl->keyStr);

	for (j = 0; (*statusP != STATUS_RS) && (j < IG_INFO_MAX); j++) {
		/* nn1  */
		for (i = 0; i < (sizeof(IgNn1ConstTable) / sizeof(struct IgNn1ConstTable_t)); i++) {
			if (strncmp(statusP, IgNn1ConstTable[i].key, strlen(IgNn1ConstTable[i].key)) == 0) {
				IgTable[j].indexNn1 = i;
				statusP += strlen(IgNn1ConstTable[i].key);
				break;
			}
		}
		if (i == (sizeof(IgNn1ConstTable) / sizeof(struct IgNn1ConstTable_t))) {
			return(FALSE);
		}

		/* nn2  */
		result = strX2Int(statusP, 2, &value);
		if (result == FALSE) {
			if ((((*(statusP)) == 'N') && ((*(statusP + 1)) == 'A')) == 0) {
				IgTable[j].capacity = -1;
			} else {
				return(FALSE);
			}
		} else {
			IgTable[j].capacity = value;
		}
		statusP += 2;

		/* nn3  */
		if (((*(statusP)) == 'N') && ((*(statusP + 1)) == 'A')) {
			IgTable[j].quantity = -1;
		} else {
			return(FALSE);
		}

		IgTable[j].valid = TRUE;

		statusP += 2;
		if (*statusP == STATUS_FS) {
			statusP++;
		}
	}

	return(TRUE);
}


/* ----------------------------------------------------------------
 * for CA
 */

/** "CA" ѥơ֥ν
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @return ʤ
 */
static void initCA(StatusList *sl)
{
	int		i;
	struct CaTable_t	*CaTable = sl->CaTable;

	for (i = 0; i < (sizeof(CaConstTable) / sizeof(struct CaConstTable_t)); i++) {
		CaTable[i].valid = FALSE;
	}
}

/** ɤ߹ơǡ "CA" ѥơ֥ؤ
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @param *status ơǡʸؤΥݥ
 * @param strLength ̤
 * @param *tbl ơǡѥơ֥ؤΥݥ
 * @retval TRUE 
 * @retval FALSE 
 */
static bool setCA(StatusList *sl, char *status, int strLength, StatusTable *tbl)
{
	int		i;
	char	*statusP;
	struct CaTable_t	*CaTable = sl->CaTable;

	statusP = status + strlen(tbl->keyStr);

	for (i = 0; (*statusP != STATUS_RS) && (i < (sizeof(CaConstTable) / sizeof(struct CaConstTable_t))); i++) {
		if (strncmp(statusP, CaConstTable[i].key, strlen(CaConstTable[i].key)) == 0) {
			CaTable[i].valid = TRUE;
			break;
		}
	}
	if (i == (sizeof(CaConstTable) / sizeof(struct CaConstTable_t))) {
		return(FALSE);
	}

	return(TRUE);
}


/* ---------------------------------------------------------------- */

/** ơǡѥơ֥
 * Υơ֥Ƥΰʳϥåоݳ
 * ɬפʤɲäƲ */
StatusTable statusTbl[] = {
	{ "ST:", initST, setST },
	{ "ER:", initER, setER },
	{ "WR:", initWR, setWR },
	{ "IQ:", initIQ, setIQ },
	{ "AI:", initAI, setAI },
	{ "TC:", initTC, setTC },
	{ "INK:", initINK, setINK },
	{ "IG:", initIG, setIG },
	{ "CA:", initCA, setCA },
};



/** ɤ߹ơǡνѴؿƤӽФ
 * @param *p ץ󥿥ơѹ¤ΤؤΥݥ
 * @param *status ơǡʸؤΥݥ
 * @param strLength ơǡʸĹ
 * @retval TRUE 
 * @retval FALSE 
 */
bool setStatus(void *p, char *status, int strLength)
{
	StatusList *sl = p;
	int	result;
	int	len;
	int	i;

	// ʸΥå
	result = strncmp(status, STATUS_START, sizeof(STATUS_START) - 1);
	if (result != 0) {
		return(FALSE);
	}

	len = sizeof(STATUS_START) - 1;
	status += len;
	strLength -= len;

	/*  */
	for (i = 0; i < (sizeof(statusTbl) / sizeof(StatusTable)); i++) {
		/* ƽؿθƽФ */
		statusTbl[i].init(sl);
	}

	while ((strLength > 0) && (*status != STATUS_EOR)) {
		for (i = 0; i < (sizeof(statusTbl) / sizeof(StatusTable)); i++) {
			len = strlen(statusTbl[i].keyStr);
			if (strncmp(status, statusTbl[i].keyStr, len) == 0) {
				result = statusTbl[i].set(sl, status, strLength, &statusTbl[i]);
				/* 顼̵뤷Ʋ³ */
/*
				if (result == FALSE) {
					return(FALSE);
				}
*/
				/* ֤˰¸ʤ褦ˤ뤿ᡢƬõ */
				break;
			}
		}
		if (i == (sizeof(statusTbl) / sizeof(StatusTable))) {
			/* ϿƤʤˤ
			 * ̵뤷ƼΥ쥳ɤõ
			 * TBD ####
			 * ¹ɽ줿顢н褷Ƥ
			 * ȻפäƤԶȤƵ󤬤äƤޤäΤ
			 * DEBUG Τɽˤ */
			DebugPrint("%s(%d) %s: Not entried \"%s\"\n",
					__FILE__, __LINE__, __FUNCTION__, status);
		}

		while ((strLength > 0) && (*status != STATUS_EOR) && (*status != STATUS_RS)) {
			strLength--;
			status++;
		}
		if (*status == STATUS_RS) {
			strLength--;
			status++;
		}
	}

	return(TRUE);
}

/** ơѹ¤Τγݡڤӽ
 * @param ʤ
 * @retval  ݤץ󥿥ơѹ¤ΤؤΥݥ
 * @retval Ի NULL
 */
void *initStatus(void)
{
	void *result;

	result = malloc(sizeof(StatusList));
	if (result != NULL) {
		memset(result, 0, sizeof(StatusList));
	}

	return(result);
}


/** ץ󥿥ơ̵֤ͭå
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @retval 0 Ե
 * @retval 1 ԵǤϤʤ (ʤ)
 */
int isPrinting(void *sl)
{
	int		i;
	struct StTable_t	*StTable = ((StatusList *)(sl))->StTable;

	for (i = 0; i < (sizeof(StConstTable) / sizeof(struct StConstTable_t)); i++) {
		if (StTable[i].valid == TRUE) {
			if (StConstTable[i].status == STATUS_IDLE) {
				/* Ե */
				return(0);
			}
		}
	}
	return(1);
}

/** ץ󥿥ơ̵֤ͭå
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @retval TRUE idling
 * @retval FALSE ¾
 */
int isIdling(void *sl)
{
	int		i;
	struct StTable_t	*StTable = ((StatusList *)(sl))->StTable;

	for (i = 0; i < (sizeof(StConstTable) / sizeof(struct StConstTable_t)); i++) {
		if (StTable[i].valid == TRUE) {
			if (StConstTable[i].status == STATUS_IDLE) {
				/* Ե */
				return(TRUE);
			}
		}
	}
	return(FALSE);
}


/* ---------------------------------------------------------------- */
/* makeXml Ϣؿ */

/** "PrtLocalizationCharacterSet" ؿ
 *  "106" Ǥ
 * @param *sl ̤
 * @param level ̤
 * @param flag ̤
 * @param number ̤
 * @retval  줿ʸǼХåեؤΥݥ
 * @retval Ի NULL
 */
static char *mxPrtLocalizationCharacterSet(StatusList *sl, int level, int flag, int number)
{
	char	*result;
	char	*p = "106";

	result = malloc(strlen(p) + 1);
	if (result != NULL) {
		strcpy(result, p);
	}

	return(result);
}

/** "PrtLocalizationLanguage" ؿ
 *  "JA" Ǥ
 * @param *sl ̤
 * @param level ̤
 * @param flag ̤
 * @param number ̤
 * @retval  줿ʸǼХåեؤΥݥ
 * @retval Ի NULL
 */
static char *mxPrtLocalizationLanguage(StatusList *sl, int level, int flag, int number)
{
	char	*result;
	char	*p = "JA";

	result = malloc(strlen(p) + 1);
	if (result != NULL) {
		strcpy(result, p);
	}

	return(result);
}


/** "PrtLocalizationCharacterSet" ѥơ֥ */
MakeXmlTable MXTPrtLocalizationCharacterSet = {
	"PrtLocalizationCharacterSet",
	NULL,
	mxPrtLocalizationCharacterSet,
	NULL,
	NULL
};

/** "PrtLocalizationLanguage" ѥơ֥ */
MakeXmlTable MXTPrtLocalizationLanguage = {
	"PrtLocalizationLanguage",
	NULL,
	mxPrtLocalizationLanguage,
	NULL,
	&MXTPrtLocalizationCharacterSet
};

/** "PrtLocalizationEntry" ѥơ֥ */
MakeXmlTable MXTPrtLocalizationEntry = {
	"PrtLocalizationEntry",
	NULL,
	NULL,
	&MXTPrtLocalizationLanguage,
	NULL
};


/** "PrtMarkerSuppliesLevel" (󥯻) ؿ
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @param level ̤
 * @param flag IqTable ơ֥Υǥå
 * @param number ̤
 * @retval  줿ʸǼХåեؤΥݥ
 * @retval Ի NULL
 */
static char *mxPrtMarkerSuppliesLevel(StatusList *sl, int level, int flag, int number)
{
	char	*result = NULL;
	char	s[12];
	struct IqTable_t	*IqTable = sl->IqTable;

	if ((flag < (sizeof(IqConstTable) / sizeof(struct IqConstTable_t))) &&
		(IqTable[flag].valid != FALSE)) {
		sprintf(s, "%d", IqTable[flag].quantity);
		result = malloc(strlen(s) + 1);
		if (result != NULL) {
			strcpy(result, s);
		}
	}

	return(result);
}

/** "PrtMarkerSuppliesMaxCapacity" (󥯤κ) ؿ
 *  100 Ǥ
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @param level ̤
 * @param flag IqTable ơ֥Υǥå
 * @param number ̤
 * @retval  줿ʸǼХåեؤΥݥ
 * @retval Ի NULL
 */
static char *mxPrtMarkerSuppliesMaxCapacity(StatusList *sl, int level, int flag, int number)
{
	char	*result = NULL;
	char	*s = "100";
	struct IqTable_t	*IqTable = sl->IqTable;

	/* ץ󥿤֤äƤƤΤ */
	if ((flag < (sizeof(IqConstTable) / sizeof(struct IqConstTable_t))) &&
		(IqTable[flag].valid != FALSE)) {
		result = malloc(strlen(s) + 1);
		if (result != NULL) {
			strcpy(result, s);
		}
	}

	return(result);
}

/** "PrtMarkerSuppliesSupplyUnit" ؿ
 *  19 Ǥ
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @param level ̤
 * @param flag IqTable ơ֥Υǥå
 * @param number ̤
 * @retval  줿ʸǼХåեؤΥݥ
 * @retval Ի NULL
 */
static char *mxPrtMarkerSuppliesSupplyUnit(StatusList *sl, int level, int flag, int number)
{
	char	*result = NULL;
	char	s[12];
	struct IqTable_t	*IqTable = sl->IqTable;

	if ((flag < (sizeof(IqConstTable) / sizeof(struct IqConstTable_t))) &&
		(IqTable[flag].valid != FALSE)) {
		sprintf(s, "%d", 19);
		result = malloc(strlen(s) + 1);
		if (result != NULL) {
			strcpy(result, s);
		}
	}

	return(result);
}

/** "PrtMarkerSuppliesDescription" () ؿ
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @param level ̤
 * @param flag IqTable, IqConstTable ơ֥Υǥå
 * @param number ̤
 * @retval  줿ʸǼХåեؤΥݥ
 * @retval Ի NULL
 */
static char *mxPrtMarkerSuppliesDescription(StatusList *sl, int level, int flag, int number)
{
	char	*result = NULL;
	struct IqTable_t	*IqTable = sl->IqTable;
	char	*po;
	char	*conv;

	if ((flag < (sizeof(IqConstTable) / sizeof(struct IqConstTable_t))) &&
		(IqTable[flag].valid != FALSE)) {
		po = _(IqConstTable[flag].color);
		conv = Utf8Conv(po);
		if (conv != NULL) {
			result = malloc(strlen(conv) + 1);
			if (result != NULL) {
				strcpy(result, conv);
			}
			free(conv);
		} else {
			po = IqConstTable[flag].color;
			result = malloc(strlen(po) + 1);
			if (result != NULL) {
				strcpy(result, po);
			}
		}
	}

	return(result);
}

/** "PrtMarkerSuppliesType" ؿ
 *  6 Ǥ
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @param level ̤
 * @param flag IqTable ơ֥Υǥå
 * @param number ̤
 * @retval  줿ʸǼХåեؤΥݥ
 * @retval Ի NULL
 */
static char *mxPrtMarkerSuppliesType(StatusList *sl, int level, int flag, int number)
{
	char	*result = NULL;
	char	s[12];
	struct IqTable_t	*IqTable = sl->IqTable;

	if ((flag < (sizeof(IqConstTable) / sizeof(struct IqConstTable_t))) &&
		(IqTable[flag].valid != FALSE)) {
		sprintf(s, "%d", 6);
		result = malloc(strlen(s) + 1);
		if (result != NULL) {
			strcpy(result, s);
		}
	}

	return(result);
}

/** "PrtMarkerSuppliesIndex" ؿ
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @param level ̤
 * @param flag IqTable ơ֥Υǥå
 * @param number PrtMarkerSuppliesIndex
 * @retval  줿ʸǼХåեؤΥݥ
 * @retval Ի NULL
 */
static char *mxPrtMarkerSuppliesIndex(StatusList *sl, int level, int flag, int number)
{
	char	*result = NULL;
	char	s[12];
	struct IqTable_t	*IqTable = sl->IqTable;

	if ((flag < (sizeof(IqConstTable) / sizeof(struct IqConstTable_t))) &&
		(IqTable[flag].valid != FALSE)) {
		sprintf(s, "%d", number);
		result = malloc(strlen(s) + 1);
		if (result != NULL) {
			strcpy(result, s);
		}
	}

	return(result);
}

/** "PrtMarkerSuppliesLevel" ѥơ֥ */
MakeXmlTable MXTPrtMarkerSuppliesLevel = {
	"PrtMarkerSuppliesLevel",
	NULL,
	mxPrtMarkerSuppliesLevel,
	NULL,
	NULL
};

/** "PrtMarkerSuppliesMaxCapacity" ѥơ֥ */
MakeXmlTable MXTPrtMarkerSuppliesMaxCapacity = {
	"PrtMarkerSuppliesMaxCapacity",
	NULL,
	mxPrtMarkerSuppliesMaxCapacity,
	NULL,
	&MXTPrtMarkerSuppliesLevel
};

/** "PrtMarkerSuppliesSupplyUnit" ѥơ֥ */
MakeXmlTable MXTPrtMarkerSuppliesSupplyUnit = {
	"PrtMarkerSuppliesSupplyUnit",
	NULL,
	mxPrtMarkerSuppliesSupplyUnit,
	NULL,
	&MXTPrtMarkerSuppliesMaxCapacity
};

/** "PrtMarkerSuppliesDescription" ѥơ֥ */
MakeXmlTable MXTPrtMarkerSuppliesDescription = {
	"PrtMarkerSuppliesDescription",
	NULL,
	mxPrtMarkerSuppliesDescription,
	NULL,
	&MXTPrtMarkerSuppliesSupplyUnit
};

/** "PrtMarkerSuppliesType" ѥơ֥ */
MakeXmlTable MXTPrtMarkerSuppliesType = {
	"PrtMarkerSuppliesType",
	NULL,
	mxPrtMarkerSuppliesType,
	NULL,
	&MXTPrtMarkerSuppliesDescription
};

/** "PrtMarkerSuppliesIndex" ѥơ֥ */
MakeXmlTable MXTPrtMarkerSuppliesIndex = {
	"PrtMarkerSuppliesIndex",
	NULL,
	mxPrtMarkerSuppliesIndex,
	NULL,
	&MXTPrtMarkerSuppliesType
};

/** "PrtMarkerSuppliesEntry" ѥơ֥ */
MakeXmlTable MXTPrtMarkerSuppliesEntryFunc = {
	"PrtMarkerSuppliesEntry",
	NULL,
	NULL,
	&MXTPrtMarkerSuppliesIndex,
	NULL
};


static char *makeXmlSub(StatusList *sl, MakeXmlTable *tbl, int level, int flag, int number);


/** "PrtMarkerSuppliesEntry" Ϣؿ
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @param level XML ΥǥȤο
 * @param flag ̤
 * @param number ̤
 * @retval  줿ʸǼХåեؤΥݥ
 * @retval Ի NULL
 */
static char *mxPrtMarkerSuppliesEntry(StatusList *sl, int level, int flag, int number)
{
	char	*result = NULL;
	char	*res;
	int		len;
	int		i;
	int		j = 1;
	struct IqTable_t	*IqTable = sl->IqTable;

	for (i = 0; i < (sizeof(IqConstTable) / sizeof(struct IqConstTable_t)); i++) {
		if (IqTable[i].valid != FALSE) {
			res = makeXmlSub(sl, &MXTPrtMarkerSuppliesEntryFunc, level, i, j++);
			if (res != NULL) {
				len = strlen(res);
				if (result != NULL) {
					len += strlen(result);
					result = realloc(result, len + 1);
					strcat(result, res);
					free(res);
				} else {
					result = res;
				}
			}
		}
	}

	return(result);
}

/* ---------------------------------------------------------------- */

/** 顼 "PrtAlertSeverityLevel" ؿ
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @param level ̤
 * @param flag ErTable, ErConstTable ơ֥Υǥå
 * @param number ̤
 * @retval  줿ʸǼХåեؤΥݥ
 * @retval Ի NULL
 */
static char *mxPrtAlertSeverityLevelEr(StatusList *sl, int level, int flag, int number)
{
	char	*result = NULL;
	char	s[12];
	struct ErTable_t	*ErTable = sl->ErTable;

	if ((flag < (sizeof(ErConstTable) / sizeof(struct ErConstTable_t))) &&
		(ErTable[flag].valid != FALSE)) {
		sprintf(s, "%d", ErConstTable[flag].level);
		result = malloc(strlen(s) + 1);
		if (result != NULL) {
			strcpy(result, s);
		}
	}

	return(result);
}

/** 顼 "PrtAlertGroup" ؿ
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @param level ̤
 * @param flag ErTable, ErConstTable ơ֥Υǥå
 * @param number ̤
 * @retval  줿ʸǼХåեؤΥݥ
 * @retval Ի NULL
 */
static char *mxPrtAlertGroupEr(StatusList *sl, int level, int flag, int number)
{
	char	*result = NULL;
	char	s[12];
	struct ErTable_t	*ErTable = sl->ErTable;

	if ((flag < (sizeof(ErConstTable) / sizeof(struct ErConstTable_t))) &&
		(ErTable[flag].valid != FALSE)) {
		sprintf(s, "%d", ErConstTable[flag].group);
		result = malloc(strlen(s) + 1);
		if (result != NULL) {
			strcpy(result, s);
		}
	}

	return(result);
}

/** 顼 "PrtAlertGroupIndex" ؿ
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @param level ̤
 * @param flag ErTable ơ֥Υǥå
 * @param number XML ǤΥǥå
 * @retval  줿ʸǼХåեؤΥݥ
 * @retval Ի NULL
 */
static char *mxPrtAlertGroupIndexEr(StatusList *sl, int level, int flag, int number)
{
	char	*result = NULL;
	char	s[12];
	struct ErTable_t	*ErTable = sl->ErTable;

	if ((flag < (sizeof(ErConstTable) / sizeof(struct ErConstTable_t))) &&
		(ErTable[flag].valid != FALSE)) {
		sprintf(s, "%d", number);
		result = malloc(strlen(s) + 1);
		if (result != NULL) {
			strcpy(result, s);
		}
	}

	return(result);
}

/** 顼 "PrtAlertCode" ؿ
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @param level ̤
 * @param flag ErTable, ErConstTable ơ֥Υǥå
 * @param number ̤
 * @retval  줿ʸǼХåեؤΥݥ
 * @retval Ի NULL
 */
static char *mxPrtAlertCodeEr(StatusList *sl, int level, int flag, int number)
{
	char	*result = NULL;
	char	s[12];
	struct ErTable_t	*ErTable = sl->ErTable;

	if ((flag < (sizeof(ErConstTable) / sizeof(struct ErConstTable_t))) &&
		(ErTable[flag].valid != FALSE)) {
		sprintf(s, "%d", ErConstTable[flag].code);
		result = malloc(strlen(s) + 1);
		if (result != NULL) {
			strcpy(result, s);
		}
	}

	return(result);
}

/** 顼 "PrtAlertDescription" ؿ
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @param level ̤
 * @param flag ErTable, ErConstTable ơ֥Υǥå
 * @param number ̤
 * @retval  줿ʸǼХåեؤΥݥ
 * @retval Ի NULL
 */
static char *mxPrtAlertDescriptionEr(StatusList *sl, int level, int flag, int number)
{
	char	*result = NULL;
	struct ErTable_t	*ErTable = sl->ErTable;
	char	*po;
	char	*conv;

	if ((flag < (sizeof(ErConstTable) / sizeof(struct ErConstTable_t))) &&
		(ErTable[flag].valid != FALSE)) {
		po = _(ErConstTable[flag].str);
		conv = Utf8Conv(po);
		if (conv != NULL) {
			result = malloc(strlen(conv) + 1);
			if (result != NULL) {
				strcpy(result, conv);
			}
			free(conv);
		} else {
			po = ErConstTable[flag].str;
			result = malloc(strlen(po) + 1);
			if (result != NULL) {
				strcpy(result, po);
			}
		}
	}

	return(result);
}

/** 顼 "PrtAlertDescription" ѥơ֥ */
MakeXmlTable MXTPrtAlertDescriptionEr = {
	"PrtAlertDescription",
	NULL,
	mxPrtAlertDescriptionEr,
	NULL,
	NULL
};

/** 顼 "PrtAlertCode" ѥơ֥ */
MakeXmlTable MXTPrtAlertCodeEr = {
	"PrtAlertCode",
	NULL,
	mxPrtAlertCodeEr,
	NULL,
	&MXTPrtAlertDescriptionEr
};

/** 顼 "PrtAlertGroupIndex" ѥơ֥ */
MakeXmlTable MXTPrtAlertGroupIndexEr = {
	"PrtAlertGroupIndex",
	NULL,
	mxPrtAlertGroupIndexEr,
	NULL,
	&MXTPrtAlertCodeEr
};

/** 顼 "PrtAlertGroup" ѥơ֥ */
MakeXmlTable MXTPrtAlertGroupEr = {
	"PrtAlertGroup",
	NULL,
	mxPrtAlertGroupEr,
	NULL,
	&MXTPrtAlertGroupIndexEr
};

/** 顼 "PrtAlertSeverityLevel" ѥơ֥ */
MakeXmlTable MXTPrtAlertSeverityLevelEr = {
	"PrtAlertSeverityLevel",
	NULL,
	mxPrtAlertSeverityLevelEr,
	NULL,
	&MXTPrtAlertGroupEr
};

/** 顼 "PrtAlertEntry" Ϣѥơ֥ */
MakeXmlTable MXTPrtAlertEntryErFunc = {
	"PrtAlertEntry",
	NULL,
	NULL,
	&MXTPrtAlertSeverityLevelEr,
	NULL
};

/* ---------------- */

/** ˥󥰤 "PrtAlertSeverityLevel" ؿ
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @param level ̤
 * @param flag WrTable, WrConstTable ơ֥Υǥå
 * @param number ̤
 * @retval  줿ʸǼХåեؤΥݥ
 * @retval Ի NULL
 */
static char *mxPrtAlertSeverityLevelWr(StatusList *sl, int level, int flag, int number)
{
	char	*result = NULL;
	char	s[12];
	struct WrTable_t	*WrTable = sl->WrTable;

	if ((flag < (sizeof(WrConstTable) / sizeof(struct WrConstTable_t))) &&
		(WrTable[flag].valid != FALSE)) {
		sprintf(s, "%d", WrConstTable[flag].level);
		result = malloc(strlen(s) + 1);
		if (result != NULL) {
			strcpy(result, s);
		}
	}

	return(result);
}

/** ˥󥰤 "PrtAlertGroup" ؿ
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @param level ̤
 * @param flag WrTable, WrConstTable ơ֥Υǥå
 * @param number ̤
 * @retval  줿ʸǼХåեؤΥݥ
 * @retval Ի NULL
 */
static char *mxPrtAlertGroupWr(StatusList *sl, int level, int flag, int number)
{
	char	*result = NULL;
	char	s[12];
	struct WrTable_t	*WrTable = sl->WrTable;

	if ((flag < (sizeof(WrConstTable) / sizeof(struct WrConstTable_t))) &&
		(WrTable[flag].valid != FALSE)) {
		sprintf(s, "%d", WrConstTable[flag].group);
		result = malloc(strlen(s) + 1);
		if (result != NULL) {
			strcpy(result, s);
		}
	}

	return(result);
}

/** ˥󥰤 "PrtAlertGroupIndex" ؿ
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @param level ̤
 * @param flag WrTable ơ֥Υǥå
 * @param number XML ǤΥǥå
 * @retval  줿ʸǼХåեؤΥݥ
 * @retval Ի NULL
 */
static char *mxPrtAlertGroupIndexWr(StatusList *sl, int level, int flag, int number)
{
	char	*result = NULL;
	char	s[12];
	struct WrTable_t	*WrTable = sl->WrTable;

	if ((flag < (sizeof(WrConstTable) / sizeof(struct WrConstTable_t))) &&
		(WrTable[flag].valid != FALSE)) {
		sprintf(s, "%d", number);
		result = malloc(strlen(s) + 1);
		if (result != NULL) {
			strcpy(result, s);
		}
	}

	return(result);
}

/** ˥󥰤 "PrtAlertCode" ؿ
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @param level ̤
 * @param flag WrTable, WrConstTable ơ֥Υǥå
 * @param number ̤
 * @retval  줿ʸǼХåեؤΥݥ
 * @retval Ի NULL
 */
static char *mxPrtAlertCodeWr(StatusList *sl, int level, int flag, int number)
{
	char	*result = NULL;
	char	s[12];
	struct WrTable_t	*WrTable = sl->WrTable;

	if ((flag < (sizeof(WrConstTable) / sizeof(struct WrConstTable_t))) &&
		(WrTable[flag].valid != FALSE)) {
		sprintf(s, "%d", WrConstTable[flag].code);
		result = malloc(strlen(s) + 1);
		if (result != NULL) {
			strcpy(result, s);
		}
	}

	return(result);
}

/** ˥󥰤 "PrtAlertDescription" ؿ
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @param level ̤
 * @param flag WrTable, WrConstTable ơ֥Υǥå
 * @param number ̤
 * @retval  줿ʸǼХåեؤΥݥ
 * @retval Ի NULL
 */
static char *mxPrtAlertDescriptionWr(StatusList *sl, int level, int flag, int number)
{
	char	*result = NULL;
	struct WrTable_t	*WrTable = sl->WrTable;
	char	*po;
	char	*conv;

	if ((flag < (sizeof(WrConstTable) / sizeof(struct WrConstTable_t))) &&
		(WrTable[flag].valid != FALSE)) {
		po = _(WrConstTable[flag].str);
		conv = Utf8Conv(po);
		if (conv != NULL) {
			result = malloc(strlen(conv) + 1);
			if (result != NULL) {
				strcpy(result, conv);
			}
			free(conv);
		} else {
			po = WrConstTable[flag].str;
			result = malloc(strlen(po) + 1);
			if (result != NULL) {
				strcpy(result, po);
			}
		}
	}

	return(result);
}

/** ˥󥰤 "PrtAlertDescription" ѥơ֥ */
MakeXmlTable MXTPrtAlertDescriptionWr = {
	"PrtAlertDescription",
	NULL,
	mxPrtAlertDescriptionWr,
	NULL,
	NULL
};

/** ˥󥰤 "PrtAlertCode" ѥơ֥ */
MakeXmlTable MXTPrtAlertCodeWr = {
	"PrtAlertCode",
	NULL,
	mxPrtAlertCodeWr,
	NULL,
	&MXTPrtAlertDescriptionWr
};

/** ˥󥰤 "PrtAlertGroupIndex" ѥơ֥ */
MakeXmlTable MXTPrtAlertGroupIndexWr = {
	"PrtAlertGroupIndex",
	NULL,
	mxPrtAlertGroupIndexWr,
	NULL,
	&MXTPrtAlertCodeWr
};

/** ˥󥰤 "PrtAlertGroup" ѥơ֥ */
MakeXmlTable MXTPrtAlertGroupWr = {
	"PrtAlertGroup",
	NULL,
	mxPrtAlertGroupWr,
	NULL,
	&MXTPrtAlertGroupIndexWr
};

/** ˥󥰤 "PrtAlertSeverityLevel" ѥơ֥ */
MakeXmlTable MXTPrtAlertSeverityLevelWr = {
	"PrtAlertSeverityLevel",
	NULL,
	mxPrtAlertSeverityLevelWr,
	NULL,
	&MXTPrtAlertGroupWr
};

/** ˥󥰤 "PrtAlertEntry" Ϣѥơ֥ */
MakeXmlTable MXTPrtAlertEntryWrFunc = {
	"PrtAlertEntry",
	NULL,
	NULL,
	&MXTPrtAlertSeverityLevelWr,
	NULL
};

/** "PrtAlertEntry" Ϣؿ
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @param level XML ΥǥȤο
 * @param flag ̤
 * @param number ̤
 * @retval  줿ʸǼХåեؤΥݥ
 * @retval Ի NULL
 */
static char *mxPrtAlertEntry(StatusList *sl, int level, int flag, int number)
{
	char	*result = NULL;
	char	*res;
	int		len;
	int		i;
	int		j = 1;
	struct WrTable_t	*WrTable = sl->WrTable;
	struct ErTable_t	*ErTable = sl->ErTable;

	for (i = 0; i < (sizeof(WrConstTable) / sizeof(struct WrConstTable_t)); i++) {
		if (WrTable[i].valid != FALSE) {
			res = makeXmlSub(sl, &MXTPrtAlertEntryWrFunc, level, i, j++);
			if (res != NULL) {
				len = strlen(res);
				if (result != NULL) {
					len += strlen(result);
					result = realloc(result, len + 1);
					strcat(result, res);
					free(res);
				} else {
					result = res;
				}
			}
		}
	}

	for (i = 0; i < (sizeof(ErConstTable) / sizeof(struct ErConstTable_t)); i++) {
		if (ErTable[i].valid != FALSE) {
			res = makeXmlSub(sl, &MXTPrtAlertEntryErFunc, level, i, j++);
			if (res != NULL) {
				len = strlen(res);
				if (result != NULL) {
					len += strlen(result);
					result = realloc(result, len + 1);
					strcat(result, res);
					free(res);
				} else {
					result = res;
				}
			}
		}
	}

	return(result);
}

/* ---------------------------------------------------------------- */

/** "PrtStatusDescription" ؿ
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @param level ̤
 * @param flag ̤
 * @param number ̤
 * @retval  줿ʸǼХåեؤΥݥ
 * @retval Ի NULL
 */
static char *mxPrtStatusDescription(StatusList *sl, int level, int flag, int number)
{
	char	*result = NULL;
	int		i;
	struct StTable_t	*StTable = sl->StTable;
	char	*po;
	char	*conv;

	for (i = 0; i < (sizeof(StConstTable) / sizeof(struct StConstTable_t)); i++) {
		if (StTable[i].valid != FALSE) {
			po = _(StConstTable[i].str);
			conv = Utf8Conv(po);
			if (conv != NULL) {
				result = malloc(strlen(conv) + 1);
				if (result != NULL) {
					strcpy(result, conv);
				}
				free(conv);
			} else {
				po = StConstTable[i].str;
				result = malloc(strlen(po) + 1);
				if (result != NULL) {
					strcpy(result, po);
				}
			}
			break;
		}
	}

	return(result);
}

/** "PrtStatusType" ؿ
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @param level ̤
 * @param flag ̤
 * @param number ̤
 * @retval  줿ʸǼХåեؤΥݥ
 * @retval Ի NULL
 */
static char *mxPrtStatusType(StatusList *sl, int level, int flag, int number)
{
	char	*result = NULL;
	char	s[12];
	int		i;
	struct StTable_t	*StTable = sl->StTable;

	for (i = 0; i < (sizeof(StConstTable) / sizeof(struct StConstTable_t)); i++) {
		if (StTable[i].valid != FALSE) {
			sprintf(s, "%d", StConstTable[i].mib);
			result = malloc(strlen(s) + 1);
			if (result != NULL) {
				strcpy(result, s);
			}
			break;
		}
	}

	return(result);
}

/** "PrtStatusIndex" ؿ
 *  1 Ǥ
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @param level ̤
 * @param flag ̤
 * @param number ̤
 * @retval  줿ʸǼХåեؤΥݥ
 * @retval Ի NULL
 */
static char *mxPrtStatusIndex(StatusList *sl, int level, int flag, int number)
{
	char	*result = NULL;
	char	s[12];
	int		i;
	struct StTable_t	*StTable = sl->StTable;

	for (i = 0; i < (sizeof(StConstTable) / sizeof(struct StConstTable_t)); i++) {
		if (StTable[i].valid != FALSE) {
			sprintf(s, "%d", 1);
			result = malloc(strlen(s) + 1);
			if (result != NULL) {
				strcpy(result, s);
			}
			break;
		}
	}

	return(result);
}

/* ---------------------------------------------------------------- */
/* ơ֥ */

/** "PrtStatusDescription" ѥơ֥ */
MakeXmlTable MXTPrtStatusDescription = {
	"PrtStatusDescription",
	NULL,
	mxPrtStatusDescription,
	NULL,
	NULL
};

/** "PrtStatusType" ѥơ֥ */
MakeXmlTable MXTPrtStatusType = {
	"PrtStatusType",
	NULL,
	mxPrtStatusType,
	NULL,
	&MXTPrtStatusDescription,
};

/** "PrtStatusIndex" ѥơ֥ */
MakeXmlTable MXTPrtStatusIndex = {
	"PrtStatusIndex",
	NULL,
	mxPrtStatusIndex,
	NULL,
	&MXTPrtStatusType,
};

/** "PrtStatusEntry" ѥơ֥ */
MakeXmlTable MXTPrtStatusEntry = {
	"PrtStatusEntry",
	NULL,
	NULL,
	&MXTPrtStatusIndex,
	NULL,
};

/** "PrtStatusTable" ѥơ֥ */
MakeXmlTable MXTPrtStatusTable = {
	"PrtStatusTable",
	NULL,
	NULL,
	&MXTPrtStatusEntry,
	NULL
};

/** "PrtStatus" ѥơ֥
 *  "1" Ǥ
 */
MakeXmlTable MXTPrtStatus = {
	"PrtStatus",
	"1",
	NULL,
	NULL,
	&MXTPrtStatusTable
};

/** "PrtAlertEntry" Ϣѥơ֥ */
MakeXmlTable MXTPrtAlertEntry = {
	NULL,
	NULL,
	mxPrtAlertEntry,
	NULL,
	NULL
};

/** "PrtAlertTable" ѥơ֥ */
MakeXmlTable MXTPrtAlertTable = {
	"PrtAlertTable",
	NULL,
	NULL,
	&MXTPrtAlertEntry,
	NULL
//	&MXTPrtStatus
};

/** "PrtMarkerSuppliesEntry" Ϣѥơ֥ */
MakeXmlTable MXTPrtMarkerSuppliesEntry = {
	NULL,
	NULL,
	mxPrtMarkerSuppliesEntry,
	NULL,
	NULL
};

/** "PrtMarkerSuppliesTable" ѥơ֥ */
MakeXmlTable MXTPrtMarkerSuppliesTable = {
	"PrtMarkerSuppliesTable",
	NULL,
	NULL,
	&MXTPrtMarkerSuppliesEntry,
	&MXTPrtAlertTable
};

/** "PrtLocalizationTable" ѥơ֥ */
MakeXmlTable MXTPrtLocalizationTable = {
	"PrtLocalizationTable",
	NULL,
	NULL,
	&MXTPrtLocalizationEntry,
	&MXTPrtMarkerSuppliesTable
};



/** Ϥ줿ʸ󤫤 XML ʸ
 * @param *sl ̤
 * @param *src1 Ƭʸ
 * @param *src2 ɲäʸ
 * @param level XML ΥǥȤο
 * @param brace '<' νˡ
 * @param release *src1, *src2 뤫ݤ
 * @retval  줿ʸǼХåեؤΥݥ
 * @retval Ի NULL
 */
static char *makeXmlContent(StatusList *sl, char *src1, char *src2, int level, int brace, int release)
{
	int		i;
	int		len;
	char	*result;

	len = 0;
	if (src1 != NULL) {
		len += strlen(src1);
	}
	if (src2 != NULL) {
		len += strlen(src2);
	}
	if (len == 0) {
		DebugPrint("%s(%d) %s: String length is 0\n", __FILE__, __LINE__, __FUNCTION__);
		return(NULL);
	}
	len++;

	if (brace == MAKE_XML_CONTENT_BRACE_START) {
		len += 2;
	} else if (brace == MAKE_XML_CONTENT_BRACE_END) {
		len += 4;
	}

	if ((brace == MAKE_XML_CONTENT_BRACE_START) || (brace == MAKE_XML_CONTENT_BRACE_END)) {
		len += level;
	}

	result = malloc(len);
	if (result == NULL) {
		DebugPrint("%s(%d) %s: Not enough core\n src1: %s\n src2: %s\n",
				   __FILE__, __LINE__, __FUNCTION__, src1, src2);
		goto exitFunc;
		return(NULL);
	}

	*result = '\0';								/*  */
	if (src1 != NULL) {
		if (brace == MAKE_XML_CONTENT_BRACE_START) {
			for (i = 0; i < level; i++) {
				strcat(result, "\t");
			}
			strcat(result, "<");
		}
		strcat(result, src1);
		if (brace == MAKE_XML_CONTENT_BRACE_START) {
			strcat(result, ">");
		}
	}
	if (src2 != NULL) {
		if (brace == MAKE_XML_CONTENT_BRACE_END) {
			if ((src1 != NULL) && (src1[strlen(src1) - 1] == '\n')) {
				for (i = 0; i < level; i++) {
					strcat(result, "\t");
				}
			}
			strcat(result, "</");
		}
		strcat(result, src2);
		if (brace == MAKE_XML_CONTENT_BRACE_END) {
			strcat(result, ">\n");
		}
	}


exitFunc:
	if ((src1 != NULL) && ((release & MAKE_XML_CONTENT_RELEASE_1) != 0)) {
		free(src1);
	}
	if ((src2 != NULL) &&((release & MAKE_XML_CONTENT_RELEASE_2) != 0)) {
		free(src2);
	}

	return(result);
}

/** Ϥ줿ơ֥˱ƽ
 * ƵԤʤ
 * @param *sl ץ󥿥ơѹ¤ΤؤΥݥ
 * @param *tbl оݤ MakeXmlTable ơ֥ؤΥݥ
 * @param level XML ΥǥȤο
 * @param flag ѥե饰 (ƤӽФؿˤäưʤ)
 * @param number ѥ (ƤӽФؿˤäưʤ)
 * @retval  줿ʸǼХåեؤΥݥ
 * @retval Ի NULL
 */
static char *makeXmlSub(StatusList *sl, MakeXmlTable *tbl, int level, int flag, int number)
{
	char	*result = NULL;
	char	*res;
	char	*p;

	if (tbl->text != NULL) {
		result = makeXmlContent(sl, tbl->name, tbl->text, level, MAKE_XML_CONTENT_BRACE_START,
							 MAKE_XML_CONTENT_RELEASE_NON);
		result = makeXmlContent(sl, result, tbl->name, 0, MAKE_XML_CONTENT_BRACE_END,
							 MAKE_XML_CONTENT_RELEASE_1);
	}

	if (tbl->make != NULL) {
		res = tbl->make(sl, level, flag, number);
		if (res != NULL) {
			if (result == NULL) {
				if (tbl->name != NULL) {
					if (*res == '\t') {
						p = "\n";
					} else {
						p = NULL;
					}
					result = makeXmlContent(sl, tbl->name, p, level, MAKE_XML_CONTENT_BRACE_START,
											MAKE_XML_CONTENT_RELEASE_NON);
				}
			}
			if (result != NULL) {
				result = makeXmlContent(sl, result, res, 0, MAKE_XML_CONTENT_BRACE_NON,
										MAKE_XML_CONTENT_RELEASE_1 | MAKE_XML_CONTENT_RELEASE_2);
			} else {
				result = res;
			}
		}
	}

	if (tbl->child != NULL) {
		/* ؤå */
		res = makeXmlSub(sl, tbl->child, level + 1, flag, number);
		if (res != NULL) {
			if (result == NULL) {
				result = makeXmlContent(sl, tbl->name, "\n", level, MAKE_XML_CONTENT_BRACE_START,
										MAKE_XML_CONTENT_RELEASE_NON);
			}
			result = makeXmlContent(sl, result, res, 0, MAKE_XML_CONTENT_BRACE_NON,
									MAKE_XML_CONTENT_RELEASE_1 | MAKE_XML_CONTENT_RELEASE_2);
		}
	}

	if ((result != NULL) && (tbl->text == NULL)) {
		/* Ĥ */
		result = makeXmlContent(sl, result, tbl->name, level, MAKE_XML_CONTENT_BRACE_END,
								MAKE_XML_CONTENT_RELEASE_1);
	}

	if (tbl->next != NULL) {
		/* ƱؤΤ */
		res = makeXmlSub(sl, tbl->next, level, flag, number);
		if (res != NULL) {
			if (result != NULL) {
				result = makeXmlContent(sl, result, res, 0, MAKE_XML_CONTENT_BRACE_NON,
										MAKE_XML_CONTENT_RELEASE_1 | MAKE_XML_CONTENT_RELEASE_2);
			} else {
				result = res;
			}
		}
	}

	return(result);
}


/** XML 
 * @param *p ץ󥿥ơѹ¤ΤؤΥݥ
 * @param *xml ʸǼХåեؤΥݥ
 * @param xmlLength *xml Ĺ
 * @param *pLang ʸθ
 * @retval TRUE 
 * @retval FALSE 
 */
bool makeXml(void *p, char *xml, int xmlLength, char *pLang)
{
	StatusList *sl = p;
	char	*result;
//	int		res;
	int		size;
	int		len;

	/* locale б */
	setlocale(LC_ALL, pLang);
	bindtextdomain(PACKAGE, LOCALE_DIR);
	textdomain(PACKAGE);

	/* XML Υإåʬ */
	size = sizeof(XML_HEADER) - 1;
	xmlLength -= (size + 1);
	if (xmlLength < 0) {
		return(FALSE);
	}
	memcpy(xml, XML_HEADER, size + 1);
	xml += size;

	/* Printer Υإåʬ */
	size = sizeof(XML_PMIBS) - 1;
	xmlLength -= (size + 1);
	if (xmlLength < 0) {
		return(FALSE);
	}
	memcpy(xml, XML_PMIBS, size + 1);
	xml += size;


	/* Printer-MIB Υإåʬ */
	size = sizeof(XML_MIBS) - 1;
	xmlLength -= size;
	if (xmlLength < 0) {
		return(FALSE);
	}
	memcpy(xml, XML_MIBS, size + 1);
	xml += size;

	/* Printer-MIB γϥơ֥ꤹ */
	result = makeXmlSub(sl, &MXTPrtLocalizationTable, 1, 0, 0);
	if (result == NULL) {
		return(FALSE);
	}

	len = strlen(result);
	if (xmlLength < len) {
		free(result);
		return(FALSE);
	}
	if (len < 0) {
		return(FALSE);
	}

	memcpy(xml, result, len + 1);
	free(result);

	xmlLength -= len;
	xml += len;

	/* Printer-MIB Υեåʬ */
	size = sizeof(XML_MIBE) - 1;
	xmlLength -= size;
	if (xmlLength < 0) {
		return(FALSE);
	}
	memcpy(xml, XML_MIBE, size + 1);
	xml += size;


	/* PrtStatus γϥơ֥ꤹ */
	result = makeXmlSub(sl, &MXTPrtStatus, 0, 0, 0);
	if (result == NULL) {
		return(FALSE);
	}

	len = strlen(result);
	if (xmlLength < len) {
		free(result);
		return(FALSE);
	}
	if (len < 0) {
		return(FALSE);
	}

	memcpy(xml, result, len + 1);				/* NULL ޤ */
	free(result);

	xmlLength -= len;
	xml += len;

	/* Printer Υեåʬ */
	size = sizeof(XML_PMIBE) - 1;
	xmlLength -= (size + 1);
	if (xmlLength < 0) {
		return(FALSE);
	}
	memcpy(xml, XML_PMIBE, size + 1);
	xml += size;

	return(TRUE);
}

/* end of bidiEpsonStatus.c */
