/* bidiPXV700.c -- Bidi for EPSON Ink Jet Printer.
 * Copyright (C) 2003 EPSON KOWA Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: bidiPXV700.c,v 1.25 2004/02/17 18:44:38 toko Exp $ */

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <errno.h>
#include <string.h>
#include <signal.h>
#include <setjmp.h>

#include <pthread.h>
#include <sys/select.h>

#if defined(DEBUG)
#include <stdarg.h>
#endif /* DEBUG */

#include "bidiPlugin.h"
#include "bidiEpson.h"
#include "bidiLibEpson.h"
#include "bidiEpsonStatus.h"



#define READ_RETRY_NUM					2		/**< ץ󥿤饹ơɤ߽Ф˼ԤκƻԲ */
#define THREAD_READ_WAIT_TIME			5		/**< read thread ưֳ (unit is second) */
//#define THREAD_READ_WAIT_TIME			2		/**< read thread ưֳ (unit is second) */
#define THREAD_WRITE_WAIT_TIME			(10 * 1000)	/**< write thread Ԥ (unit is micro second) */
#define PRINTER_STATUS_INVALID_TIME		1		/**< ץ󥿤饹ơɤ߽Ф˼ԤΥ顼ȸޤǤλ (unit is second) */


int signalCatchedFlag;						/**< signal ­  signal ID  */


/** Debugѥåϴؿ
 * @param printf ؿƱΰȤ
 * @return ̵
 */
static void DebugPrint(const char *fmt, ...)
{
#ifdef DEBUG
	va_list ap;

	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	fflush(stderr);
#endif
	return;
}

/** DebugѤ16ʤɽ
 * @param *p ǡؤΥݥ
 * @param length ѴĹ
 * @return ̵
 */
static void PrintHex(char *p, int length)
{
#ifdef DEBUG
	int		i, j;

	for (i = 0; i < length;  ) {
		fprintf(stderr, "%04X ", i);
		for (j = 0; (j < 16) && ((i + j) < length); j++) {
			if (j == 8) {
				fprintf(stderr, " ");
			}
			fprintf(stderr, "%02X ", p[i + j]);
		}
		for ( ; j < 16; j++) {
			if (j == 8) {
				fprintf(stderr, " ");
			}
			fprintf(stderr, "   ");
		}
		fprintf(stderr, "  ");
		for (j = 0; (j < 16) && ((i + j) < length); j++) {
			if (j == 8) {
				fprintf(stderr, " ");
			}
			if ((p[i + j] >= ' ') && (p[i + j] < 0x7f)) {
				fprintf(stderr, "%c", p[i + j]);
			} else {
				fprintf(stderr, ".");
			}
		}
		fprintf(stderr, "\n");
		i += j;
	}

#endif
	return;
}


/** wait Ѵؿ
 * sleep ؿδ select ؿѤ
 * @param *timeout Ԥ
 * @retval 0 
 * @retval -1 顼ȯ
 */
static int SafeSleep(struct timeval *timeout)
{
	int	result;

	result = select(0, NULL, NULL, NULL, timeout);

	return(result);
}


/** signal ȯν
 * SIGALRM  SIGHUP ξϡޤ줿򵭲ΤߤǸ롣
 * ¾ξϻ
 * @param signalId ȯʥ
 * @return ̵
 */
static void SignalHandler(int signalId)
{
	DebugPrint("%s(%d) %s: signalId = %3d, %08X\n", __FILE__, __LINE__, __FUNCTION__, signalId, signalId);

	if ((signalId == SIGALRM) || (signalId == SIGHUP)) {
		signalCatchedFlag = signalId;
		return;
	}

	/* 嵭ʳξľ˻ */
	DebugPrint("\t interrupted, exit\n");
	exit(1);
}

/** ꤵƤ signal 򤹤
 * @param *ss ʥ
 * @retval TRUE Ǥ
 * @retval FALSE ˼Ԥ
 */
static bool SaveSignal(savedSignal *ss)
{

	/* Library פ Process פξб */
	ss->term = signal(SIGTERM, SignalHandler);
	if (ss->term == SIG_ERR) {
		DebugPrint("%s(%d) %s: %d, %s\n", __FILE__, __LINE__, __FUNCTION__, errno, strerror(errno));
		return(FALSE);
	}
	ss->pipe = signal(SIGPIPE, SignalHandler);
	if (ss->pipe == SIG_ERR) {
		DebugPrint("%s(%d) %s: %d, %s\n", __FILE__, __LINE__, __FUNCTION__, errno, strerror(errno));
		/* ᤹ */
		signal(SIGTERM, ss->term);
		return(FALSE);
	}
	ss->hup = signal(SIGHUP, SignalHandler);
	if (ss->hup == SIG_ERR) {
		DebugPrint("%s(%d) %s: %d, %s\n", __FILE__, __LINE__, __FUNCTION__, errno, strerror(errno));
		/* ᤹ */
		signal(SIGTERM, ss->term);
		signal(SIGHUP, ss->pipe);
		return(FALSE);
	}
	ss->alrm = signal(SIGALRM, SignalHandler);
	if (ss->alrm == SIG_ERR) {
		DebugPrint("%s(%d) %s: %d, %s\n", __FILE__, __LINE__, __FUNCTION__, errno, strerror(errno));
		/* ᤹ */
		signal(SIGTERM, ss->term);
		signal(SIGHUP, ss->pipe);
		signal(SIGHUP, ss->hup);
		return(FALSE);
	}
	return(TRUE);
}

/** 򤷤 signal 򸵤᤹
 * @param *ss ʥ
 * @retval TRUE ᤻
 * @retval FALSE ᤹˼Ԥ
 */
static bool RestoreSignal(savedSignal *ss)
{
	int	result;

	result = TRUE;

	if (signal(SIGTERM, ss->term) == SIG_ERR) {
		DebugPrint("%s(%d) %s: %d, %s\n", __FILE__, __LINE__, __FUNCTION__, errno, strerror(errno));
		result = FALSE;
	}
	if (signal(SIGPIPE, ss->pipe) == SIG_ERR) {
		DebugPrint("%s(%d) %s: %d, %s\n", __FILE__, __LINE__, __FUNCTION__, errno, strerror(errno));
		result = FALSE;
	}
	if (signal(SIGHUP, ss->hup) == SIG_ERR) {
		DebugPrint("%s(%d) %s: %d, %s\n", __FILE__, __LINE__, __FUNCTION__, errno, strerror(errno));
		result = FALSE;
	}
	if (signal(SIGALRM, ss->alrm) == SIG_ERR) {
		DebugPrint("%s(%d) %s: %d, %s\n", __FILE__, __LINE__, __FUNCTION__, errno, strerror(errno));
		result = FALSE;
	}

	return(result);
}


/** ˻֤ĹʤԤ
 * @param *waitVal Ԥ
 * @return ̵
 */
static void Wait4EcbtWrite(struct timeval *waitVal)
{
	struct timeval tmp;


	if (waitVal->tv_usec > (100 * 1000)) {
		waitVal->tv_usec += (100 * 1000);
		if (waitVal->tv_usec >= (1000 * 1000)) {
			waitVal->tv_usec = (100 * 1000) + 1;
			waitVal->tv_sec++;
		}
	} else {
		waitVal->tv_usec += THREAD_WRITE_WAIT_TIME;
		waitVal->tv_sec = 0;
	}


/*
	waitVal->tv_usec += THREAD_WRITE_WAIT_TIME;
	if (waitVal->tv_usec >= (1000 * 1000)) {
		waitVal->tv_sec++;
		waitVal->tv_usec = THREAD_WRITE_WAIT_TIME;
	}
*/

/*
	if (waitVal->tv_sec > 0) {
		waitVal->tv_usec = 0;
		waitVal->tv_sec *= 2;
		if (waitVal->tv_sec > 2) {
			waitVal->tv_sec = 2;
		}
	} else {
		waitVal->tv_usec *= 2;
		if (waitVal->tv_usec >= (1000 * 1000)) {
			waitVal->tv_sec = 1;
			waitVal->tv_usec = 0;
		}
	}
*/

	tmp.tv_sec = waitVal->tv_sec;
	tmp.tv_usec = waitVal->tv_usec;

	SafeSleep(&tmp);
}



/** ץ󥿤Ȥ̿狼ݤ䤤碌
 * @param *pBidiLibC ѹ¤ΤؤΥݥ
 * @retval TRUE ̿顼ȯƤ
 * @retval FALSE ̿顼ȯƤʤ
 */
bool IsPrinterCommunicationError(void *pBidiLibC)
{
	BidiLibC *p = (BidiLibC *)pBidiLibC;

	return(p->commError);
}

/** ץ󥿤饹ơɤ߽Ф
 * @param *cbtHdl CBT ѥϥɥ
 * @param **buf ɤ߽ФơǼݥ󥿤ؤΥݥ (ƽФ¦˶)
 * @param *rsize ɤ߽ФơĹǼݥ (ƽФ¦˶)
 * @param *status ߥơͭݤɽե饰
 * @retval ECBT_ERR_NOERR 
 * @retval ECBT_ERR_NOERRʳ 
 */
static EcbtErrCode getStatus(BidiLibC *p, void *cbtHdl, char **buf, int *rsize)
{
	int		result;
	int		count;
	int		size;
	int		offset;
//	int		retryNum;
	char	GetStatusCommand[] = { 's', 't', 0x01, 0x00, 0x01 };
	struct timeval		waitVal;

	size = sizeof(GetStatusCommand);
	offset = 0;
	waitVal.tv_sec = 0;
	waitVal.tv_usec = THREAD_WRITE_WAIT_TIME;
//	retryNum = PRINTER_STATUS_INVALID_TIME;
	while (size > 0) {
		count = 0;
		result = EcbtWrite(cbtHdl, CHID_COMMAND, GetStatusCommand + offset, size, &count);
		if ((result == ECBT_ERR_NOERR) || (result == ECBT_ERR_BUSY) || (result == ECBT_ERR_SIGNAL)) {
			offset += count;
			size -= count;
		} else {
			DebugPrint("%s(%d) %s: EcbtWrite() error = %d, signalCatchedFlag = %d\n",
					   __FILE__, __LINE__, __FUNCTION__, result, signalCatchedFlag);
			return(result);
		}
		Wait4EcbtWrite(&waitVal);
//		retryNum--;
		if (waitVal.tv_sec > PRINTER_STATUS_INVALID_TIME) {
			p->statusRead = FALSE;
		}
	}

	if (signalCatchedFlag == 0) {
		result = EcbtRead(cbtHdl, CHID_COMMAND, NULL, 511, rsize);
	}
	if ((signalCatchedFlag != 0) || (result != ECBT_ERR_NOERR)) {
		DebugPrint("%s(%d) %s: EcbtRead() error = %d, signalCatchedFlag = %d\n",
				   __FILE__, __LINE__, __FUNCTION__, result, signalCatchedFlag);
		return(result);
	}
	/* core Ǥʤ;ʬ˳ݤ */
	if ((*buf = malloc(*rsize + 64 + 1)) == NULL) {
		DebugPrint("%s(%d) %s: Not enough core\n", __FILE__, __LINE__, __FUNCTION__);
		return(ECBT_ERR_NOMEM);
	}
	memset(*buf, 0, *rsize + 64 + 1);
	memset(*buf, STATUS_RS, *rsize + 64);
	if (signalCatchedFlag == 0) {
		result = EcbtRead(cbtHdl, CHID_COMMAND, *buf, *rsize, &count);
	}
	if ((signalCatchedFlag != 0) || (result != ECBT_ERR_NOERR)) {
		DebugPrint("%s(%d) %s: EcbtRead() error = %d, signalCatchedFlag = %d\n",
				   __FILE__, __LINE__, __FUNCTION__, result, signalCatchedFlag);
		free(*buf);
		*buf = NULL;
		return(result);
	}
	DebugPrint("%s(%d) %s: *rsize = %d, count = %d\n", __FILE__, __LINE__, __FUNCTION__, *rsize, count);
	PrintHex(*buf, count);

	return(ECBT_ERR_NOERR);
}



/** ͥ write ׵ѹ¤Τ
 * @param *pBidiLibC ѹ¤ΤؤΥݥ
 * @param channel data or command ͥ
 * @return ʤ
 */
static void InitWriteRequestInfo(BidiLibC *pBidiLibC, ChID channel)
{
	writeRequestList	*list;

	if (channel == CHID_DATA) {
		list = &(pBidiLibC->dataWriteRequestList);
	} else {
		list = &(pBidiLibC->commandWriteRequestList);
	}
	list->number = 0;
	list->startIndex = 0;
	list->appendIndex = 0;
}

/** ͥ write ׵θĿ
 * @param *pBidiLibC ѹ¤ΤؤΥݥ
 * @param channel data or command ͥ
 * @return write ׵θĿ
 */
static int GetWriteRequestNumber(BidiLibC *pBidiLibC, ChID channel)
{
	writeRequestList	*list;

	if (channel == CHID_DATA) {
		list = &(pBidiLibC->dataWriteRequestList);
	} else {
		list = &(pBidiLibC->commandWriteRequestList);
	}
	return(list->number);
}

/** ͥؤ write ׵᤬ǽݤ֤
 * @param *pBidiLibC ѹ¤ΤؤΥݥ
 * @param channel data or command ͥ
 * @retval TRUE ǽ
 * @retval FALSE ǽ
 */
static bool CanWriteRequest(BidiLibC *pBidiLibC, ChID channel)
{
	if (GetWriteRequestNumber(pBidiLibC, channel) < THREAD_WRITE_LIST_MAX) {
		return(TRUE);
	}
	return(FALSE);
}

/** data  write ׵;͵뤫ݤ
 * @param *pBidiLibC ¤ΤؤΥݥ
 * @param *nWaitVal Ԥ֡0ʤԤʤ
 * @retval TRUE ;͵
 * @retval FALSE ;͵ʤ
 */
bool CanDataWriteRequest(void *pBidiLibC, struct timespec *nWaitVal)
{
	BidiLibC *p = (BidiLibC *)pBidiLibC;
	bool	result;
	int		ret;
	struct timeval	now;

	result = CanWriteRequest(p, CHID_DATA);
	if ((result == FALSE) && ((nWaitVal->tv_sec != 0) || (nWaitVal->tv_nsec != 0))) {
		pthread_mutex_lock(&(p->mutexBuffer));

		gettimeofday(&now, NULL);
		nWaitVal->tv_sec += now.tv_sec;
		nWaitVal->tv_nsec += now.tv_usec * 1000;
		ret = 0;
		while ((result == FALSE) && (ret != ETIMEDOUT)) {
			DebugPrint("%s(%d) %s: in while loop\n", __FILE__, __LINE__, __FUNCTION__);
			ret = pthread_cond_timedwait(&(p->condDataBuffer), &(p->mutexBuffer), nWaitVal);
			result = CanWriteRequest(p, CHID_DATA);
			DebugPrint("%s(%d) %s: in while loop, result = %d, ret = %d\n",
					   __FILE__, __LINE__, __FUNCTION__, result, ret);
		}
		if (ret == ETIMEDOUT) {
			/* ॢ */
			result = FALSE;
		} else {
			result = TRUE;
		}
		pthread_mutex_unlock(&(p->mutexBuffer));
	}

	return(result);
}

/** ͥ write ׵ξ֤
 * @param *pBidiLibC ѹ¤ΤؤΥݥ
 * @param channel data or command ͥ
 * @return write ׵ξؤΥݥ
 */
static writeRequestInfo *GetWriteRequestInfo(BidiLibC *pBidiLibC, ChID channel)
{
	writeRequestList	*list;

	if (channel == CHID_DATA) {
		list = &(pBidiLibC->dataWriteRequestList);
	} else {
		list = &(pBidiLibC->commandWriteRequestList);
	}
	return(&list->wrq[list->startIndex]);
}


/** ͥ ߤ write ׵
 * @param *pBidiLibC ѹ¤ΤؤΥݥ
 * @param channel data or command ͥ
 * @return  write ׵θĿ
 */
static int DeleteWriteRequestInfo(BidiLibC *pBidiLibC, ChID channel)
{
	writeRequestList	*list;
	writeRequestInfo	*wrq;

	DebugPrint("%s(%d) %s\n", __FILE__, __LINE__, __FUNCTION__);

	if (channel == CHID_DATA) {
		list = &(pBidiLibC->dataWriteRequestList);
	} else {
		list = &(pBidiLibC->commandWriteRequestList);
	}

	if (GetWriteRequestNumber(pBidiLibC, channel) == 0) {
		/* 񤭹׵᤬¸ߤʤ */
		DebugPrint("%s(%d) %s: !!! no write request !!!\n", __FILE__, __LINE__, __FUNCTION__);
		return(0);
	}

	pthread_mutex_lock(&(pBidiLibC->mutexBuffer));

	wrq = GetWriteRequestInfo(pBidiLibC, channel);

	free(wrq->buf);

	list->number--;
	list->startIndex++;
	if (list->startIndex >= THREAD_WRITE_LIST_MAX) {
		list->startIndex = 0;
	}

	pthread_cond_broadcast(&(pBidiLibC->condDataBuffer));

	pthread_mutex_unlock(&(pBidiLibC->mutexBuffer));

	return(list->number);
}

/** ͥ write ׵ɲä
 * @param *pBidiLibC ѹ¤ΤؤΥݥ
 * @param channel data or command ͥ
 * @return write ׵ξؤΥݥ
 */
static writeRequestInfo *AppendWriteRequestInfo(BidiLibC *pBidiLibC, ChID channel)
{
	writeRequestList	*list;
	writeRequestInfo	*wrq;

	DebugPrint("%s(%d) %s\n", __FILE__, __LINE__, __FUNCTION__);

	if (channel == CHID_DATA) {
		list = &(pBidiLibC->dataWriteRequestList);
	} else {
		list = &(pBidiLibC->commandWriteRequestList);
	}

	if (list->number >= THREAD_WRITE_LIST_MAX) {
		DebugPrint("%s(%d) %s: Not enough core\n", __FILE__, __LINE__, __FUNCTION__);
		return(NULL);
	}

	pthread_mutex_lock(&(pBidiLibC->mutexBuffer));

	wrq = &list->wrq[list->appendIndex];
	memset(wrq, 0, sizeof(writeRequestInfo));
	wrq->buf = NULL;

	list->number++;
	list->appendIndex++;
	if (list->appendIndex >= THREAD_WRITE_LIST_MAX) {
		list->appendIndex = 0;
	}

	if (list->number < THREAD_WRITE_LIST_MAX) {
		/* Τ롣(ݸ) */
		pthread_cond_broadcast(&(pBidiLibC->condDataBuffer));
	}

	pthread_mutex_unlock(&(pBidiLibC->mutexBuffer));

	return(wrq);
}


/** thread  ޥɥͥؤ write ꤹ
 * @param *pBidiLibC ѹ¤ΤؤΥݥ
 * @param *pBuf write ٤ǡؤΥݥ
 * @param offset buf γϰ
 * @param size 񤭹Ĺ
 * @retval TRUE 
 * @retval FALSE 
 */
static int CommandWriteRequest(void *pBidiLibC, char *pBuf, int offset, int size)
{
	BidiLibC *p = (BidiLibC *)pBidiLibC;
	writeRequestInfo	*wrq;

	DebugPrint("%s(%d) %s: pBuf = %08X, offset = %d, size = %d\n",
			   __FILE__, __LINE__, __FUNCTION__, pBuf, offset, size);

	/* Ǥ lock ݤ 롣
	 * ɬפʤ̤ǹԤʤ */
	wrq = AppendWriteRequestInfo(p, CHID_COMMAND);
	if (wrq == NULL) {
		DebugPrint("%s(%d) %s: Not enough core\n", __FILE__, __LINE__, __FUNCTION__);
		return(FALSE);
	}

	wrq->buf = pBuf;
	wrq->offset = offset;
	wrq->size = size;

	return(TRUE);
}

/** thread  ǡͥؤ write ꤹ
 * @param *pBidiLibC ѹ¤ΤؤΥݥ
 * @param *pBuf write ٤ǡؤΥݥ
 * @param offset buf γϰ
 * @param size 񤭹Ĺ
 * @retval TRUE 
 * @retval FALSE 
 */
static int DataWriteRequest(BidiLibC *pBidiLibC, char *pBuf, int offset, int size)
{
	writeRequestInfo	*wrq;

	DebugPrint("%s(%d) %s: pBuf = %08X, offset = %d, size = %d\n",
			   __FILE__, __LINE__, __FUNCTION__, pBuf, offset, size);

	wrq = AppendWriteRequestInfo(pBidiLibC, CHID_DATA);
	if (wrq == NULL) {
		return(BIDI_ERROR);
	}

	wrq->buf = pBuf;
	wrq->offset = offset;
	wrq->size = size;

	return(BIDI_OK);
}



/** thread  read Ԥʤ
 * ܥץबλޤư³
 * @param *pBidiLibC ѹ¤ΤؤΥݥ
 * @return ʤ
 */
static void bidiLibReadThread(void *pBidiLibC)
{
	int result;
	int rsize;
	char *buf;

	int	retryNum = 0;
	struct timeval	waitVal;

	BidiLibC *p = (BidiLibC *)pBidiLibC;

	DebugPrint("%s(%d) %s\n", __FILE__, __LINE__, __FUNCTION__);

	/* λʤ */
	while ( 1 ) {

		if ((signalCatchedFlag != 0) || (p->commError != FALSE)) {
			/* signal or ̿顼 ȯ */
			goto wait;
		}

		/* Status ɤ߹ */
		result = ECBT_ERR_NOERR;
		buf = NULL;
		pthread_mutex_lock(&(p->mutexIO));
		if (signalCatchedFlag == 0) {
			result = getStatus(p, p->cbtHdl, &buf, &rsize);
			if (result == ECBT_ERR_NOCONNECT) {
				/* ̿顼ȯƤ */
				DebugPrint("%s(%d) %s: printer communication error = %d\n",
						   __FILE__, __LINE__, __FUNCTION__, result);
				p->commError = TRUE;
			}
		}
		pthread_mutex_unlock(&(p->mutexIO));
		if (signalCatchedFlag != 0) {
			if (buf != NULL) {
				free(buf);
			}
			goto wait;
		}
		if (result != ECBT_ERR_NOERR) {
			if (result == ECBT_ERR_BUSY) {
				retryNum++;
				DebugPrint("%s(%d) %s: getStatus() error (busy): retryNum = %d.\n",
						   __FILE__, __LINE__, __FUNCTION__, retryNum);
				if (retryNum > READ_RETRY_NUM) {
					p->statusRead = FALSE;
					retryNum = 0;
				}
			} else {
				p->statusRead = FALSE;
				DebugPrint("%s(%d) %s: getStatus() error: don't retry.\n",
						   __FILE__, __LINE__, __FUNCTION__);
			}

		} else {

			retryNum = 0;

			/* åݤƤ顢ɤ߹ơ򥻥åȤ */
			pthread_mutex_lock(&(p->mutexStatus));
			if (signalCatchedFlag == 0) {
				result = setStatus(p->printerStatus, buf, rsize);
			}
			pthread_mutex_unlock(&(p->mutexStatus));
			if (buf != NULL) {
				free(buf);
			}
			if (signalCatchedFlag != 0) {
				goto wait;
			}

			p->statusRead = TRUE;
		}

		if ((signalCatchedFlag != 0) || (p->commError != FALSE)) {
			/* signal or ̿顼 ȯ */
			goto wait;
		}

		/* data channel  open/close  */
		result = ECBT_ERR_NOERR;
		if ((p->dataChannelOpend == FALSE) && (p->dataChannelOpenRequest != FALSE)) {
			if (isIdling(p->printerStatus) != FALSE) {
				/* DATAͥ open  */
				pthread_mutex_lock(&(p->mutexIO));
				if (signalCatchedFlag == 0) {
					result = EcbtChannelOpen(p->cbtHdl, CHID_DATA);
					if (result == ECBT_ERR_NOCONNECT) {
						/* ̿顼ȯƤ */
						DebugPrint("%s(%d) %s: printer communication error = %d\n",
								   __FILE__, __LINE__, __FUNCTION__, result);
						p->commError = TRUE;
					}
				}
				pthread_mutex_unlock(&(p->mutexIO));
				if (signalCatchedFlag != 0) {
					goto wait;
				}
				if ((result == ECBT_ERR_NOERR) || (result == ECBT_ERR_CHREOPN)) {
					p->dataChannelOpend = TRUE;
				}
			}
		} else if ((p->dataChannelOpend != FALSE) && (p->dataChannelOpenRequest == FALSE)) {
			/* DATAͥ close  */
			pthread_mutex_lock(&(p->mutexIO));
			if (signalCatchedFlag == 0) {
				result = EcbtChannelClose(p->cbtHdl, CHID_DATA);
				if (result == ECBT_ERR_NOCONNECT) {
					/* ̿顼ȯƤ */
					DebugPrint("%s(%d) %s: printer communication error = %d\n",
							   __FILE__, __LINE__, __FUNCTION__, result);
					p->commError = TRUE;
				}
			}
			pthread_mutex_unlock(&(p->mutexIO));
			if (signalCatchedFlag != 0) {
				goto wait;
			}
			if ((result == ECBT_ERR_NOERR) || (result == ECBT_ERR_CHNOPN)) {
				p->dataChannelOpend = FALSE;
			}
		}

wait:
		/* Ԥ */
		waitVal.tv_sec = THREAD_READ_WAIT_TIME;
		waitVal.tv_usec = 0;
		result = SafeSleep(&waitVal);
		if ((result < 0) && (errno != EINTR)) {
			DebugPrint("%s(%d) %s: SafeSleep() error.\n", __FILE__, __LINE__, __FUNCTION__);
		}
	}
}

/** thread  write Ԥʤ
 * ܥץबλޤư³
 * @param *pBidiLibC ѹ¤ΤؤΥݥ
 * @return ʤ
 */
static void bidiLibWriteThread(void *pBidiLibC)
{
	int		result;
	int		count;
	char	rBuf[512];							/* ɤ߼ΤѥХåե */
//	int		rsize;
	writeRequestInfo	*wrq;
	struct timeval		waitVal;
	BidiLibC	*p = (BidiLibC *)pBidiLibC;

	DebugPrint("%s(%d) %s\n", __FILE__, __LINE__, __FUNCTION__);

	/* λʤ */
	while ( 1 ) {

		if ((signalCatchedFlag != 0) || (p->commError != FALSE)) {
			/* signal or ̿顼 ȯ */
			goto wait;
		}

		/* write flag Υå */
		if ((GetWriteRequestNumber(p, CHID_COMMAND) == 0) &&
			(GetWriteRequestNumber(p, CHID_DATA) == 0)) {
			/* interrupted or no write request */
			count = 1;
			goto wait;
		}

		if (GetWriteRequestNumber(p, CHID_COMMAND) != 0) {
			/* command 񤭹 */

			wrq = GetWriteRequestInfo(p, CHID_COMMAND);

			pthread_mutex_lock(&(p->mutexIO));

			if (signalCatchedFlag == 0) {

				waitVal.tv_sec = 0;
				waitVal.tv_usec = THREAD_WRITE_WAIT_TIME;
				while (wrq->size > 0) {
					count = 0;
					result = ECBT_ERR_NOERR;
					result = EcbtWrite(p->cbtHdl, CHID_COMMAND, wrq->buf + wrq->offset, wrq->size, &count);
					if ((result == ECBT_ERR_NOERR) || (result == ECBT_ERR_BUSY) || (result == ECBT_ERR_SIGNAL)) {
						wrq->offset += count;
						wrq->size -= count;
						if (wrq->size == 0) {
							/* end of write data */
							DeleteWriteRequestInfo(p, CHID_COMMAND);

							/* Ԥɤ߽Ф
							 *  reset ˤΤбʤΤ ɤ߼ΤƤ롣
							 * bidiCtrl бϡǡΰ˹פɬ */
							result = EcbtRead(p->cbtHdl, CHID_COMMAND, rBuf, sizeof(rBuf), &count);
							if (result == ECBT_ERR_NOCONNECT) {
								/* ̿顼ȯƤ */
								DebugPrint("%s(%d) %s: printer communication error = %d\n",
										   __FILE__, __LINE__, __FUNCTION__, result);
								p->commError = TRUE;
							}
							continue;
						}
					} else if (result == ECBT_ERR_NOCONNECT) {
						/* ̿顼ȯƤ */
						DebugPrint("%s(%d) %s: printer communication error = %d\n",
								   __FILE__, __LINE__, __FUNCTION__, result);
						p->commError = TRUE;
					} else {
						DebugPrint("%s(%d) %s: EcbtWrite() error = %d, signalCatchedFlag = %d\n",
								   __FILE__, __LINE__, __FUNCTION__, result, signalCatchedFlag);
					}
					Wait4EcbtWrite(&waitVal);
					if (waitVal.tv_sec > PRINTER_STATUS_INVALID_TIME) {
						p->statusRead = FALSE;
					}
				}
			}
			pthread_mutex_unlock(&(p->mutexIO));
			count = 1;							/* Ԥʤ褦 */
		}

		if ((signalCatchedFlag != 0) || (p->commError != FALSE)) {
			/* signal or ̿顼 ȯ */
			goto wait;
		}

		if (GetWriteRequestNumber(p, CHID_DATA) != 0) {
			/* data 񤭹 */
			count = 0;

			pthread_mutex_lock(&(p->mutexIO));

			wrq = GetWriteRequestInfo(p, CHID_DATA);

			if (signalCatchedFlag == 0) {
				result = EcbtWrite(p->cbtHdl, CHID_DATA, wrq->buf + wrq->offset, wrq->size, &count);
				if ((result == ECBT_ERR_NOERR) || (result == ECBT_ERR_BUSY) || (result == ECBT_ERR_SIGNAL)) {
					wrq->offset += count;
					wrq->size -= count;
					if (wrq->size == 0) {
						/* end of write data */
						DeleteWriteRequestInfo(p, CHID_DATA);
					}
				} else if (result == ECBT_ERR_NOCONNECT) {
					/* ̿顼ȯƤ */
					DebugPrint("%s(%d) %s: printer communication error = %d\n",
							   __FILE__, __LINE__, __FUNCTION__, result);
					p->commError = TRUE;
				} else {
					DebugPrint("%s(%d) %s: EcbtWrite() error = %d, signalCatchedFlag = %d\n",
							   __FILE__, __LINE__, __FUNCTION__, result, signalCatchedFlag);
					count = 0;
				}
			}
			pthread_mutex_unlock(&(p->mutexIO));
		}


wait:
		if ((signalCatchedFlag == 0) && (p->commError == FALSE) &&
			((GetWriteRequestNumber(p, CHID_COMMAND) != 0) ||
			 (GetWriteRequestNumber(p, CHID_DATA) != 0))) {
			waitVal.tv_sec = 0;
			waitVal.tv_usec = THREAD_WRITE_WAIT_TIME;
			continue;
		}

		if (count == 0) {
			Wait4EcbtWrite(&waitVal);
		} else {
			waitVal.tv_sec = 0;
			waitVal.tv_usec = THREAD_WRITE_WAIT_TIME;
		}
	}
}


/** printer  reset ޥɤ
 * @param *pBidiLibC ѹ¤ΤؤΥݥ
 * @retval TRUE 
 * @retval FALSE 
 */
static bool PrinterReset(void *pBidiLibC)
{
	int		result;
	char	*buf;
	static char	cancelData[] = {
		'r', 's', 0x01, 0x00, 0x01,				/* ץ󥿤Υꥻåȥޥ */
	};

	if ((buf = malloc(sizeof(cancelData))) == NULL) {
		DebugPrint("%s(%d) %s: Not enough core\n", __FILE__, __LINE__, __FUNCTION__);
		return(FALSE);
	}

	memcpy(buf, cancelData, sizeof(cancelData));

	result = CommandWriteRequest(pBidiLibC, buf, 0, sizeof(cancelData));
	if (result == FALSE) {
		free(buf);
	}

	return(result);
}


/** bidiNew ΥץѼ
 * create a new object
 * @param fdRead file descriptor for read
 * @param fdWrite file descriptor for write
 * @param pURI URI of device
 * @retval  pointer of a new object
 * @retval  NULL
 */
void *bidiLibNew(int fdRead, int fdWrite, char *pURI)
{
	BidiLibC	*p;
	char		*uri;
	int			result;
	savedSignal	ss;

	DebugPrint("%s(%d) %s: fdRead = %d, pURI = \"%s\"\n", __FILE__, __LINE__, __FUNCTION__, fdRead, pURI);

	/* EPSON ξϡ fdRead  fdWrite ʤФʤʤ */
	if (fdRead != fdWrite) {
		DebugPrint("%s(%d) %s: fdRead = %d, fdWrite = %d\n",
				   __FILE__, __LINE__, __FUNCTION__, fdRead, fdWrite);
		return(NULL);
	}

	// BidiLibC ΰƤ
	p = (BidiLibC *)malloc(sizeof(BidiLibC));
	if (p == NULL) {
		DebugPrint("%s(%d) %s: %d, %s\n", __FILE__, __LINE__, __FUNCTION__, errno, strerror(errno));
		return(NULL);
	}
	memset(p, 0, sizeof(BidiLibC));

	// ץ󥿤ɤ߽Фơΰν
	p->printerStatus = initStatus();
	if (p->printerStatus == NULL) {
		DebugPrint("%s(%d) %s: %d, %s\n", __FILE__, __LINE__, __FUNCTION__, errno, strerror(errno));
		free(p);
		return(NULL);
	}

	/* signal  */
	signalCatchedFlag = 0;
	result = SaveSignal(&ss);
	if (result == FALSE) {
		free(p);
		return(NULL);
	}


	// ǥХȤ̿򳫻Ϥ
	result = EcbtOpen(&(p->cbtHdl), fdRead);
	if ((signalCatchedFlag != 0) || (result != ECBT_ERR_NOERR)) {
		free(p->printerStatus);
		free(p);
		RestoreSignal(&ss);
		return(NULL);
	}

	/* COMMANDͥ򥪡ץ󤹤 */
	result = EcbtChannelOpen(p->cbtHdl, CHID_COMMAND);
	if ((signalCatchedFlag != 0) || (result != ECBT_ERR_NOERR)) {
		free(p->printerStatus);
		free(p);
		RestoreSignal(&ss);
		return(NULL);
	}

	// pURI Ƥ򥳥ԡ
	if (pURI != NULL) {
		uri = strdup(pURI);
		if (uri == NULL) {
			DebugPrint("%s(%d) %s: %d, %s\n", __FILE__, __LINE__, __FUNCTION__, errno, strerror(errno));
			free(p->printerStatus);
			free(p);
			RestoreSignal(&ss);
			return(NULL);
		}
	} else {
		uri = NULL;
	}

	// BidiLibC 
	p->fdRead = fdRead;
	p->fdWrite = fdWrite;
	p->uri = uri;
	p->isPrinterUp = 0;
	p->idJob = 0;
	p->xmlSize = 0;
	p->pXmlBuf = 0;

	InitWriteRequestInfo(p, CHID_DATA);
	InitWriteRequestInfo(p, CHID_COMMAND);

	p->dataChannelOpend = FALSE;
	pthread_mutex_init(&(p->mutexStatus), NULL);
	pthread_mutex_init(&(p->mutexIO), NULL);
	pthread_mutex_init(&(p->mutexBuffer), NULL);
	pthread_cond_init(&(p->condDataBuffer), NULL);


	/* EPSON ǤǤ read/write ƱǹԤʤ碌뤿 pthread Ȥ */
	// read  thread κ
	result = pthread_create(&(p->RdThread), NULL, (void *)bidiLibReadThread, (void *)p);
	if ((signalCatchedFlag != 0) || (result != 0)) {
		DebugPrint("%s(%d) %s: read pthread_create() error. %d, %s\n",
				   __FILE__, __LINE__, __FUNCTION__, errno, strerror(errno));
		free(p->printerStatus);
		free(p);
		p = NULL;
		RestoreSignal(&ss);
		return(NULL);
	}
	p->RdThreadValid = TRUE;

	// write  thread κ
	result = pthread_create(&(p->WrThread), NULL, (void *)bidiLibWriteThread, (void *)p);
	if ((signalCatchedFlag != 0) || (result != 0)) {
		DebugPrint("%s(%d) %s: write pthread_create() error. %d, %s\n",
				   __FILE__, __LINE__, __FUNCTION__, errno, strerror(errno));

		result = pthread_cancel(p->RdThread);	/* read thread νλ */

		free(p->printerStatus);
		free(p);
		p = NULL;
		return(NULL);
	}
	p->WrThreadValid = TRUE;


	RestoreSignal(&ss);
	return(p);
}


/** bidiDestroy ΥץѼ */
void bidiLibDestroy(void *pBidiLibC)
{
	BidiLibC *p = (BidiLibC *)pBidiLibC;
	int					result;
	struct timeval		timeout;
	savedSignal			ss;

	DebugPrint("%s(%d) %s\n", __FILE__, __LINE__, __FUNCTION__);

	/* signal  */
	result = SaveSignal(&ss);
	if (result == FALSE) {
		/* 顼ȯƤ̵뤷Ƽ¹ԤɤΤʡ */
//		return;
	}

	/* read  thread νλ顼̵ */
	result = -1;
	while ((p->RdThreadValid == TRUE) && (signalCatchedFlag == 0) && (result != 0)) {
		result = pthread_cancel(p->RdThread);
		if (result != 0) {
			DebugPrint("%s(%d) %s: error : %s (%d)\n", __FILE__, __LINE__, __FUNCTION__, strerror(errno), errno);
			timeout.tv_sec = 0;
			timeout.tv_usec = 10 * 1000;
			result = SafeSleep(&timeout);
			result = -1;
		}
	}
	result = -1;
	while ((p->WrThreadValid == TRUE) && (signalCatchedFlag == 0) && (result != 0)) {
		result = pthread_cancel(p->WrThread);
		if (result != 0) {
			DebugPrint("%s(%d) %s: error : %s (%d)\n", __FILE__, __LINE__, __FUNCTION__, strerror(errno), errno);
			timeout.tv_sec = 0;
			timeout.tv_usec = 10 * 1000;
			result = SafeSleep(&timeout);
			result = -1;
		}
	}

	// open Ƥơclose Ƥ
	// 顼̵
	if (signalCatchedFlag == 0) {
		/* ץ󥿤 reset ʤ */
		if (p->commError == FALSE) {
			DebugPrint("%s(%d) %s: prev EcbtClose()\n", __FILE__, __LINE__, __FUNCTION__);
			/* ľʤΤǡǤϥ顼̵뤹 */
			EcbtClose(p->cbtHdl);
			DebugPrint("%s(%d) %s: after EcbtClose()\n", __FILE__, __LINE__, __FUNCTION__);
		}
	}

	// Ƥΰγ
	do {
		result = DeleteWriteRequestInfo(p, CHID_DATA);
	} while (result != 0);
	do {
		result = DeleteWriteRequestInfo(p, CHID_COMMAND);
	} while (result != 0);

	if (p->uri != NULL) {
		free(p->uri);
	}
	free(p->printerStatus);

	free(pBidiLibC);

	RestoreSignal(&ss);
}

/** bidiGetCap ΥץѼ */
int bidiLibGetCap(void *pBidiLibC, BidiCap cap)
{
	BidiLibC *p = (BidiLibC *)pBidiLibC;

	DebugPrint("%s(%d) %s: cap = 0x%0x\n", __FILE__, __LINE__, __FUNCTION__, cap);

	if (p->commError != FALSE) {
		/* ̿顼ȯƤΤ̵ */
		DebugPrint("\t p->commError = %d\n", p->commError);
		return(BIDI_ERROR);
	}

	switch (cap) {
		case BIDI_CAP_WRITE:
			return BIDI_TRUE;
		case BIDI_CAP_JOB:
			return BIDI_TRUE;
		case BIDI_CAP_CTRL:
			return BIDI_FALSE;					/* ϡʤ */
//			return BIDI_TRUE;
		default:
			DebugPrint("%s(%d) %s: %s\n", __FILE__, __LINE__, __FUNCTION__, "illegal cap case.");
			return BIDI_ERROR;
	}
}

/** bidiStartJob ΥץѼ */
int bidiLibStartJob(void *pBidiLibC, int idJob)
{
	BidiLibC *p = (BidiLibC *)pBidiLibC;

	DebugPrint("%s(%d) %s: idJob = %d, p->dataChannelOpend = %d\n", __FILE__, __LINE__, __FUNCTION__, idJob, p->dataChannelOpend);

	if (p->commError != FALSE) {
		/* ̿顼ȯƤΤ̵ */
		DebugPrint("\t p->commError = %d\n", p->commError);
		return(BIDI_ERROR);
	}

	p->isPrinterUp = 0;
	p->idJob = idJob;
	p->dataChannelOpenRequest = TRUE;

	return(BIDI_OK);
}

/** bidiEndJob ΥץѼ */
int bidiLibEndJob(void *pBidiLibC)
{
	int stat;
	BidiLibC *p = (BidiLibC *)pBidiLibC;

	DebugPrint("%s(%d) %s\n", __FILE__, __LINE__, __FUNCTION__);

	if (p->commError != FALSE) {
		/* ̿顼ȯƤΤ̵ */
		DebugPrint("\t p->commError = %d\n", p->commError);
		return(BIDI_ERROR);
	}

	// ɥ֤ä isPrinterUp  FALSE ˤ
	stat = isPrinting(p->printerStatus);
	if ((stat == 0) && (p->isPrinterUp != 0)) {
		p->isPrinterUp = 0;
	}

	DebugPrint("%s(%d) %s: stat = %d, p->isPrinterUp = %d\n", __FILE__, __LINE__, __FUNCTION__, stat, p->isPrinterUp);

	if ((stat != 0) || (p->isPrinterUp != 0)) {
		return(BIDI_EPROGRESS);
	}

	p->dataChannelOpenRequest = FALSE;

	return(BIDI_OK);
}

/** bidiCancelJob ΥץѼ */
int bidiLibCancelJob(void *pBidiLibC, int idJob)
{
	BidiLibC *p = (BidiLibC *)pBidiLibC;
	int		result;

	DebugPrint("%s(%d) %s: idJob = %d\n", __FILE__, __LINE__, __FUNCTION__, idJob);

	if (p->commError != FALSE) {
		/* ̿顼ȯƤΤ̵ */
		DebugPrint("\t p->commError = %d\n", p->commError);
		return(BIDI_ERROR);
	}

	if (p->idJob != idJob) {
		return(BIDI_ENOJOB);
	}

	if (pthread_mutex_trylock(&(p->mutexIO)) == EBUSY) {
		return(BIDI_EPROGRESS);
	}

	do {
		result = DeleteWriteRequestInfo(p, CHID_DATA);
	} while (result != 0);
	do {
		result = DeleteWriteRequestInfo(p, CHID_COMMAND);
	} while (result != 0);

	/* ץ󥿤˥ꥻåȤ */
	if (PrinterReset(p) == FALSE) {
		return(BIDI_EPROGRESS);
	}

	p->dataChannelOpenRequest = FALSE;

	pthread_mutex_unlock(&(p->mutexIO));

	return(BIDI_OK);
}

/** bidiGetReadFD ΥץѼ */
int bidiLibGetReadFD(void *pBidiLibC)
{
	BidiLibC *p = (BidiLibC *)pBidiLibC;

	DebugPrint("%s(%d) %s\n", __FILE__, __LINE__, __FUNCTION__);

	if (p->commError != FALSE) {
		/* ̿顼ȯƤΤ̵ */
		DebugPrint("\t p->commError = %d\n", p->commError);
		return(BIDI_ERROR);
	}

	return(p->fdRead);
}

/** bidiStartRead ΥץѼ */
int bidiLibStartRead(void *pBidiLibC, BidiReadMode idReadMode, char *pLang)
{
	int result;
	int	retryNum;
	struct timeval	waitVal;
	BidiLibC *p = (BidiLibC *)pBidiLibC;
	savedSignal	ss;

	DebugPrint("%s(%d) %s: idReadMode = %08X, pLang = \"%s\"\n", __FILE__, __LINE__, __FUNCTION__, idReadMode, pLang);

	if (p->commError != FALSE) {
		/* ̿顼ȯƤΤ̵ */
		DebugPrint("\t p->commError = %d\n", p->commError);
		return(BIDI_ERROR);
	}

	// Read ModeΥå
	if (idReadMode != BIDI_READ_PRT_MIB_SUMMARY) {
		return(BIDI_ERROR);
	}

	/* signal  */
//	signalCatchedFlag = 0;
	result = SaveSignal(&ss);
	if (result == FALSE) {
		return(BIDI_ERROR);
	}

	retryNum = 2;
	while ((signalCatchedFlag == 0) && ((p->statusRead == FALSE) && (retryNum > 0))) {
		waitVal.tv_sec = THREAD_READ_WAIT_TIME;
		waitVal.tv_usec = 0;
		result = SafeSleep(&waitVal);
		if ((result < 0) && (errno != EINTR)) {
			/* 顼Ȥ */
			DebugPrint("%s(%d) %s: SafeSleep() error.\n", __FILE__, __LINE__, __FUNCTION__);
			RestoreSignal(&ss);
			return(BIDI_ERROR);
		}
		retryNum--;
	}
	if (signalCatchedFlag != 0) {
		RestoreSignal(&ss);
		return(BIDI_EINTR);
	}
	if (p->statusRead == FALSE) {
		/* ԤäƤ⡢ԤäƤ⡢ */
		RestoreSignal(&ss);
		return(BIDI_ERROR);
	}

	/* ɤ߹ߺѤ */

	pthread_mutex_lock(&(p->mutexStatus));
	p->xmlBuf[0] = '\0';
	result = makeXml(p->printerStatus, p->xmlBuf, XML_BUFFER_SIZE, pLang);
	pthread_mutex_unlock(&(p->mutexStatus));

	p->xmlSize = strlen(p->xmlBuf);
	p->pXmlBuf = 0;

	RestoreSignal(&ss);
	return(BIDI_OK);
}

/** bidiRead ΥץѼ */
int bidiLibRead(void *pBidiLibC, void *pBuf, int nBufBytes)
{
	BidiLibC *p = (BidiLibC *) pBidiLibC;
	int i;
	int left, size;
	int	result;
	savedSignal	ss;

	DebugPrint("%s(%d) %s: nBufBytes = %d\n", __FILE__, __LINE__, __FUNCTION__, nBufBytes);

	if (p->commError != FALSE) {
		/* ̿顼ȯƤΤ̵ */
		DebugPrint("\t p->commError = %d\n", p->commError);
		return(BIDI_ERROR);
	}

	/* signal  */
	result = SaveSignal(&ss);
	if (result == FALSE) {
		return(BIDI_ERROR);
	}

	left = p->xmlSize - p->pXmlBuf;
	size = left > nBufBytes ? nBufBytes : left;
//	memcpy(pBuf, p->xmlBuf + p->pXmlBuf, size);
	for (i = 0; (i < size) && (signalCatchedFlag == 0); i++) {
		*((char *)pBuf + i) = p->xmlBuf[p->pXmlBuf + i];
	}
	p->pXmlBuf += i;

	RestoreSignal(&ss);
	return(i);
}

/** bidiEndRead ΥץѼ */
int bidiLibEndRead(void *pBidiLibC)
{
	BidiLibC *p = (BidiLibC *) pBidiLibC;

	DebugPrint("%s(%d) %s\n", __FILE__, __LINE__, __FUNCTION__);

	if (p->commError != FALSE) {
		/* ̿顼ȯƤΤ̵ */
		DebugPrint("\t p->commError = %d\n", p->commError);
		return(BIDI_ERROR);
	}

	p->xmlSize = 0;
	p->pXmlBuf = 0;
	return(BIDI_OK);
}

/** bidiGetWriteFD ΥץѼ */
int bidiLibGetWriteFD(void *pBidiLibC)
{
	BidiLibC *p = (BidiLibC *) pBidiLibC;

	DebugPrint("%s(%d) %s\n", __FILE__, __LINE__, __FUNCTION__);

	if (p->commError != FALSE) {
		/* ̿顼ȯƤΤ̵ */
		DebugPrint("\t p->commError = %d\n", p->commError);
		return(BIDI_ERROR);
	}

	return(p->fdWrite);
}

/* ---------------------------------------------------------------- */

/* EPSON Ǥɬ */

/** bidiStartWrite ΥץѼ */
int bidiLibStartWrite(void *pBidiLibC)
{
	BidiLibC *p = (BidiLibC *) pBidiLibC;
	struct timeval	waitVal;
	int	retryNum;
	int	result;
	savedSignal	ss;

	DebugPrint("%s(%d) %s\n", __FILE__, __LINE__, __FUNCTION__);

	if (p->commError != FALSE) {
		/* ̿顼ȯƤΤ̵ */
		DebugPrint("\t p->commError = %d\n", p->commError);
		return(BIDI_ERROR);
	}

	/* signal  */
	result = SaveSignal(&ss);
	if (result == FALSE) {
		return(BIDI_ERROR);
	}

	retryNum = 10;
	while ((signalCatchedFlag == 0) && ((p->dataChannelOpend == FALSE) && (retryNum > 0))) {

		DebugPrint("%s(%d) %s: wait in while loop\n", __FILE__, __LINE__, __FUNCTION__);

		/* Ԥ */
		waitVal.tv_sec = THREAD_READ_WAIT_TIME;
		waitVal.tv_usec = 0;
		result = SafeSleep(&waitVal);
		if ((result < 0) && (errno != EINTR)) {
			DebugPrint("%s(%d) %s: SafeSleep() error.\n", __FILE__, __LINE__, __FUNCTION__);
		}
		retryNum--;
	}
	if (signalCatchedFlag != 0) {
		RestoreSignal(&ss);
		return(BIDI_EINTR);
	}
	if (p->dataChannelOpend == FALSE) {
		/* ǡͥ뤬ץ󤵤ƤʤΤǡ顼Ȥ */
		RestoreSignal(&ss);
		return(BIDI_ERROR);
	}

	p->isPrinterUp = 1;

	RestoreSignal(&ss);
	return(BIDI_OK);
}

/** bidiWrite ΥץѼ */
int bidiLibWrite(void *pBidiLibC, void *pBuf, int nBufBytes)
{
	BidiLibC *p = (BidiLibC *) pBidiLibC;
	int length;
	int result;
	savedSignal	ss;

	DebugPrint("%s(%d) %s\n", __FILE__, __LINE__, __FUNCTION__);

	if (p->commError != FALSE) {
		/* ̿顼ȯƤΤ̵ */
		DebugPrint("\t p->commError = %d\n", p->commError);
		return(BIDI_ERROR);
	}

	/* signal  */
	result = SaveSignal(&ss);
	if (result == FALSE) {
		return(BIDI_ERROR);
	}

	length = 0;
	result = CanWriteRequest(pBidiLibC, CHID_DATA);
	if (result != FALSE) {
		result = DataWriteRequest(p, pBuf, 0, nBufBytes);
		/* Ǥϥ顼ȯʤ */
		length = nBufBytes;
	}

	RestoreSignal(&ss);
	return(length);
}

/** bidiEndWrite ΥץѼ */
int bidiLibEndWrite(void *pBidiLibC)
{
	BidiLibC *p = (BidiLibC *) pBidiLibC;

	DebugPrint("%s(%d) %s\n", __FILE__, __LINE__, __FUNCTION__);

	if (p->commError != FALSE) {
		/* ̿顼ȯƤΤ̵ */
		DebugPrint("\t p->commError = %d\n", p->commError);
		return(BIDI_ERROR);
	}

	p->isPrinterUp = 0;

	return(BIDI_OK);
}

/** bidiCtrl ΥץѼ
 * Ȥ
 */
int bidiLibCtrl(void *pBidiLibC, int idRequest, void *pData, int nDataBytes)
{
	BidiLibC *p = (BidiLibC *) pBidiLibC;

	DebugPrint("%s(%d) %s\n", __FILE__, __LINE__, __FUNCTION__);

	if (p->commError != FALSE) {
		/* ̿顼ȯƤΤ̵ */
		DebugPrint("\t p->commError = %d\n", p->commError);
		return(BIDI_ERROR);
	}

	return(BIDI_ERROR);

	return(BIDI_OK);
}

/* end of bidiPXV700.c */
