/* bidiProc.c -- Bidi Process Type.
 * Copyright (C) 2003 EPSON KOWA Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: bidiProc.c,v 1.20 2004/02/17 18:04:09 toko Exp $ */

/** \mainpage
 *
 * ϡbidi Process ѥեǤ롣
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <fcntl.h>
#include <string.h>
#include <sys/select.h>
#include <signal.h>
#include <setjmp.h>
#include <errno.h>
#include <netinet/in.h>

#include <pthread.h>

#if defined(DEBUG)
#include <stdarg.h>
#endif /* DEBUG */

#include "bidiCommon.h"
#include "bidiProc.h"

#include "bidiLibEpson.h"



/** Process ʬδѹ¤ */
typedef struct BidiProcC_t {
	int	dataWriteFD;		 		/**< data write ѥեǥץ */
	int	dataReadFD;					/**< data read ѥեǥץ */
	int	cmdWriteFD;					/**< command write ѥեǥץ */
	int	cmdReadFD;					/**< command read ѥեǥץ */
	int	outputFD;					/**< ץ󥿤ؤν񤭹ѥեǥץ */
	int	inputFD;					/**< ץ󥿤ɤ߽Фѥեǥץ */

	char	*uri;					/**< uri  */

	int	cmdReadBuf[64];				/**< caller (Plugin) μѥХåե */

	int	idJob;						/**< ߤ job id */

	void *pLibC;					/**< Library ʬδѹ¤ΤؤΥݥ */

	BidiCxt	context;				/**< current context */
	BidiCxt	saveContext;			/**< previous context */

} BidiProcC;


extern int signalCatchedFlag;		/**< signal ­  signal ID  */



/** Debugѥåϴؿ
 * @param printf ؿƱΰȤ
 * @return ̵
 */
static void DebugPrint(const char *fmt, ...)
{
#ifdef DEBUG
	va_list ap;

	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	fflush(stderr);
#endif
	return;
}

/* ---------------------------------------------------------------- */

/** wait Ѵؿ
 * sleep ؿδ select ؿѤ
 * @param *timeout Ԥ
 * @retval 0 
 * @retval -1 顼ȯ
 */
static int SafeSleep(struct timeval *timeout)
{
	int	result;

	result = select(0, NULL, NULL, NULL, timeout);

	return(result);
}


/** ǡ FD å
 * TBD #### ΥɤǤưʤ
 * ΤLinux ʤΤ
 * */
#if 0
static bool SetLock(int fd)
{
	int				result;
	struct flock	lock;

	memset(&lock, 0, sizeof(struct flock));
	lock.l_type = F_RDLCK;
	result = fcntl(fd, F_SETLK, &lock);
	if (result != 0) {
		/* error */
		DebugPrint("%s(%d) %s: result = %d: %d: \"%s\"\n",
				   __FILE__, __LINE__, __FUNCTION__, result, errno, strerror(errno));
		return(FALSE);
	}
	DebugPrint("%s(%d) %s: lock succeeded\n",
			   __FILE__, __LINE__, __FUNCTION__);

	return(TRUE);
}

/** ǡ FD Υå
 * TBD #### SetLock() ưʤа̣ʤ
 * */
static bool UnsetLock(int fd)
{
	int				result;
	struct flock	lock;

	memset(&lock, 0, sizeof(struct flock));
	lock.l_type = F_UNLCK;
	result = fcntl(fd, F_SETLK, &lock);
	if (result != 0) {
		/* error */
		DebugPrint("%s(%d) %s: result = %d: %d: \"%s\"\n",
				   __FILE__, __LINE__, __FUNCTION__, result, errno, strerror(errno));
		return(FALSE);
	}
	DebugPrint("%s(%d) %s: unlock succeeded\n",
			   __FILE__, __LINE__, __FUNCTION__);

	return(TRUE);
}
#endif


/** "--data-write-fd" ν */
static bool setDataWriteFD(BidiProcC *pBidiProcC, char *str)
{
	pBidiProcC->dataWriteFD = atol(str);
	return(TRUE);
}


/** "--data-read-fd" ν */
static bool setDataReadFD(BidiProcC *pBidiProcC, char *str)
{
	pBidiProcC->dataReadFD = atol(str);
	return(TRUE);
}

/** "--cmd-write-fd" ν */
static bool setCmdWriteFD(BidiProcC *pBidiProcC, char *str)
{
	pBidiProcC->cmdWriteFD = atol(str);
	return(TRUE);
}

/** "--cmd-read-fd" ν */
static bool setCmdReadFD(BidiProcC *pBidiProcC, char *str)
{
	pBidiProcC->cmdReadFD = atol(str);
	return(TRUE);
}

/** "--output-fd" ν */
static bool setOutputFD(BidiProcC *pBidiProcC, char *str)
{
	pBidiProcC->outputFD = atol(str);
	return(TRUE);
}

/** "--input-fd" ν */
static bool setInputFD(BidiProcC *pBidiProcC, char *str)
{
	pBidiProcC->inputFD = atol(str);
	return(TRUE);
}

/** "--printer-uri" ν */
static bool setUri(BidiProcC *pBidiProcC, char *str)
{
	if ((pBidiProcC->uri = malloc(strlen(str) + 1)) == NULL) {
		DebugPrint("%s(%d) %s: not enough core for \"%s\"\n",
				   __FILE__, __LINE__, __FUNCTION__, str);
		return(FALSE);
	}
	strcpy(pBidiProcC->uri, str);
	return(TRUE);
}


/** ޥɥ饤Ѥ */
struct cmdOption_t {
	bool	valid;								/**< ͭ (ꤵ줿) ݤ */
	char	*name;								/**< ץʸ */
	bool	(*set)(BidiProcC *st, char *str);	/**< ؿ */
} cmdOption[] = {
	{ FALSE, "--data-write-fd", setDataWriteFD },
	{ FALSE, "--data-read-fd", setDataReadFD },
	{ FALSE, "--cmd-write-fd", setCmdWriteFD },
	{ FALSE, "--cmd-read-fd", setCmdReadFD },
	{ FALSE, "--output-fd", setOutputFD },
	{ FALSE, "--input-fd", setInputFD },
	{ FALSE, "--printer-uri", setUri },
};


/** ʸñĹ (Хȿ) ֤
 * @param *str оʸ
 * @param *sep ѥ졼
 * @param *buf ̤
 * @param size ̤
 * @retval  ñĹ
 * @retval -1 顼 (ȯʤȦ)
 */
static int getWord(char *str, char *sep, char *buf, int size)
{
	int i;
	int j;

	for (i = 0; str[i] != '\0'; i++) {
		for (j = 0; sep[j] != '\0'; j++) {
			if (str[i] == sep[j]) {
				return(i);
			}
		}
	}
	return(-1);
}


/** ޥɥ饤Υå
 * @param argc ޥɥ饤θĿ
 * @param *argv[] ޥɥ饤
 * @param *pBidiProcC ѹ¤ΤؤΥݥ
 * @retval TRUE ä
 * @retval FALSE ʰä
 */
static bool argCheck(int argc, char *argv[], BidiProcC *pBidiProcC)
{
	int	i, j, k;
	int	result;
	int	index;
	int	len;
	char	*p;
	char	tmpBuf[64];
	char	*sep = " =";

	DebugPrint("%s(%d) %s\n", __FILE__, __LINE__, __FUNCTION__);

	i = 1;
	while (i < argc) {

		len = getWord(argv[i], sep, tmpBuf, sizeof(tmpBuf));
		if (len < 0) {
			if ((i + 1) >= argc) {
				DebugPrint("%s(%d) %s: too few argment: \"%s\"\n",
						   __FILE__, __LINE__, __FUNCTION__, argv[i]);
				return(FALSE);
			}
			len = strlen(argv[i]);
			index = i + 1;
			p = argv[i + 1];
		} else {
			index = i;
			p = argv[i] + len;
		}
		if (len >= sizeof(tmpBuf)) {
			DebugPrint("%s(%d) %s: too long: \"%s\"\n",
					   __FILE__, __LINE__, __FUNCTION__, argv[i]);
			return(FALSE);
		}
		strncpy(tmpBuf, argv[i], len);
		tmpBuf[len] = '\0';

		for (j = 0; j < (sizeof(cmdOption) / sizeof(struct cmdOption_t)); j++) {
			if (strcmp(tmpBuf, cmdOption[j].name) == 0) {
				while (*p != '\0') {
					for (k = 0; sep[k] != '\0'; k++) {
						if (*p == sep[k]) {
							break;
						}
					}
					if (sep[k] == '\0') {
						break;
					}
					p++;
				}

				result = cmdOption[j].set(pBidiProcC, p);
				if (result == FALSE) {
					return(FALSE);
				}
				cmdOption[j].valid = TRUE;
				break;
			}
		}
		if (j == (sizeof(cmdOption) / sizeof(struct cmdOption_t))) {
			DebugPrint("%s(%d) %s: illegal option: \"%s\"\n",
					   __FILE__, __LINE__, __FUNCTION__, argv[i]);
			return(FALSE);
		}
		i = index + 1;
	}

	i = 0;
	for (j = 0; j < (sizeof(cmdOption) / sizeof(struct cmdOption_t)); j++) {
		if (cmdOption[j].valid != FALSE) {

			DebugPrint("%s(%d) %s: %d: %s\n",
					   __FILE__, __LINE__, __FUNCTION__, j, cmdOption[j].name);

			i++;
		}
	}
	if (i != ((sizeof(cmdOption) / sizeof(struct cmdOption_t)) - 1)) {
		DebugPrint("%s(%d) %s: too few options\n", __FILE__, __LINE__, __FUNCTION__);
		return(FALSE);
	}

/*
	DebugPrint("%s(%d) %s: dataWriteFD = %d\n", __FILE__, __LINE__, __FUNCTION__, pBidiProcC->dataWriteFD);
	DebugPrint("%s(%d) %s: dataReadFD = %d\n", __FILE__, __LINE__, __FUNCTION__, pBidiProcC->dataReadFD);
	DebugPrint("%s(%d) %s: cmdWriteFD = %d\n", __FILE__, __LINE__, __FUNCTION__, pBidiProcC->cmdWriteFD);
	DebugPrint("%s(%d) %s: cmdReadFD = %d\n", __FILE__, __LINE__, __FUNCTION__, pBidiProcC->cmdReadFD);
	DebugPrint("%s(%d) %s: outputFD = %d\n", __FILE__, __LINE__, __FUNCTION__, pBidiProcC->outputFD);
	DebugPrint("%s(%d) %s: inputFD = %d\n", __FILE__, __LINE__, __FUNCTION__, pBidiProcC->inputFD);
*/

	return(TRUE);
}


/** ѹ¤Τ
 * @param *pBidiProcC ѹ¤ΤؤΥݥ
 * @return ʤ
 */
static void init(BidiProcC *pBidiProcC)
{
	memset(pBidiProcC, 0, sizeof(struct BidiProcC_t));
}



/* ---------------------------------------------------------------- */



/** signal äݤåɬפʤԤʤ
 * @param *pBidiProcC ѹ¤ΤؤΥݥ
 * @return ̵
 * */
static bool SignalCheck(BidiProcC *pBidiProcC)
{
//	int		result;

	if (signalCatchedFlag != 0) {
		DebugPrint("%s(%d) %s\n", __FILE__, __LINE__, __FUNCTION__);

		if (signalCatchedFlag == SIGHUP) {
/*
			result = bidiLibEndRead(pBidiProcC->pLibC);

			result = BIDI_EPROGRESS;
			// ȽλǤޤǴĥ
			while (result == BIDI_EPROGRESS) {
				result = bidiLibCancelJob(pBidiProcC->pLibC, pBidiProcC->idJob);
			}
*/
			signalCatchedFlag = 0;
		}
	}

	return(TRUE);
}

/** signal ȯν
 * SIGHUP ξϡޤ줿򵭲ΤߤǸ
 * @param signalId ȯʥ
 * @return ̵
 * */
static void SignalHandler(int signalId)
{
	DebugPrint("%s(%d) %s: signalId = %3d, %08X\n", __FILE__, __LINE__, __FUNCTION__, signalId, signalId);

	signalCatchedFlag = signalId;

	switch (signalId) {
		case SIGPIPE:
		case SIGTERM:
			DebugPrint("\t interrupted, exit\n");
			exit(1);
			break;

		case SIGHUP:
			DebugPrint("\t continue\n");
//			result = bidiLibCancelJob(pBidiProcC->pLibC, pBidiProcC->idJob);
			break;

		default:
			/* ʥʥξϽλ */
			DebugPrint("\t unknown interrupted, exit\n");
			exit(1);
			break;
	}
}


/** signal ȯνˡꤹ
 * @param ʤ
 * @retval TRUE Ǥ
 * @retval FALSE ˼Ԥ
 */
static bool SetSignal(void)
{
	sig_t	result;

	result = signal(SIGTERM, SignalHandler);
	if (result == SIG_ERR) {
		DebugPrint("%s(%d) %s: %d, %s\n", __FILE__, __LINE__, __FUNCTION__, errno, strerror(errno));
		return(FALSE);
	}
	result = signal(SIGPIPE, SignalHandler);
	if (result == SIG_ERR) {
		DebugPrint("%s(%d) %s: %d, %s\n", __FILE__, __LINE__, __FUNCTION__, errno, strerror(errno));
		return(FALSE);
	}
	result = signal(SIGHUP, SignalHandler);
	if (result == SIG_ERR) {
		DebugPrint("%s(%d) %s: %d, %s\n", __FILE__, __LINE__, __FUNCTION__, errno, strerror(errno));
		return(FALSE);
	}

	return(TRUE);
}


/** bidiNew  process Ǥμ
 * create a new object
 * @param *pBidiProcC ѹ¤ΤؤΥݥ
 * @param len ƽФ¦äޥɤĹ (Хȿ)
 * @retval BIDI_OK 
 * @retval BIDI_OKʳ Ԥξϼºݤˤ餺ΤԤ
 */
static int bidiProcNew(BidiProcC *pBidiProcC, int len)
{
	int	*buf;
	int	length;
	int	args[1];
	int	result;

	void *pLibC;			// handle used by lib module API

	DebugPrint("%s(%d) %s\n", __FILE__, __LINE__, __FUNCTION__);

	buf = pBidiProcC->cmdReadBuf;

	if (len != 4) {
		DebugPrint("%s(%d) %s: illegal length %d.\n", __FILE__, __LINE__, __FUNCTION__, len);
		(void)writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_ERROR, 0, NULL);
		goto errWait;
	}

	if (buf[0] > BIDI_PLUGIN_VERSION) {
		/* ϸŤ */
		DebugPrint("%s(%d) %s: requested vesion is new %d.\n", __FILE__, __LINE__, __FUNCTION__, buf[0]);
		goto errSend;
	}

	/* signal handler  */
	result = SetSignal();
	if (result == FALSE) {
		goto errSend;
	}

	// call bidiLibNew func
	pLibC = bidiLibNew(pBidiProcC->inputFD, pBidiProcC->outputFD, pBidiProcC->uri);
	if (pLibC == NULL) {
		DebugPrint("%s(%d) %s: error occur in bidiLibNew().\n", __FILE__, __LINE__, __FUNCTION__);
		goto errSend;
	}
	DebugPrint("%s(%d) %s: after bidiLibNew().\n", __FILE__, __LINE__, __FUNCTION__);


	// generating BidiC
	pBidiProcC->pLibC = pLibC;
	pBidiProcC->context = BIDI_CXT_NORMAL;
	pBidiProcC->saveContext = BIDI_CXT_NORMAL;


	if (signalCatchedFlag == 0) {
		args[0] = BIDI_OK;
		length = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_OK, 1, args);

		return(BIDI_OK);
	}


errSend:
	if (signalCatchedFlag == 0) {
		args[0] = BIDI_ERROR;
		(void)writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_OK, 1, args);
	}

errWait:
	/* ޤ̵¥롼 */
	SafeSleep(NULL);

	exit(1);

	/* un reached */
	return(FALSE);
}

/** bidiDestroy  process Ǥμ
 * ܴؿǤϥ顼̵뤷ǤäɤΤǤϡ
 */
static int bidiProcDestroy(BidiProcC *pBidiProcC, int len)
{
//	int	length;

	DebugPrint("%s(%d) %s: len = %d\n", __FILE__, __LINE__, __FUNCTION__, len);

	if (len != 0) {
		DebugPrint("%s(%d) %s: illegal length %d.\n", __FILE__, __LINE__, __FUNCTION__, len);
		/* TBD ####
		 * ǥ顼֤Ƥ̵̣ʤ褦ʡ */
//		length = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_ERROR, 0, NULL);
//		return(BIDI_ERROR);
	}

	// call bidiLibDestroy func
	bidiLibDestroy(pBidiProcC->pLibC);

//	if (signalCatchedFlag == 0) {
		(void)writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_OK, 0, NULL);
//	}

	DebugPrint("%s(%d) %s: process exit\n", __FILE__, __LINE__, __FUNCTION__);

	exit(0);
}

/** bidiGetCap  process Ǥμ */
static int bidiProcGetCap(BidiProcC *pBidiProcC, int len)
{
	int	*buf;
	int	length;
	int	result;
	int	args[1];

	DebugPrint("%s(%d) %s: len = %d\n", __FILE__, __LINE__, __FUNCTION__, len);

	if (IsPrinterCommunicationError(pBidiProcC->pLibC) != FALSE) {
		/* ̿顼ȯƤΤ̵ */
		DebugPrint("%s(%d) %s: printer communication error.\n", __FILE__, __LINE__, __FUNCTION__);
		args[0] = BIDI_ERROR;
		length = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_OK, 1, args);
		return(BIDI_ERROR);
	}

	buf = pBidiProcC->cmdReadBuf;

	if (len != 4) {
		DebugPrint("%s(%d) %s: illegal length %d.\n", __FILE__, __LINE__, __FUNCTION__, len);
		length = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_ERROR, 0, NULL);
		return(BIDI_ERROR);
	}

	result = bidiLibGetCap(pBidiProcC->pLibC, buf[0]);

	args[0] = result;
	length = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_OK, 1, args);

	return(BIDI_OK);
}

/** bidiStartJob  process Ǥμ */
static int bidiProcStartJob(BidiProcC *pBidiProcC, int len)
{
//	int	idJob;
	int	*buf;
	int	length;
	int	result;
	int	args[1];

	DebugPrint("%s(%d) %s: len = %d\n", __FILE__, __LINE__, __FUNCTION__, len);

	if (IsPrinterCommunicationError(pBidiProcC->pLibC) != FALSE) {
		/* ̿顼ȯƤΤ̵ */
		DebugPrint("%s(%d) %s: printer communication error.\n", __FILE__, __LINE__, __FUNCTION__);
		args[0] = BIDI_ERROR;
		length = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_OK, 1, args);
		return(BIDI_ERROR);
	}

	buf = pBidiProcC->cmdReadBuf;

	if (len != 4) {
		DebugPrint("%s(%d) %s: illegal length %d.\n", __FILE__, __LINE__, __FUNCTION__, len);
		length = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_ERROR, 0, NULL);
		return(BIDI_ERROR);
	}

	// check context
	if (pBidiProcC->context == BIDI_CXT_JOBIN) {
		DebugPrint("%s(%d) %s: illegal call.\n", __FILE__, __LINE__, __FUNCTION__);
		length = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_ERROR, 0, NULL);
		return(BIDI_ERROR);
	}

	// set context
	pBidiProcC->saveContext = pBidiProcC->context;
	pBidiProcC->context = BIDI_CXT_JOBIN;

	pBidiProcC->idJob = buf[0];

	// call bidiLibStartJob func
	result = bidiLibStartJob(pBidiProcC->pLibC, pBidiProcC->idJob);
	if (result != BIDI_OK) {
		DebugPrint("%s(%d) %s: result = %d\n", __FILE__, __LINE__, __FUNCTION__, result);
	}

	args[0] = result;
	length = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_OK, 1, args);

	return(result);
}

/** bidiEndJob  process Ǥμ */
static int bidiProcEndJob(BidiProcC *pBidiProcC, int len)
{
	int	*buf;
	int	length;
	int	result;
	int	args[1];

	DebugPrint("%s(%d) %s\n", __FILE__, __LINE__, __FUNCTION__);

	if (IsPrinterCommunicationError(pBidiProcC->pLibC) != FALSE) {
		/* ̿顼ȯƤΤ̵ */
		DebugPrint("%s(%d) %s: printer communication error.\n", __FILE__, __LINE__, __FUNCTION__);
		args[0] = BIDI_ERROR;
		length = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_OK, 1, args);
		return(BIDI_ERROR);
	}

	buf = pBidiProcC->cmdReadBuf;

	if (len != 0) {
		DebugPrint("%s(%d) %s: illegal length %d.\n", __FILE__, __LINE__, __FUNCTION__, len);
		length = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_ERROR, 0, NULL);
		return(BIDI_ERROR);
	}

	// check context
	if (pBidiProcC->context != BIDI_CXT_JOBIN) {
		DebugPrint("%s(%d) %s: illegal call.\n", __FILE__, __LINE__, __FUNCTION__);
		length = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_ERROR, 0, NULL);
		return(BIDI_ERROR);
	}

	result = bidiLibEndJob(pBidiProcC->pLibC);
	if (result != BIDI_OK) {
		DebugPrint("%s(%d) %s: result = %d\n", __FILE__, __LINE__, __FUNCTION__, result);
	}

	args[0] = result;
	length = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_OK, 1, args);

	if ((result != BIDI_OK) && (result != BIDI_EPROGRESS)) {
		return(result);
	}

	// set contextonly case of BIDI_OK
	if (result == BIDI_EPROGRESS) {
		result = BIDI_OK;
	} else {
		pBidiProcC->context = pBidiProcC->saveContext = BIDI_CXT_NORMAL;
//		pBidiProcC->context = pBidiProcC->saveContext;

		/* #### for profile */
//		exit(0);
	}

	return(BIDI_OK);
}

/** bidiCancelJob  process Ǥμ */
static int bidiProcCancelJob(BidiProcC *pBidiProcC, int len)
{
	int	idJob;
	int	*buf;
	int	length;
	int	result;
	int	args[1];

	DebugPrint("%s(%d) %s\n", __FILE__, __LINE__, __FUNCTION__);

	if (IsPrinterCommunicationError(pBidiProcC->pLibC) != FALSE) {
		/* ̿顼ȯƤΤ̵ */
		DebugPrint("%s(%d) %s: printer communication error.\n", __FILE__, __LINE__, __FUNCTION__);
		args[0] = BIDI_ERROR;
		length = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_OK, 1, args);
		return(BIDI_ERROR);
	}

	buf = pBidiProcC->cmdReadBuf;

	if (len != 4) {
		DebugPrint("%s(%d) %s: illegal length %d.\n", __FILE__, __LINE__, __FUNCTION__, len);
		length = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_ERROR, 0, NULL);
		return(BIDI_ERROR);
	}

	idJob = buf[0];

	// call bidiLibCancelJob func
	result = bidiLibCancelJob(pBidiProcC->pLibC, idJob);
	if (result != BIDI_OK) {
		DebugPrint("%s(%d) %s: result = %d\n", __FILE__, __LINE__, __FUNCTION__, result);
	}

	args[0] = result;
	length = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_OK, 1, args);

	if ((result == BIDI_ENOJOB) || (result == BIDI_EPROGRESS)) {
		result = BIDI_OK;
	}

	return(result);
}

/** bidiStartRead  process Ǥμ */
static int bidiProcStartRead(BidiProcC *pBidiProcC, int len)
{
	BidiReadMode	idReadMode;
	char	*pLang;
	char	*s;
	int	i;
	int	*buf;
	int	length;
	int	result;
	int	args[1];

	DebugPrint("%s(%d) %s: len = %d\n", __FILE__, __LINE__, __FUNCTION__, len);

	if (IsPrinterCommunicationError(pBidiProcC->pLibC) != FALSE) {
		/* ̿顼ȯƤΤ̵ */
		DebugPrint("%s(%d) %s: printer communication error.\n", __FILE__, __LINE__, __FUNCTION__);
		args[0] = BIDI_ERROR;
		length = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_OK, 1, args);
		return(BIDI_ERROR);
	}

	buf = pBidiProcC->cmdReadBuf;

	if (len < (4 + 4)) {
		DebugPrint("%s(%d) %s: illegal length %d.\n", __FILE__, __LINE__, __FUNCTION__, len);
		length = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_ERROR, 0, NULL);
		return(BIDI_ERROR);
	}

	// check context
	if (pBidiProcC->context == BIDI_CXT_READIN || pBidiProcC->context == BIDI_CXT_WRITEIN) {
		DebugPrint("%s(%d) %s: illegal call.\n", __FILE__, __LINE__, __FUNCTION__);
		length = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_ERROR, 0, NULL);
		return(BIDI_ERROR);
	}

	idReadMode = buf[0];
	/* pLang ɤ߹ */
	if ((len - 4 - 4) > 0) {

		if ((pLang = malloc(len - 4 - 4 + 1)) == NULL) {
			DebugPrint("%s(%d) %s: not enough core: len = %d.\n", __FILE__, __LINE__, __FUNCTION__, len);
			args[0] = BIDI_ERROR;
			length = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_OK, 1, args);
			return(BIDI_ERROR);
		}
		/* ʸ swap */
		for (i = 2; i < ((len + 3) / 4); i++) {
			buf[i] = ntohl(buf[i]);
		}
		buf[i] = 0;

		s = (char *)&buf[2];
		for (i = 0; i < (len - 4 - 4); i++) {
			pLang[i] = s[i];
		}
		pLang[i] = '\0';
	} else {
		pLang = NULL;
	}

	// set context
	pBidiProcC->saveContext = pBidiProcC->context;
	pBidiProcC->context = BIDI_CXT_READIN;

	// call bidiLibStartRead func
	result = bidiLibStartRead(pBidiProcC->pLibC, idReadMode, pLang);
	if (result != BIDI_OK) {
		DebugPrint("%s(%d) %s: result = %d\n", __FILE__, __LINE__, __FUNCTION__, result);
	}

	if (signalCatchedFlag == 0) {
		args[0] = result;
		length = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_OK, 1, args);
		if (result == BIDI_OK) {
			args[0] = 0;
			length = write(pBidiProcC->dataReadFD, (char *)&args, 1);
		}
	}

	DebugPrint("%s(%d) %s: result = %d, length = %d\n",
			   __FILE__, __LINE__, __FUNCTION__, result, length);

	if (pLang != NULL) {
		free(pLang);
	}

	return(result);
}

/** bidiRead  process Ǥμ */
static int bidiProcRead(BidiProcC *pBidiProcC, int len)
{
	int length;
	int	nBufBytes;

	char	*pBuf;
	int	*buf;
	int	result;
	int	args[1];

	DebugPrint("%s(%d) %s: len = %d\n", __FILE__, __LINE__, __FUNCTION__, len);

	if (IsPrinterCommunicationError(pBidiProcC->pLibC) != FALSE) {
		/* ̿顼ȯƤΤ̵ */
		DebugPrint("%s(%d) %s: printer communication error.\n", __FILE__, __LINE__, __FUNCTION__);
		args[0] = BIDI_ERROR;
		length = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_OK, 1, args);
		return(BIDI_ERROR);
	}

	buf = pBidiProcC->cmdReadBuf;

	if (len != 4) {
		DebugPrint("%s(%d) %s: illegal length %d.\n", __FILE__, __LINE__, __FUNCTION__, len);
		result = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_ERROR, 0, NULL);
		return(BIDI_ERROR);
	}

	nBufBytes = buf[0];

	// check context
	if (pBidiProcC->context != BIDI_CXT_READIN) {
		DebugPrint("%s(%d) %s: illegal call.\n", __FILE__, __LINE__, __FUNCTION__);
		result = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_ERROR, 0, NULL);
		return(BIDI_ERROR);
	}

	if ((pBuf = malloc(nBufBytes)) == NULL) {
		DebugPrint("%s(%d) %s: not enough core.\n", __FILE__, __LINE__, __FUNCTION__);
		args[0] = BIDI_ERROR;
		result = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_OK, 1, args);
		return(BIDI_ERROR);
	}

	length = bidiLibRead(pBidiProcC->pLibC, pBuf, nBufBytes);

	if (signalCatchedFlag == 0) {
		args[0] = length;
		result = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_OK, 1, args);

		result = write(pBidiProcC->dataReadFD, pBuf, length);
	}

	free(pBuf);

	return(result);
}

/** bidiEndRead  process Ǥμ */
static int bidiProcEndRead(BidiProcC *pBidiProcC, int len)
{
	int	*buf;
	int	result;
	int	length;
	int	args[1];

	DebugPrint("%s(%d) %s\n", __FILE__, __LINE__, __FUNCTION__);

	if (IsPrinterCommunicationError(pBidiProcC->pLibC) != FALSE) {
		/* ̿顼ȯƤΤ̵ */
		DebugPrint("%s(%d) %s: printer communication error.\n", __FILE__, __LINE__, __FUNCTION__);
		args[0] = BIDI_ERROR;
		length = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_OK, 1, args);
		return(BIDI_ERROR);
	}

	buf = pBidiProcC->cmdReadBuf;

	if (len != 0) {
		DebugPrint("%s(%d) %s: illegal length %d.\n", __FILE__, __LINE__, __FUNCTION__, len);
		length = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_ERROR, 0, NULL);
		return(BIDI_ERROR);
	}

	// check context
	if (pBidiProcC->context != BIDI_CXT_READIN) {
		DebugPrint("%s(%d) %s: illegal call.\n", __FILE__, __LINE__, __FUNCTION__);
		length = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_ERROR, 0, NULL);
		return(BIDI_ERROR);
	}

	// set context
	pBidiProcC->context = pBidiProcC->saveContext;
	pBidiProcC->saveContext = BIDI_CXT_NORMAL;

	// call bidiLibEndRead func
	result = bidiLibEndRead(pBidiProcC->pLibC);
	if (result != BIDI_OK) {
		DebugPrint("%s(%d) %s: result = %d\n", __FILE__, __LINE__, __FUNCTION__, result);
	}

	if (signalCatchedFlag == 0) {
		args[0] = result;
		length = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_OK, 1, args);
	}

	return(result);
}

/** bidiStartWrite  process Ǥμ */
static int bidiProcStartWrite(BidiProcC *pBidiProcC, int len)
{
	int	*buf;
	int	result;
	int	length;
	int	args[1];

	DebugPrint("%s(%d) %s\n", __FILE__, __LINE__, __FUNCTION__);

	if (IsPrinterCommunicationError(pBidiProcC->pLibC) != FALSE) {
		/* ̿顼ȯƤΤ̵ */
		DebugPrint("%s(%d) %s: printer communication error.\n", __FILE__, __LINE__, __FUNCTION__);
		args[0] = BIDI_ERROR;
		length = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_OK, 1, args);
		return(BIDI_ERROR);
	}

	buf = pBidiProcC->cmdReadBuf;

	if (len != 0) {
		DebugPrint("%s(%d) %s: illegal length %d.\n", __FILE__, __LINE__, __FUNCTION__, len);
		length = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_ERROR, 0, NULL);
		return(BIDI_ERROR);
	}

	// check context
	if (pBidiProcC->context == BIDI_CXT_READIN || pBidiProcC->context == BIDI_CXT_WRITEIN) {
		DebugPrint("%s(%d) %s: illegal call.\n", __FILE__, __LINE__, __FUNCTION__);
		length = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_ERROR, 0, NULL);
		return(BIDI_ERROR);
	}

	// set context
	pBidiProcC->saveContext = pBidiProcC->context;
	pBidiProcC->context = BIDI_CXT_WRITEIN;

	// call bidiLibStartWrite func
	result = bidiLibStartWrite(pBidiProcC->pLibC);
	if (result != BIDI_OK) {
		DebugPrint("%s(%d) %s: bidiLibStartWrite() result = %d\n", __FILE__, __LINE__, __FUNCTION__, result);
	}

	if (signalCatchedFlag == 0) {
		args[0] = result;
		length = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_OK, 1, args);
	}

	return(result);
}

/** bidiWrite  process Ǥμ */
static int bidiProcWrite(BidiProcC *pBidiProcC, int len)
{
	int	nBufBytes;
	char	*pBuf = NULL;
	int	length;
	int	result;
	int	*buf;
	int	args[1];
	struct timeval	uWaitVal;
	struct timespec	nWaitVal;

	static int	wc = 0;

	DebugPrint("%s(%d) %s: len = %d\n", __FILE__, __LINE__, __FUNCTION__, len);

	if (IsPrinterCommunicationError(pBidiProcC->pLibC) != FALSE) {
		/* ̿顼ȯƤΤ̵ */
		DebugPrint("%s(%d) %s: printer communication error.\n", __FILE__, __LINE__, __FUNCTION__);
		args[0] = BIDI_ERROR;
		length = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_OK, 1, args);
		return(BIDI_ERROR);
	}

	length = 0;

	wc++;

	buf = pBidiProcC->cmdReadBuf;

	if (len != 4) {
		DebugPrint("%s(%d) %s: illegal length %d.\n", __FILE__, __LINE__, __FUNCTION__, len);
		result = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_ERROR, 0, NULL);
		return(BIDI_ERROR);
	}

	// check context
	if (pBidiProcC->context != BIDI_CXT_WRITEIN) {
		DebugPrint("%s(%d) %s: illegal call.\n", __FILE__, __LINE__, __FUNCTION__);
		result = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_ERROR, 0, NULL);
		return(BIDI_ERROR);
	}

	/* ХåեΥå */
	length = 0;
	nWaitVal.tv_sec = 2;		/* 2Ԥ  */
    nWaitVal.tv_nsec = 0;
	result = CanDataWriteRequest(pBidiProcC->pLibC, &nWaitVal);
	if (result != FALSE) {
		/* ꥹȤ˶ */
		nBufBytes = buf[0];
		/* ݤ׵̤ */
		if ((pBuf = malloc(nBufBytes)) != NULL) {
			length = nBufBytes;
		}
	}
	if (signalCatchedFlag != 0) {
		if (pBuf != NULL) {
			free(pBuf);
		}
		result = BIDI_ERROR;
		goto errRet;
	}

	args[0] = length;
	result = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_OK, 1, args);
	if (length == 0) {
		DebugPrint("%s(%d) %s: Can't write, wc = %d.\n", __FILE__, __LINE__, __FUNCTION__, wc);
		return(BIDI_OK);
	}


	DebugPrint("%s(%d) %s: Can write, wc = %d.\n", __FILE__, __LINE__, __FUNCTION__, wc);


	/* ####
	 * ʹߤϡ顼ȯƤ caller ˥顼ɤ֤ˡʤΤ
	 * signal Ԥ */
	if (signalCatchedFlag == 0) {
		result = read(pBidiProcC->dataWriteFD, pBuf, length);
	}
	if ((signalCatchedFlag != 0) || (result < 0)) {
		/* error */
		if (pBuf != NULL) {
			free(pBuf);
		}
		result = BIDI_ERROR;
		goto errRet;
	}

	result = bidiLibWrite(pBidiProcC->pLibC, pBuf, result);
	/* Ǥɬ񤭹 */

	return(result);

errRet:

	/* Ҥ signal Ԥ */
	while (signalCatchedFlag == 0) {
		uWaitVal.tv_sec = 0;
		uWaitVal.tv_usec = 500 * 1000;
		result = SafeSleep(&uWaitVal);
	}

	return(result);
}

/** bidiEndWrite  process Ǥμ */
static int bidiProcEndWrite(BidiProcC *pBidiProcC, int len)
{
	int	*buf;
	int	length;
	int	result;
	int	args[1];

	DebugPrint("%s(%d) %s\n", __FILE__, __LINE__, __FUNCTION__);

	if (IsPrinterCommunicationError(pBidiProcC->pLibC) != FALSE) {
		/* ̿顼ȯƤΤ̵ */
		DebugPrint("%s(%d) %s: printer communication error.\n", __FILE__, __LINE__, __FUNCTION__);
		args[0] = BIDI_ERROR;
		length = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_OK, 1, args);
		return(BIDI_ERROR);
	}

	buf = pBidiProcC->cmdReadBuf;

	// check context
	if (pBidiProcC->context != BIDI_CXT_WRITEIN) {
		DebugPrint("%s(%d) %s: illegal call.\n", __FILE__, __LINE__, __FUNCTION__);
		length = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_ERROR, 0, NULL);
		return(BIDI_ERROR);
	}

	// set context
	pBidiProcC->context = pBidiProcC->saveContext;
	pBidiProcC->saveContext = BIDI_CXT_NORMAL;


	// call bidiLibEndWrite func
	result = bidiLibEndWrite(pBidiProcC->pLibC);
	if (result != BIDI_OK) {
		DebugPrint("%s(%d) %s: result = %d\n", __FILE__, __LINE__, __FUNCTION__, result);
	}

	if (signalCatchedFlag == 0) {
		args[0] = result;
		length = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_OK, 1, args);
	}

	return(result);
}

/** bidiCtrl  process Ǥμ ()
 * ͳ˸ľɬ
 */
static int bidiProcCtrl(BidiProcC *pBidiProcC, int len)
{
	int	idRequest;
	int	nDataBytes;
	char	*pBuf = NULL;
	int	result;

	int	nBufBytes;
	int	*buf;
	int	length;
	int	args[1];

	DebugPrint("%s(%d) %s\n", __FILE__, __LINE__, __FUNCTION__);

	buf = pBidiProcC->cmdReadBuf;

	if (IsPrinterCommunicationError(pBidiProcC->pLibC) != FALSE) {
		/* ̿顼ȯƤΤ̵ */
		DebugPrint("%s(%d) %s: printer communication error.\n", __FILE__, __LINE__, __FUNCTION__);
		args[0] = BIDI_ERROR;
		length = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_OK, 1, args);
		return(BIDI_ERROR);
	}

	if (len != 8) {
		DebugPrint("%s(%d) %s: illegal length %d.\n", __FILE__, __LINE__, __FUNCTION__, len);
		length = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_ERROR, 0, NULL);
		return(BIDI_ERROR);
	}

	idRequest = buf[0];
	nBufBytes = buf[1];

	if ((pBuf = malloc(nBufBytes)) == NULL) {
		args[0] = BIDI_ERROR;
		length = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_OK, 1, args);
		return(BIDI_ERROR);
	}

	length = read(pBidiProcC->dataWriteFD, pBuf, nBufBytes);
	if (length < 0) {
		/* error */
		args[0] = BIDI_ERROR;
		length = writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_OK, 1, args);
		result = BIDI_ERROR;
		goto errRet;
	}

	// call bidiLibCtrl func
	result = bidiLibCtrl(pBidiProcC->pLibC, idRequest, pBuf, nDataBytes);

errRet:
	if (pBuf != NULL) {
		free(pBuf);
	}

	return(result);
}

/* ---------------------------------------------------------------- */

/** caller (Plugin) Υޥɤ
 * @param *pBidiProcC ѹ¤ΤؤΥݥ
 * @return ºݤˤ餺ޤư³
 */
static bool parseBidiCmd(BidiProcC *pBidiProcC)
{
	int	result;
	fd_set	waitFd;
	int	maxFd;
	int cmdID;
	int	len;

	DebugPrint("%s(%d) %s\n", __FILE__, __LINE__, __FUNCTION__);

	signalCatchedFlag = 0;

	if (pBidiProcC->dataWriteFD > pBidiProcC->cmdWriteFD) {
		maxFd = pBidiProcC->dataWriteFD;
	} else {
		maxFd = pBidiProcC->cmdWriteFD;
	}
	FD_ZERO(&waitFd);

	while (1) {

		/* caller ΥޥԤ */
		FD_SET(pBidiProcC->cmdWriteFD, &waitFd);
		result = select(maxFd + 1, &waitFd, NULL, NULL, NULL); /* ̵¥롼 */
		if (result < 0) {
			DebugPrint("%s(%d) %s: select() error.\n", __FILE__, __LINE__, __FUNCTION__);
			continue;
		}
		if ((result > 0) && (FD_ISSET(pBidiProcC->cmdWriteFD, &waitFd) == 0)) {
			continue;
		}

		/* caller 鲿褿 */

		/* 褺 CommandIDData Length ɤ */
		len = cmdRead(pBidiProcC->cmdWriteFD, &cmdID, pBidiProcC->cmdReadBuf, sizeof(pBidiProcC->cmdReadBuf));
		if (len < 0) {
			/* TBD ####
			 * 顼ɤɤ褦 */
			DebugPrint("%s(%d) %s: cmdRead fail cmdID = %08X, len = %d.\n",
					   __FILE__, __LINE__, __FUNCTION__, cmdID, len);
			continue;
//			return(BIDI_ERROR);
		}
		switch (cmdID) {
			case BIDI_CMD_NEW:
				result = bidiProcNew(pBidiProcC, len);
				break;

			case BIDI_CMD_DESTROY:
				result = bidiProcDestroy(pBidiProcC, len);
				break;

			case BIDI_CMD_GETCAP:
				result = bidiProcGetCap(pBidiProcC, len);
				break;

			case BIDI_CMD_STARTJOB:
				result = bidiProcStartJob(pBidiProcC, len);
				break;

			case BIDI_CMD_ENDJOB:
				result = bidiProcEndJob(pBidiProcC, len);
				break;

			case BIDI_CMD_CANCELJOB:
				result = bidiProcCancelJob(pBidiProcC, len);
				break;

			case BIDI_CMD_STARTREAD:
				result = bidiProcStartRead(pBidiProcC, len);
				break;

			case BIDI_CMD_ENDREAD:
				result = bidiProcEndRead(pBidiProcC, len);
				break;

			case BIDI_CMD_READ:
				result = bidiProcRead(pBidiProcC, len);
				break;

			case BIDI_CMD_STARTWRITE:
				result = bidiProcStartWrite(pBidiProcC, len);
				break;

			case BIDI_CMD_ENDWRITE:
				result = bidiProcEndWrite(pBidiProcC, len);
				break;

			case BIDI_CMD_WRITE:
				result = bidiProcWrite(pBidiProcC, len);
				break;

			case BIDI_CMD_CTRL:
				result = bidiProcCtrl(pBidiProcC, len);
				break;

			default:
				/* ̤ݡȤξ */
				DebugPrint("%s(%d) %s\n", __FILE__, __LINE__, __FUNCTION__);
				(void)writePacket(pBidiProcC->cmdReadFD, BIDI_CMD_ERROR, 0, NULL);
				result = BIDI_ERROR;
				break;
		}

		/* signal äݤΥå */
		SignalCheck(pBidiProcC);
	}

	/* ؤʤ */
	return(result);
}


/** bidi process Ǥ ȥå״ؿ
 * @param argc ޥɥ饤θĿ
 * @param *argv[] ޥɥ饤
 * @return åǤΥ顼Τܴؿǽλ롣
 * ̤ parseBidiCmd ؿ餺ޤư³
 */
int main(int argc, char *argv[])
{
	BidiProcC	pBidiProcC;
	int result;

	DebugPrint("%s(%d) %s\n", __FILE__, __LINE__, __FUNCTION__);

	/*  */
	init(&pBidiProcC);

#ifdef DEBUG
	for (result = 0; result < argc; result++) {
		fprintf(stderr, "%d: \"%s\"\n", result, argv[result]);
	}
#endif

	/* å */
	result = argCheck(argc, argv, &pBidiProcC);
	if (result == FALSE) {
		/* TBD ####
		 * ȯʤȦ
		 * caller ˥顼֤ */
		exit(1);
	}

	result = parseBidiCmd(&pBidiProcC);

	/* äʤ */

	exit(result);
}
