/*  test_main.c -- ECBTΥƥѥץ
 *  Copyright (C) 2003  EPSON KOWA Corporation
 *
 *  NOTE: FIFOեϡͽƤ¹ԤƤ
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/select.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <time.h>
#include "bidiEpson.h"

#define FIFO_FILEPATH "./fifo0"
#define USB_FILEPATH "/dev/usb/lp2"

static int
loop(EcbtHandle hl)
{
	char get_status_command[] = { 's', 't', 0x01, 0x00, 0x01 };
/* 	char timer_set_command[] = { 't', 'i', 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }; */
	int fd, max;
	fd_set  watchset;
	time_t after;
	int result = 0;
	int count = 0;
	char rbuf[512];
	char wbuf[4084];


	fd = open(FIFO_FILEPATH, O_RDONLY | O_NONBLOCK);
	if (fd < 0) {
		perror(FIFO_FILEPATH);
		return -1;
	}

	FD_ZERO(&watchset);
	FD_SET(fd, &watchset);
	max = fd;
	
	after = time(NULL);

	for(;;) {
		struct timeval tv;
		fd_set  inset;
	
		inset = watchset;
		tv.tv_sec = 2;
		tv.tv_usec = 0;
		
		if(select(max + 1, &inset, NULL, NULL, &tv) < 0) {
			perror("select");
			close(fd);
			return -1;
		}

		/* Get status */
		if (after < time(NULL)) {
			int rcount;

			after +=2;

			result = EcbtWrite(hl, CHID_COMMAND, get_status_command, sizeof(get_status_command), &rcount);
			if (result == ECBT_ERR_NOERR) {
				int rsize;
				
				result = EcbtRead(hl, CHID_COMMAND, NULL, 511, &rsize);
				if (result == ECBT_ERR_NOERR) {
					result = EcbtRead(hl, CHID_COMMAND, rbuf, rsize, &rcount);
					if (result == ECBT_ERR_NOERR) {
						rbuf[rsize] = '\0';
						fprintf(stderr, "[PrinterStatus]\"%s\"\n", rbuf);
					} else {
						fprintf(stderr, "ERROR:EcbtRead[Get Status](%d)\n", result);
					}
				} else {
					fprintf(stderr, "ERROR:EcbtRead[Get Status - SIZE](%d)\n", result);
				}
			} else {
				fprintf(stderr, "ERROR:EcbtWrite[Get Status](%d)\n", result);
			}
			
			if (result != ECBT_ERR_NOERR) {
				return -1;
			}
		}
		
		/* Check FIFO buffer */
		if (FD_ISSET(fd, &inset)) {
			int wcount;
			if (count == 0) {
				count = read(fd, wbuf, 4084);
				if (count < 0) {
					perror("read");
					return -1;
				}
			}

			result = EcbtWrite(hl, CHID_DATA, wbuf, count, &wcount);
			if (result != ECBT_ERR_NOERR && result != ECBT_ERR_BUSY) {
				fprintf(stderr, "ECBT WRITE_ERROR (%d->%d) err=%d\n", count, wcount, result);
			} else if (wcount < count) {
				int i;

				for (i = 0; i < (count - wcount); i++) {
					wbuf[i] = wbuf[wcount + i];
				}
				count -= wcount;
			} else {
				count = 0;
			}
		}
	}

	return 0;
}


int
main(int argc, char *argv[])
{
	EcbtHandle hl;
	int fd;
	int result = 0;


	fd = open(USB_FILEPATH, O_RDWR /*| O_NONBLOCK*/);
	if (fd < 0) {
		perror(USB_FILEPATH);
		return -1;
	}

	result = EcbtOpen(&hl, fd);
	if (result != ECBT_ERR_NOERR) {
		fprintf(stderr, "ERROR:EcbtOpen(%d)\n", result);
	}
	
	if (result == ECBT_ERR_NOERR) {
		result = EcbtChannelOpen(hl, CHID_COMMAND);
		if (result != ECBT_ERR_NOERR) {
			fprintf(stderr, "ERROR:EcbtChannelOpen[COMMAND](%d)\n", result);
		}
	}

	if (result == ECBT_ERR_NOERR) {
		do {
			result = EcbtChannelOpen(hl, CHID_DATA);
			if (result != ECBT_ERR_NOERR) {
				fprintf(stderr, "ERROR:EcbtChannelOpen[DATA](%d)\n", result);
			}
			if (result == ECBT_ERR_BUSY) {
				sleep(1);
			}
		} while (result == ECBT_ERR_BUSY);
	}
	
	if (result == ECBT_ERR_NOERR) {
		result = loop(hl);
	}

	if (result == ECBT_ERR_NOERR) {
		result = EcbtChannelClose(hl, CHID_DATA);
		if (result != ECBT_ERR_NOERR) {
			fprintf(stderr, "ERROR:EcbtChannelClose[DATA](%d)\n", result);
		}
	}

	if (result == ECBT_ERR_NOERR) {
		result = EcbtChannelClose(hl, CHID_COMMAND);
		if (result != ECBT_ERR_NOERR) {
			fprintf(stderr, "ERROR:EcbtChannelClose[COMMAND](%d)\n", result);
		}
	}
	
	if (result == ECBT_ERR_NOERR) {
		result = EcbtClose(hl);
		if (result != ECBT_ERR_NOERR) {
			fprintf(stderr, "ERROR:EcbtClose(%d)\n", result);
		}
	}


	if (result != ECBT_ERR_NOERR) {
		fprintf(stderr, "ECBT TEST: *** NG ***\n");
		return -1;
	}

	fprintf(stderr, "ECBT TEST: *** OK ***\n");
	return 0;
}
