/*
 *  CUPS add-on module for OPVP ver.0.1
 *  Copyright CANON INC. 2001-2004
 *  All Rights Reserved.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <cups/cups.h>
#include <cups/ppd.h>
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <ctype.h>
#include <errno.h>


#include "canonopt.h"
#include "paramlist.h"


//#define DEBUG_SLEEP
//#define DEBUG_PS


#define	SHELL_PATH		"/bin/"
#define	SHELL_NAME		"sh"
#define	DATA_BUF_SIZE	(1024 * 256)

#define PS_PATH			"/tmp/canon_opvp_data%d.ps"
#define BEGIN_FEATURE	"\%\%BeginFeature"
#define PAGE_DEV_BEGIN	"<<"
#define PAGE_DEV_END	">>setpagedevice"
#define	OPVP_MODEL		"*opvpModel"
#define	OPVP_DRIVER		"*opvpDriver"
#define	OPVP_DEVICE		"*opvpDevice"

//#define	USE_CUSTOM_SIZE
#ifdef	USE_CUSTOM_SIZE
#define	GS_PARAMS		"gs -r%d -g%dx%d -q -dBATCH -dSAFER \
-dQUIET -dNOPAUSE -sDEVICE=%s -sDriver=%s -sModel=%s"
#else
#define GS_PARAMS		"gs -r%d -sPAPERSIZE=%s -q -dBATCH -dSAFER \
-dQUIET -dNOPAUSE -sDEVICE=%s -sDriver=%s -sModel=%s"
#endif
#define	DOC_PARAM	" -sDocInfo="
#define	GS_OUTPUT	" -sOutputFile=- -"

#define IS_BLANK(c)		(c == ' '  || c == '\t')
#define IS_RETURN(c)	(c == '\r' || c == '\n')


static
ParamList *get_ps_params(int *p_ifd, char **p_ps_path)
{
	FILE *ps_fp;
	int ifd = *p_ifd;
	int cc;
	ParamList *p_list = NULL;
	unsigned char *p_read_buf;
	unsigned char *p_data_buf;
	int read_byte;
	static char ps_path[256];
	char key[256];
	char value[256];

	sprintf(ps_path, PS_PATH, getpid());
	ps_fp = fopen(ps_path, "w+");
	*p_ps_path = ps_path;

	if( ps_fp == NULL )
		return NULL;

	p_read_buf = (char*)malloc(DATA_BUF_SIZE);
	p_data_buf = (char*)malloc(DATA_BUF_SIZE);

	if( p_read_buf != NULL && p_data_buf != NULL )
	{
		while( (read_byte = read(ifd, p_read_buf, DATA_BUF_SIZE)) > 0 )
		{
			int read_pos = 0;
			int data_byte = 0;

			while( read_pos < read_byte )
			{
				p_data_buf[data_byte++] = cc = p_read_buf[read_pos++];

				if( data_byte == DATA_BUF_SIZE )
				{
					if( fwrite(p_data_buf, 1, data_byte, ps_fp) != data_byte )
						goto write_error;
					data_byte = 0;
				}
				else if( IS_RETURN(cc) )
				{
					int begin_len = strlen(BEGIN_FEATURE);
					int dev_len = strlen(PAGE_DEV_BEGIN);
					int end_len = strlen(PAGE_DEV_END);

					// Interpret %%BeginFeature comment.
					if( strncmp(p_data_buf, BEGIN_FEATURE, begin_len) == 0 )
					{
						int pos = begin_len;
						int key_pos = 0;
						int val_pos = 0;

						while( IS_BLANK(p_data_buf[pos])
							|| p_data_buf[pos] == ':'
							|| p_data_buf[pos] == '*' )	pos++;

						while( !IS_BLANK(p_data_buf[pos])
							&& !IS_RETURN(p_data_buf[pos])
							&& key_pos < 255 )
								key[key_pos++] = p_data_buf[pos++];
						key[key_pos++] = 0;

						while(IS_BLANK(p_data_buf[pos]))
							pos++;

						while( !IS_BLANK(p_data_buf[pos])
							&& !IS_RETURN(p_data_buf[pos])
							&& val_pos < 255 )
							value[val_pos++] = p_data_buf[pos++];
						value[val_pos++] = 0;

						param_list_add_multi(&p_list, key, value, val_pos+1, 1);
					}
					// Interpret <</***(...)>>setpagedevice.
					else if( strncmp(p_data_buf, PAGE_DEV_BEGIN, dev_len) == 0
					 && (data_byte - 1 - end_len) )
					{
						if( strncmp(&p_data_buf[data_byte - 1 - end_len],
							 PAGE_DEV_END, end_len) == 0 )
						{
							int pos = dev_len;

							while( !IS_RETURN(p_data_buf[pos]) )
							{
								int key_pos = 0;
								int val_pos = 0;

								while( p_data_buf[pos] != '/'
									&& !IS_RETURN(p_data_buf[pos]) ) pos++;

								if( p_data_buf[pos] == '/' ) pos++;
								else continue;

								while( isalnum(p_data_buf[pos])
									&& key_pos < 255 )
										key[key_pos++] = p_data_buf[pos++];
								key[key_pos++] = 0;

								if( p_data_buf[pos] == '(' )
								{
									pos++;

									while( p_data_buf[pos] != ')'
										&& !IS_BLANK(p_data_buf[pos])
										&& !IS_RETURN(p_data_buf[pos])
										&& val_pos < 255 )
										value[val_pos++] = p_data_buf[pos++];
									value[val_pos++] = 0;

									if( p_data_buf[pos] == ')' ) pos++;
									else continue;

									if( !strcmp(key, "CNPageSizeName") )
										strcpy(key, "PageSize");
								}
								else continue;

								param_list_add_multi(
									&p_list, key, value, val_pos+1, 1);
							}
						}
					}

					if( data_byte > 0 )
					{
						if( fwrite(p_data_buf, 1, data_byte, ps_fp)
														 != data_byte )
							goto write_error;
					}
					data_byte = 0;
				}
			}
			if( data_byte > 0 )
			{
				if( fwrite(p_data_buf, 1, data_byte, ps_fp) != data_byte )
					goto write_error;
			}
		}
	}

	if( p_read_buf != NULL )
		free(p_read_buf);
	if( p_data_buf != NULL )
		free(p_data_buf);

	if( *p_ifd  != 0 )
		close(*p_ifd);

	fseek(ps_fp, 0L, SEEK_SET);
	*p_ifd = fileno(ps_fp);

	return p_list;

write_error:
	if( p_list != NULL )
		param_list_free(p_list);

	if( *p_ifd  != 0 )
		close(*p_ifd);
	*p_ifd = -1;

	return NULL;
}

static
char *strtolower(char *str)
{
	unsigned char *cc = str;
	while( *cc )
	{
		*cc = tolower(*cc);
		cc++;
	}
	return str;
}

static
void mark_ps_param_options(ppd_file_t *p_ppd, ParamList *p_param)
{
	int num = param_list_num(p_param);

	if( num > 0 )
	{
		cups_option_t *options
			= (cups_option_t*)malloc(num * sizeof(cups_option_t));
		if( options )
		{
			int i;
			for( i = 0 ; i < num ; i++ )
			{
				options[i].name = p_param->key;
				options[i].value = strtolower(p_param->value);
				p_param = p_param->next;
			}
			for( i = num - 1 ; i >= 0 ; i-- )
			{
				cupsMarkOptions(p_ppd, 1, &options[i]);
			}

			free(options);
		}
	}
}


static
int get_string(char *p_line_buf, char *p_value_buf, int value_len)
{
	int pos = 0;
	int val_pos = 0;

	while( IS_BLANK(p_line_buf[pos]) )
		pos++;
	while( p_line_buf[pos] == ':' )
		pos++;
	while( IS_BLANK(p_line_buf[pos]) )
		pos++;
	while( p_line_buf[pos] == '"')
		pos++;

	while( !IS_BLANK(p_line_buf[pos])
		&& !IS_RETURN(p_line_buf[pos])
		&& p_line_buf[pos] != '"'
		&& p_line_buf[pos] != '\n'
		&& val_pos < 255 )
			p_value_buf[val_pos++] = p_line_buf[pos++];
	p_value_buf[val_pos++] = 0;

	return val_pos;
}

#define	LINE_BUF_SIZE	1024
#define	READ_DRIVER		1
#define	READ_MODEL		2
#define	READ_DEVICE		4
#define	READ_DONE		(READ_DRIVER | READ_MODEL | READ_DEVICE)

static
int get_driver_model(char *ppd_file,
	char *p_driver, char *p_model, char *p_device, int len)
{
	FILE *fp = fopen(ppd_file, "r");
	char line_buf[LINE_BUF_SIZE];
	int flag = 0;

	if( fp )
	{
		int model_len = strlen(OPVP_MODEL);
		int driver_len = strlen(OPVP_DRIVER);
		int device_len = strlen(OPVP_DEVICE);
		int cc;
		int pos = 0;

		do
		{
			cc = fgetc(fp);

			if( cc != '\n' && cc != EOF )
			{
				if( pos < LINE_BUF_SIZE - 1 )
					line_buf[pos++] = cc;
			}
			else
			{
				line_buf[pos] = 0;

				if( strncmp(line_buf, OPVP_MODEL, model_len) == 0 )
				{
					if( get_string(line_buf + model_len, p_model, len) > 0 )
						flag |= READ_MODEL;
				}
				else if( strncmp(line_buf, OPVP_DRIVER, driver_len) == 0 )
				{
					if( get_string(line_buf + driver_len, p_driver, len) > 0 )
						flag |= READ_DRIVER;
				}
				else if( strncmp(line_buf, OPVP_DEVICE, device_len) == 0 )
				{
					if( get_string(line_buf + device_len, p_device, len) > 0 )
						flag |= READ_DEVICE;
				}

				if( flag == READ_DONE )
					break;

				pos = 0;
			}
		}
		while( cc != EOF );

		fclose(fp);
	}
	return (flag == READ_DONE)? 0 : -1;
}


static
int add_cmd_line_params(char *p_cmd_buf, ParamList *p_list, int cmd_buf_len)
{
	if( strlen(p_cmd_buf) + strlen(DOC_PARAM) + 1 < cmd_buf_len - 1 )
	{
		strcat(p_cmd_buf, DOC_PARAM);
		strcat(p_cmd_buf, "\"");

		while( p_list )
		{
			int len = strlen(p_list->key) + strlen(p_list->value) + 2;
			if( strlen(p_cmd_buf) + len < cmd_buf_len - 1 )
			{
				strcat(p_cmd_buf, p_list->key);
				strcat(p_cmd_buf, "=");
				strcat(p_cmd_buf, p_list->value);
			}
			else
			{
				p_cmd_buf[0] = 0;
				return -1;
			}

			p_list = p_list->next;

			if( p_list )
				strcat(p_cmd_buf, ",");
		}
		strcat(p_cmd_buf, "\"");
		return strlen(p_cmd_buf);
	}
	else
	{
		p_cmd_buf[0] = 0;
		return -1;
	}
}

#define	OPT_BUF_SIZE	1024

static
int make_cmd_line(cups_option_t *p_opt, int num_opt,
	ParamList *p_param, char *cmd_buf, int cmd_buf_len)
{
	int cmd_len = -1;
	char *ppd_name = getenv("PPD");
	ppd_file_t *p_ppd;

	if( (p_ppd = ppdOpenFile(ppd_name)) != NULL )
	{
		PpdToOptKey *p_opt_key_table;
		char model_name[256];
		char driver_name[256];
		char device_name[256];

		model_name[0] = '\0';
		driver_name[0] = '\0';

		// Obtain driver and model name.
		if( get_driver_model(ppd_name,
			driver_name, model_name, device_name, 256) < 0 )
		{
			ppdClose(p_ppd);
			return -1;
		}

		p_opt_key_table = alloc_opt_key_table(ppd_name);

		if( p_opt_key_table != NULL )
		{
			char *opt_key;
			ParamList *p_list = NULL;
			ppd_choice_t *p_choice;
			ppd_size_t *p_size;
			char *p_size_name;
			int reso;
			char opt_buf[OPT_BUF_SIZE];
			int opt_len;

			opt_buf[0] = '\0';

			ppdMarkDefaults(p_ppd);
			mark_ps_param_options(p_ppd, p_param);
			cupsMarkOptions(p_ppd, num_opt, p_opt);

			// Obtain page size setting.
			p_choice = ppdFindMarkedChoice(p_ppd, "PageSize");
			p_size_name = p_choice->choice;
			p_size = ppdPageSize(p_ppd, p_size_name);

			opt_key = ppd_to_opt_key(p_opt_key_table, "PageSize");
			param_list_add(&p_list, opt_key,
					p_choice->choice, strlen(p_choice->choice) + 1);

			// Obtain resolution setting.
			p_choice = ppdFindMarkedChoice(p_ppd, "Resolution");
			reso = atoi(p_choice->choice);

			opt_key = ppd_to_opt_key(p_opt_key_table, "Resolution");
			param_list_add(&p_list, opt_key,
					p_choice->choice, strlen(p_choice->choice) + 1);

			// Convert all options to the driver options.
			while( num_opt-- > 0 )
			{
				opt_key = ppd_to_opt_key(p_opt_key_table, p_opt->name);

				if( opt_key != NULL )
				{
					char *value = p_opt->value;
					param_list_add(&p_list, opt_key,
										value, strlen(value) + 1);
				}
				p_opt++;
			}

			// Make the driver options.
			opt_len = add_cmd_line_params(opt_buf, p_list, OPT_BUF_SIZE);
			cmd_len = strlen(GS_PARAMS) + opt_len + strlen(GS_OUTPUT);

			if( cmd_len < cmd_buf_len )
			{
#ifdef USE_CUSTOM_SIZE
				sprintf(cmd_buf, GS_PARAMS, 
					reso,
					(int)(p_size->width * (float)reso / 72.0),
  					(int)(p_size->length * (float)reso / 72.0),
					device_name,
					driver_name,
					model_name);
#else
				sprintf(cmd_buf, GS_PARAMS,
					reso,
					p_size_name,
					device_name,
					driver_name,
					model_name);
#endif
				strcat(cmd_buf, opt_buf);
				strcat(cmd_buf, GS_OUTPUT);
			}
			else
			{
				cmd_buf[0] = 0;
				cmd_len = -1;
			}

			free_opt_key_table(p_opt_key_table);
		}
		ppdClose(p_ppd);
	}
	return cmd_len;
}

int exec_filter(char* cmd_path, char* cmd_params[], int ofd, int fds[2])
{
	int status = 0;
	int	child_pid = -1;

	if( pipe(fds) >= 0 )
	{
		child_pid = fork();

		if( child_pid == 0 )
		{
			close(0);
			dup2(fds[0], 0);
			close(fds[0]);
			close(fds[1]);

			if( ofd != 1 )
			{
				close(1);
				dup2(ofd, 1);
				close(ofd);
			}

			execv(cmd_path, cmd_params);
						
			fprintf(stderr, "execl() error\n");
			status = -1;
		}
		else if( child_pid != -1 )
		{
			close(fds[0]);
		}
	}
	return child_pid;
}

#define CMD_BUF_SIZE	1024

int	main(int argc, char *argv[])
{
	cups_option_t *p_opt = NULL;
	ParamList *p_param = NULL;
	char *p_data_buf;
	int num_opt = 0;
	int filter_pid = -1;
	int ifd;
	int fds[2];
	char *ps_path;
	char filter_path[256];
	char *filter_param[4];
	char cmd_buf[CMD_BUF_SIZE];

#ifdef	DEBUG_SLEEP
	sleep(1000);
#endif

	setbuf(stderr, NULL);
	fprintf(stderr, "DEBUG: pstocanonopvp start.\n");

	signal(SIGCHLD, SIG_DFL);
	signal(SIGPIPE, SIG_DFL);

	if( argc < 6 || argc > 7 )
	{
		fputs("ERROR: pstocanonopvp illegal parameter number.\n", stderr);
		return 1;
	}

	if( argv[5] != NULL )
	{
		num_opt = cupsParseOptions(argv[5], 0, &p_opt);
		if( num_opt < 0 )
		{
			perror("ERROR: illegal option - ");
			return 1;
		}
	}

	if( argc == 7 )
	{
		if( (ifd = open(argv[6], O_RDONLY)) == -1 )
		{
			perror("ERROR: can't open file - ");
			return 1;
		}
	}
	else
	{
		ifd = 0;
	}

	p_param = get_ps_params(&ifd, &ps_path);
	if( ifd == -1 )
	{
		perror("ERROR: can't open data file - ");
		return 1;
	}
	fprintf(stderr, "DEBUG: pstocanonopvp ps_ifd,%d.\n", ifd);

	if( make_cmd_line(p_opt, num_opt, p_param, cmd_buf, CMD_BUF_SIZE) < 0 )
	{
		if( ifd != 0 )
			close(ifd);

		perror("ERROR: can't make parameter - ");
		return 1;
	}

	strcpy(filter_path, SHELL_PATH);
	strcat(filter_path, SHELL_NAME);

	filter_param[0] = SHELL_NAME;
	filter_param[1] = "-c";
	filter_param[2] = cmd_buf;
	filter_param[3] = NULL;

	p_data_buf = (char*)malloc(DATA_BUF_SIZE);

	if( p_data_buf != NULL )
	{
#ifdef	DEBUG_PS
		FILE *log_fp = fopen("/tmp/opvp_data_log.ps", "w");
#endif
		filter_pid = exec_filter(filter_path, filter_param, 1, fds);

		for(;;)
		{
			int read_bytes = read(ifd, p_data_buf, DATA_BUF_SIZE);

			if( read_bytes > 0 )
			{
				if( write(fds[1], p_data_buf, read_bytes) < 0 )
				{
					fprintf(stderr,
						"ERROR: pstocanonopvp write error,%d.\n", errno);
					break;
				}
#ifdef	DEBUG_PS
				fwrite(p_data_buf, read_bytes, 1, log_fp);
#endif
			}
			else if( read_bytes < 0 )
			{
				fprintf(stderr, "ERROR: pstocanonopvp read error,%d.\n", errno);
				break;
			}
			else{
				break;
			}
		}

		free(p_data_buf);

#ifdef	DEBUG_PS
		if( log_fp )	fclose(log_fp);
#endif
	}

	if( p_param != NULL )
		param_list_free(p_param);

	if( ifd != 0 )
	{
		close(ifd);
		unlink(ps_path);
	}

	close(fds[1]);

	if( filter_pid != -1 )
		waitpid(filter_pid, NULL, 0);

	return 0;
}

