#  openprinting-jp.spec -- creates an RPM package from source
#  Copyright (C) 2004  EPSON KOWA Corporation

#  If you want to build the whole she-bang as a single RPM (for ease
#  of installation or some such), set this macro to 1.
#
%define single_binary 0

#  With this macro set to 1, the %%pre will forcefully --erase the
#  ghostscript package.  This simulates what is called a Replaces:
#  in Debian circles because we provide (and obsolete) ghostscript.
#  This approach is better, IMHO, than silently overwriting files
#  owned by the ghostscript package.
#  Of course, when uninstalling we have no way of reinstalling the
#  ghostscript package unless we can assume there is a ghostscript
#  package somewhere below the default RPM tree :-{
#
%define forced_replace 1

Name:    openprinting-jp
Version: 0.1.3
Release: 0

%define gs_v 7.06
%define hp_v 1.3.1

Summary:  OpenPrinting Japan API Sample Implementation
Group:    Applications/Publishing
Vendor:   EPSON KOWA Corporation
Packager: Olaf Meeuwissen <olaf@epkowa.co.jp>
Source:	  %{name}-%{version}.tar.gz
License:  GPL/MIT, IPA Font License

#  These are from the turbolinux 10 ghostscript.spec file, but note
#  that the toplevel configure requires autoconf 2.54 or better AND
#  executes it when the ghostscript sources are unpacked.
#
BuildRequires: autoconf >= 2.54
BuildRequires: zlib-devel, libpng-devel, unzip, gtk+-devel
BuildRequires: glib-devel, XFree86-devel, libtiff-devel, libjpeg-devel
BuildRequires: gimp-print-devel, openssl-devel
BuildRequires: cups-devel >= 1.1.13
BuildRequires: libxml-devel

BuildRoot: %{_tmppath}/%{name}-%{version}-root

%if !%{single_binary}
%description
This package provides printer drivers and several supporting utilities
that demonstrate the capabilities of OpenPrinting API developed by the
Japanese branch of the OpenPrinting work group.

The API covers vector printer drivers for use by *any* renderer and
bi-directional communication between user-space applications and the
printer (as opposed to the spooler).

The drivers support several printers and are used by extended versions
of the Xprint and Ghostscript renderers provided by the %{name}-xprint
and %{name}-ghostscript packages.  You should install at least one of
these packages to be able to print anything.


%package ghostscript

Summary: A PostScript(TM) interpreter and renderer.
Group:   Applications/Publishing
License: GPL
%endif

Provides:  ghostscript = %{gs_v}
Obsoletes: ghostscript
Provides:  hpijs = %{hp_v}
Obsoletes: hpijs

#  These are from the turbolinux 10 ghostscript.spec file.
#
Requires: zlib, libpng, glib, libjpeg, libtiff
Requires: urw-fonts >= 1.1, ghostscript-fonts
Requires: gimp-print >= 4.2.5
Requires: cups-libs >= 1.1.13
Requires: XFree86-libs

%if !%{single_binary}
%description ghostscript
Ghostscript is a set of software that provides a PostScript(TM)
interpreter, a set of C procedures (the Ghostscript library, which
implements the graphics capabilities in the PostScript language) and
an interpreter for Portable Document Format (PDF) files.  Ghostscript
translates PostScript code into many common, bitmapped formats, like
those understood by your printer or screen.  Ghostscript is normally
used to display PostScript files and to print PostScript files to
non-PostScript printers.

This version has been extended by the OpenPrinting Japan work group to
support vector printers as well.  The printer drivers required to use
this extension as well as several supporting utilities are provided by
the %{name} package.

If you need to display PostScript files or print them to
non-PostScript printers, you should install this package.  If you
install this package, you also need to install the ghostscript-fonts
package.
%else				# single_binary
%description
This package provides printer drivers and several supporting utilities
that demonstrate the capabilities of OpenPrinting API developed by the
Japanese branch of the OpenPrinting work group.

The drivers support several printers and are used by extended versions
of the Xprint and Ghostscript renderers included in this package.
%endif

%if %{forced_replace}
Note that this package erases your ghostscript package but at the same
time provides a full replacement for it.  If you want to uninstall it,
you will (most likely) have to use the --nodeps option to rpm and then
reinstall the ghostscript package to satisfy the dependencies of other
packages installed on your system.
%endif

%if !%{single_binary}
%package xprint
Summary: Xprint
Group:   X11/Publishing
License: MIT

%description xprint 
Xprint provides an advanched print API for X11-based (incl. CDE,
Xt/Motif-, Xt/LessTif-, Xt/Athena-, Qt- and Mozilla-based)
applications.

This version has been extended by the OpenPrinting Japan work group to
support vector printers as well.  The printer drivers required to use
this extension as well as several supporting utilities are provided by
the %{name} package.
%endif


%prep
%setup -q


%build
%configure --enable-glue=yes \
    --with-fontpath=%{_datadir}/fonts:%{_prefix}/X11R6/lib/X11/fonts
make


%install
if test -n "${RPM_BUILD_ROOT}" && test "/" != "${RPM_BUILD_ROOT}"; then
    rm -rf ${RPM_BUILD_ROOT}
fi
make install DESTDIR=${RPM_BUILD_ROOT} \
    xp_tar_dir=%{name}-%{version}-xprint

#  Incorporate some tweaks from ghostscript.spec that are not handled
#  by the upstream build system (see glue/Makefile.am for details).
#
mkdir -p ${RPM_BUILD_ROOT}%{_mandir}/man1
ln -fs gs.1.gz ${RPM_BUILD_ROOT}%{_mandir}/man1/ghostscript.1.gz
ln -fs gs ${RPM_BUILD_ROOT}%{_bindir}/ghostscript
rm -f ${RPM_BUILD_ROOT}%{_bindir}/ijs_client_example
rm -f ${RPM_BUILD_ROOT}%{_bindir}/sysvlp.sh
mkdir -p ${RPM_BUILD_ROOT}%{_datadir}/ghostscript/Resource
mkdir -p ${RPM_BUILD_ROOT}%{_datadir}/ghostscript/Resource/CIDFont
mkdir -p ${RPM_BUILD_ROOT}%{_datadir}/ghostscript/Resource/Font

#  Zap some installed ghostscript related files/directories that are
#  not shipped with the turbolinux ghostscript package (this enables
#  rpmbuild with a Fascist build policy).
#
rm -fr ${RPM_BUILD_ROOT}%{_includedir}/ijs
rm -f  ${RPM_BUILD_ROOT}%{_libdir}/libijs.a
rm -fr ${RPM_BUILD_ROOT}%{_datadir}/doc/ghostscript-%{gs_v}
rm -fr ${RPM_BUILD_ROOT}%{_datadir}/doc/hpijs-%{hp_v}
rm -fr ${RPM_BUILD_ROOT}%{_mandir}/de

#  Zap all(!) the files that are provided by ghostscript-cups, so we
#  don't have to conflict with it.  Ours are the same anyway.
#
rm -f ${RPM_BUILD_ROOT}%{_sysconfdir}/cups/pstoraster.convs
rm -f ${RPM_BUILD_ROOT}%{_libdir}/cups/filter/pstoraster

#  Rename the Xprt binary to avoid clobbering the one from XFree86.
#  We'll fix this up in the %%post and revert in the %%postun.
#
mv  ${RPM_BUILD_ROOT}/usr/X11R6/bin/Xprt \
    ${RPM_BUILD_ROOT}/usr/X11R6/bin/Xprt_opvp

#  Build a list of Xprint files that get installed (coordinate with
#  glue/Makefile.am!) for use by the %%files section.
#  - rename the Xprt binary (see above)
#  - remove directories to avoid double inclusion
#  - suffix man pages because they are gzipped
#
tar tzf ${TMPDIR-/tmp}/%{name}-%{version}-xprint.tar.gz \
    | sed '/^xprint\/install/!d' \
    | sed 's,^xprint\/install,,' \
    | sed '/usr\/X11R6\/lib\/X11\/xserver\/SecurityPolicy/d' \
    | sed '/usr\/X11R6\/lib\/X11\/xserver\/README/d' \
    | sed 's,/bin/Xprt$,/bin/Xprt_opvp,' \
    | sed '/\/$/d' \
    | sed '/man\/man1/s,$,.gz,' \
    > %{_tmppath}/%{name}-%{version}-xprint-files


%clean
if test -n "${RPM_BUILD_ROOT}" && test "/" != "${RPM_BUILD_ROOT}"; then
    rm -rf ${RPM_BUILD_ROOT}
fi
make clean


#  WARNING: the %%pre, %%post, %%preun and %%postun are rather order
#         dependent to cater to the complexities associated with the
#         single_binary macro

%if %{forced_replace}
%pre
target=openprinting-jp
%if !%{single_binary}
%pre ghostscript
target=openprinting-jp-ghostscript
%endif
if test 1 = "$1"; then		# _replace_ the bastard
    if test -n "$(rpm -qa ghostscript)"; then
	rpm --nodeps --erase ghostscript
	ln -s ${target} /usr/share/doc/packages/ghostscript || true
    fi
fi
%if !%{single_binary}
%pre xprint
%endif
%endif				# forced_replace


%post
/sbin/ldconfig
if [ -x /usr/bin/fc-cache ] ; then
  HOME=/root /usr/bin/fc-cache %{_datadir}/fonts
fi

if [ -x /usr/sbin/chkfontpath ]; then
  /usr/sbin/chkfontpath -q -a %{_datadir}/fonts/ja/TrueType
fi
%if !%{single_binary}
%post xprint
%endif
if test 1 = "$1"; then
    cd /usr/X11R6/bin
    if test -x Xprt; then
	mv Xprt Xprt_XFree86
    fi
    ln -fs Xprt_opvp Xprt
    /sbin/chkconfig --add xprint
    /sbin/service xprint start
else
    /sbin/service xprint condrestart
fi
%if !%{single_binary}
%post ghostscript
/sbin/ldconfig
%endif


%postun
/sbin/ldconfig
if [ -x /usr/bin/fc-cache ]; then
  HOME=/root /usr/bin/fc-cache %{_datadir}/fonts
fi

if [ "$1" = "0" ]; then
  if [ -x /usr/sbin/chkfontpath ]; then
     /usr/sbin/chkfontpath -q -r %{_datadir}/fonts/ja/TrueType
  fi
fi
%if !%{single_binary}
%postun ghostscript
/sbin/ldconfig
%endif
%if %{forced_replace}
if test 0 = "$1"; then
    if test -h /usr/share/doc/packages/ghostscript; then
	rm -f /usr/share/doc/packages/ghostscript
    fi
fi
%endif
%if !%{single_binary}
%postun xprint
%endif
if test 0 = "$1"; then
    cd /usr/X11R6/bin
    if test ! -x Xprt; then
	if test -x Xprt_XFree86; then
	    mv Xprt_XFree86 Xprt
	fi
    fi
fi


%preun
%if !%{single_binary}
%preun ghostscript
%endif
%if !%{single_binary}
%preun xprint
%endif
if test 0 = "$1"; then
    /sbin/service xprint stop
    /sbin/chkconfig --del xprint
else
    /sbin/service xprint condrestart
fi


#  These files are courtesy of the openprinting-jp code.
#
%if %{single_binary}
%files -f %{_tmppath}/%{name}-%{version}-xprint-files
%else
%files
%endif
%defattr(-,root,root)
%{_bindir}/bidiProc
%{_bindir}/oppd-conf
%{_bindir}/oppd-listen
%{_bindir}/oppd-output
%{_bindir}/oppd-spool
%{_bindir}/oppd-status
%{_bindir}/opvp_rpc_EPSON_PX-V700_server
%{_bindir}/opvp_rpc_OPVPBJPixus850_server
%{_bindir}/printerStatusTypeA
%{_bindir}/printerStatusTypeB
#
#  FIXME: Should we really ship these include files?  As far as I am
#         concerned, these should probably not even be installed!
#
%{_includedir}/bidiEntry.h
%{_includedir}/bidiEntryCUPS.h
%{_includedir}/bidiParser.h
%{_includedir}/bidiTypes.h
%{_libdir}/bidiPIXUS850i.a
%{_libdir}/bidiPIXUS850i.la
%{_libdir}/bidiPIXUS850i.so
%{_libdir}/bidiPIXUS850i.so.0
%{_libdir}/bidiPIXUS850i.so.0.0.0
%{_libdir}/bidiPXV700.a
%{_libdir}/bidiPXV700.la
%{_libdir}/bidiPXV700.so
%{_libdir}/bidiPXV700.so.0
%{_libdir}/bidiPXV700.so.0.0.0
%{_libdir}/cups/backend/bidiparallel
%{_libdir}/cups/backend/bidiusb
%{_libdir}/cups/backend/op_lpd
%{_libdir}/cups/filter/pstocanonopvp
%{_libdir}/libCanonPageColor.a
%{_libdir}/libCanonPageColor.la
%{_libdir}/libCanonPageColor.so
%{_libdir}/libCanonPageColor.so.0
%{_libdir}/libCanonPageColor.so.0.0.0
%{_libdir}/libEpsonPageColor.a
%{_libdir}/libEpsonPageColor.la
%{_libdir}/libEpsonPageColor.so
%{_libdir}/libEpsonPageColor.so.0
%{_libdir}/libEpsonPageColor.so.0.0.0
%{_libdir}/libHPPageColor.a
%{_libdir}/libHPPageColor.la
%{_libdir}/libHPPageColor.so
%{_libdir}/libHPPageColor.so.0
%{_libdir}/libHPPageColor.so.0.0.0
%{_libdir}/libbidi-plugin.a
%{_libdir}/libbidi-plugin.la
%{_libdir}/libbidi-plugin.so
%{_libdir}/libbidi-plugin.so.0
%{_libdir}/libbidi-plugin.so.0.0.0
%{_libdir}/libbidiformat.a
%{_libdir}/libbidiformat.la
%{_libdir}/libbidiformat.so
%{_libdir}/libbidiformat.so.1
%{_libdir}/libbidiformat.so.1.0.0
%{_libdir}/opvp_rpc_EPSON_PX-V700.a
%{_libdir}/opvp_rpc_EPSON_PX-V700.la
%{_libdir}/opvp_rpc_EPSON_PX-V700.so
%{_libdir}/opvp_rpc_EPSON_PX-V700.so.0
%{_libdir}/opvp_rpc_EPSON_PX-V700.so.0.0.0
%{_libdir}/opvp_rpc_OPVPBJPixus850.a
%{_libdir}/opvp_rpc_OPVPBJPixus850.la
%{_libdir}/opvp_rpc_OPVPBJPixus850.so
%{_libdir}/opvp_rpc_OPVPBJPixus850.so.0
%{_libdir}/opvp_rpc_OPVPBJPixus850.so.0.0.0
%{_libdir}/pkgconfig/bidiformat.pc
%{_datadir}/cups/model/OPRP-RPC-Epson-PX_V700.ppd
%{_datadir}/cups/model/OPVP-Epson-LP_9500C.ppd
%{_datadir}/cups/model/OPVP-HP-Color_LaserJet_4600.ppd
%{_datadir}/cups/model/OPVP-HP-Color_LaserJet_5500.ppd
%{_datadir}/cups/model/opvp_canon_lbp2810.ppd
%{_datadir}/cups/model/opvp_canon_pixus850i.ppd
%lang(ja) %{_datadir}/locale/ja/LC_MESSAGES/bidiEpsonStatus.mo
%lang(ja) %{_datadir}/locale/ja/LC_MESSAGES/bidiPIXUS850i.mo
%lang(ja) %{_datadir}/locale/ja/LC_MESSAGES/oppd-conf.mo
%lang(ja) %{_datadir}/locale/ja/LC_MESSAGES/printerStatusTypeB.mo
#
# IPA font files
#
%doc ipafont/COPYING.font.ja
%{_datadir}/fonts/ja/TrueType/ipag.ttf
%{_datadir}/fonts/ja/TrueType/ipagp.ttf
%{_datadir}/fonts/ja/TrueType/ipagui.ttf
%{_datadir}/fonts/ja/TrueType/ipam.ttf
%{_datadir}/fonts/ja/TrueType/ipamp.ttf


%if !%{single_binary}
%files ghostscript
%defattr(-,root,root)
%endif
#  These have been adapted from the ghostscript %files section to
#  match our sources and notation.
#
%doc glue/ghostscript-%{gs_v}/doc/*
%doc glue/hpijs-%{hp_v}/gs_hpijs.png glue/hpijs-%{hp_v}/hpijs_readme.html
%dir %{_datadir}/ghostscript
%dir %{_datadir}/ghostscript/%{gs_v}
%dir %{_datadir}/ghostscript/Resource
%dir %{_datadir}/ghostscript/Resource/CIDFont
%dir %{_datadir}/ghostscript/Resource/Font
%{_datadir}/ghostscript/%{gs_v}/lib
%{_datadir}/ghostscript/%{gs_v}/examples
#
#%{_bindir}/*
#%{_libdir}/libijs.so*
#%{_mandir}/man?/*
#
#  The above have been replaced by an explicit list, mechanically derived
#  from the output of `rpm -qlp ghostscript-7.06-2.i586.rpm` included
#  below.  FYI, the complete command was:
#
#    rpm -qlp ghostscript-7.06-2.i586.rpm \
#	| grep -v doc/packages/ghostscript \
#	| sed 's,/usr/share/man,%{_mandir},' \
#	| sed 's,/usr/share,%{_datadir},' \
#	| sed 's,/usr/lib,%{_libdir},' \
#	| sed 's,/usr/bin,%{_bindir},' \
#	| sed '/Resource/d' \
#	| sed '/%{_datadir}\/ghostscript$/d' \
#       | sed '/%{_datadir}\/ghostscript\/7.06$/d' \
#       | sed '/%{_datadir}\/ghostscript\/7.06\/examples/d' \
#       | sed '/%{_datadir}\/ghostscript\/7.06\/lib/d' \
#	| sed 's,ghostscript/7\.06,ghostscript/%{gs_v},'
#
%{_bindir}/bdftops
%{_bindir}/dvipdf
%{_bindir}/eps2eps
%{_bindir}/fixmswrd.pl
%{_bindir}/font2c
%{_bindir}/ghostscript
%{_bindir}/gs
%{_bindir}/gsbj
%{_bindir}/gsdj
%{_bindir}/gsdj500
%{_bindir}/gslj
%{_bindir}/gslp
%{_bindir}/gsnd
%{_bindir}/hpijs
%{_bindir}/ijs-config
%{_bindir}/lprsetup.sh
%{_bindir}/pdf2dsc
%{_bindir}/pdf2ps
%{_bindir}/pdfopt
%{_bindir}/pf2afm
%{_bindir}/pfbtopfa
%{_bindir}/pj-gs.sh
%{_bindir}/printafm
%{_bindir}/ps2ascii
%{_bindir}/ps2epsi
%{_bindir}/ps2pdf
%{_bindir}/ps2pdf12
%{_bindir}/ps2pdf13
%{_bindir}/ps2pdf14
%{_bindir}/ps2pdfwr
%{_bindir}/ps2ps
%{_bindir}/pv.sh
%{_bindir}/unix-lpr.sh
%{_bindir}/wftopfa
%{_libdir}/libijs.so
%{_mandir}/man1/ansi2knr.1.gz
%{_mandir}/man1/dvipdf.1.gz
%{_mandir}/man1/eps2eps.1.gz
%{_mandir}/man1/font2c.1.gz
%{_mandir}/man1/ghostscript.1.gz
%{_mandir}/man1/gs.1.gz
%{_mandir}/man1/gsbj.1.gz
%{_mandir}/man1/gsdj.1.gz
%{_mandir}/man1/gsdj500.1.gz
%{_mandir}/man1/gslj.1.gz
%{_mandir}/man1/gslp.1.gz
%{_mandir}/man1/gsnd.1.gz
%{_mandir}/man1/pdf2dsc.1.gz
%{_mandir}/man1/pdf2ps.1.gz
%{_mandir}/man1/pdfopt.1.gz
%{_mandir}/man1/pf2afm.1.gz
%{_mandir}/man1/pfbtopfa.1.gz
%{_mandir}/man1/printafm.1.gz
%{_mandir}/man1/ps2ascii.1.gz
%{_mandir}/man1/ps2epsi.1.gz
%{_mandir}/man1/ps2pdf.1.gz
%{_mandir}/man1/ps2pdf12.1.gz
%{_mandir}/man1/ps2pdf13.1.gz
%{_mandir}/man1/ps2pdfwr.1.gz
%{_mandir}/man1/ps2ps.1.gz
%{_mandir}/man1/wftopfa.1.gz

%if !%{single_binary}
#  The list of Xprint files is computed by the %install and here we
#  simply include everything listed.
#
%files xprint -f %{_tmppath}/%{name}-%{version}-xprint-files
%defattr(-,root,root)
%endif


%changelog
* Wed Sep 22 2004  Olaf Meeuwissen <olaf@epkowa.co.jp> 0.1.3-0
  - resplit the package in the three parts we released for 0.1.1
  - folded the ipafont into the main package

* Thu Sep 16 2004  Olaf Meeuwissen <olaf@epkowa.co.jp> 0.1.2-0
  - added ipafont
  - build a single big package

* Thu Feb 26 2004  Olaf Meeuwissen <olaf@epkowa.co.jp> 0.1.1-0
  - new upstream

* Tue Feb 24 2004  Olaf Meeuwissen <olaf@epkowa.co.jp> 0.1-1
  - new upstream
  - added missing library symlinks

* Mon Feb 23 2004  Olaf Meeuwissen <olaf@epkowa.co.jp> 0.0.pre.15-0
  - initial release
