/*
 *  OpenPrinting Vector Printer Driver for Canon Pixus850i
 *  Copyright (c) 2004 E&D Inc.
 *  All rights reserved.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * NOTE:
 *  - As a special exception, this program is permissible to link with the
 *    libraries released as the binary modules.
 *  - If you write modifications of your own for these programs, it is your
 *    choice whether to permit this exception to apply to your modifications.
 *    If you do not wish that, delete this exception.
*/

/*
 * Define : Printer Control
 */

#ifndef	_PD_PRINTER_H
#define	_PD_PRINTER_H

typedef int bool;

#ifndef FALSE
#define FALSE 0
#endif

#ifndef TRUE
#define TRUE (!FALSE)
#endif

#ifndef NULL
#ifdef __cplusplus
#define NULL 0
#else
#define NULL (void *)0
#endif
#endif

#ifdef DEBUG
#define pd_DebugPrint(format,...)	fprintf(stderr,format,##__VA_ARGS__)
#else
#define pd_DebugPrint(format,...)
#endif

#define	MMETER_PER_INCH	25.4

/* GraphicsState */
#define	pd_LINEDASH_ARYMAX	16
typedef struct _pd_GraphicsState {
	pdColorSpace	colorSpace;
} pd_GraphicsState;

/* GraphicsStateList */
typedef struct _pd_GraphicsStateList {
	struct _pd_GraphicsStateList	*next;
	pd_GraphicsState	gstate;
} pd_GraphicsStateList;

/* DeviceInfo */
enum {
	pdDev_orient_NULL,
	pdDev_orient_portrait,
	pdDev_orient_landscape
};
enum {
	pdDev_pgSize_NULL,
	pdDev_pgSize_a3,
	pdDev_pgSize_a4,
	pdDev_pgSize_a5,
	pdDev_pgSize_b4,
	pdDev_pgSize_b5,
	pdDev_pgSize_b6,
	pdDev_pgSize_hagaki,
	pdDev_pgSize_letter,
	pdDev_pgSize_legal
};
typedef struct _pd_DeviceInfo {
	int		mediaCopy;
	int		orientation;	// pdDev_orient
	int		pageSizeID;	// pdDev_pgSize
	int		pageSizeX;	// mm : at !pageSizeID
	int		pageSizeY;
	int		resolutionX;	// dpi
	int		resolutionY;
} pd_DeviceInfo;

typedef struct _pd_supDevInfo {
	int		*mediaCopy;	// a,-1,b : a <= Num <= b
	int		*orientation;
	int		*pageSizeID;
	bool		pageSizeXY;	// support free size
	int		*resolution;	// X,Y, ...
} pd_supDevInfo;

/* PrinterInfo */
typedef struct _pd_PrinterInfo {
	int		printerContext;
	int		outputFD;
	pd_supDevInfo	supDev;		// support DeviceInfo
	pd_DeviceInfo	defDev;		// default DeviceInfo
	pd_DeviceInfo	curDev;		// current DeviceInfo
	bool		jobStarted;	// StartJob <-> EndJob
	pd_DeviceInfo	jobDev;
	bool		docStarted;	// StartDoc <-> EndDoc
	pd_DeviceInfo	docDev;
	bool		pageStarted;	// StartPage <-> EndPage
	pd_DeviceInfo	pageDev;
	bool		drwImgStarted;	// StartDrawImage <-> EndDrawImage
	bool		scnLinStarted;	// StartScanline <-> EndScanline
	bool		rasterStarted;	// StartRaster <-> EndRaster
	int		rasterWidth;
	pd_GraphicsStateList	gstateList;
	void		*printerData;	/* printer local data */
} pd_PrinterInfo;

/* PrinterInfoList */
typedef struct _pd_PrinterInfoList {
	struct _pd_PrinterInfoList	*next;
	pd_PrinterInfo		printer;
} pd_PrinterInfoList;

/*******************************************************************
  API entry
********************************************************************/
extern int	(*Inkjet_Procs[])();
extern pd_supDevInfo	Inkjet_supDev;

/*******************************************************************
  inside common functions
********************************************************************/

/* for Data Convert */
int	pd_DtoI( double );

/* for Data Management */
pdBrushData	*pd_CreateBrushData( pdBrushData * );
void	pd_EraseBrushData( pdBrushData ** );
void	pd_ClearGraphicsState( pd_GraphicsState * );
bool	pd_CopyGraphicsState( pd_GraphicsState *, pd_GraphicsState * );

pd_PrinterInfo	*pd_GetPrinterInfo( int );
pd_PrinterInfo	*pd_NewPrinterInfo( int );
bool	pd_DeletePrinterInfo( int );
pd_GraphicsState	*pd_GetGraphicsState( int );
void	pd_InitGraphicsState( pd_GraphicsState * );
bool	pd_PushGraphicsState( pd_PrinterInfo * );
bool	pd_PopGraphicsState( pd_PrinterInfo * );

/* for Data Out */
bool	pd_DvWrite( pd_PrinterInfo *, char *, size_t );
bool	pd_DvWriteStr( pd_PrinterInfo *, char * );

/* for Printer Control */
void	pd_SetDefDevInfo( pd_PrinterInfo * );
bool	pd_SetJobInfo( pd_PrinterInfo *, char * );
bool	pd_SetDocInfo( pd_PrinterInfo *, char * );
bool	pd_SetPageInfo( pd_PrinterInfo *, char * );

#endif
