/* Copyright (c) 2004 E-D Inc.  All rights reserved.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
/*******************************************************************

  OpenPrinting Vector Printer Driver API : for Canon

 ==================================================================

  Common Functions : LIPS Control

********************************************************************/
/* system include files */
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>

/* local include files */
#include "pdapi.h"
#include "pd_print.h"
#include "pd_lips.h"

/*******************************************************************
  for Data Convert
********************************************************************/
/* ----------------------------------------------------------------
 *
 *	pdLips_Param
 *
 * ---------------------------------------------------------------- */
void	pdLips_Param( int pdt, char *str )
{
	bool	sgn;
	int		i, j;

	sgn = TRUE;
	if( pdt < 0 ) {
		sgn = FALSE;
		pdt = -pdt;
	}
	if(      pdt < 16    )	i = 1;
	else if( pdt < 1024  )	i = 2;
	else if( pdt < 65536 )	i = 3;
	else					i = 4;
	str[i] = '\0';
	str[i - 1] = ( pdt & 0x0f ) | 0x20 | ( sgn ? 0x10 : 0x00 );
	pdt >>= 4;
	for( j = i - 2; j >= 0; j-- ) {
		str[j] = ( pdt & 0x3f ) | 0x40;
		pdt >>= 6;
	}
}
/* ----------------------------------------------------------------
 *
 *	pdLips_ParamAdd
 *
 * ---------------------------------------------------------------- */
void	pdLips_ParamAdd( int pdt, char *str )
{
	pdLips_Param( pdt, str + strlen( str ) );
}

/*******************************************************************
  for Attribute Out
********************************************************************/
/* ----------------------------------------------------------------
 *
 *	pdLips_Out_FillMode
 *
 * ---------------------------------------------------------------- */
bool	pdLips_Out_FillMode( pd_PrinterInfo *pri )
{
	pd_GraphicsState	*gs = &pri->gstateList.gstate;
	char	obuf[128];

	if( gs->fillMode == pdfillModeWinding ) {
		sprintf( obuf, "}I1%c", C_IS2 );
	} else {
		sprintf( obuf, "}I0%c", C_IS2 );
	}
	if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );
	return( TRUE );
}

/* ----------------------------------------------------------------
 *
 *	pdLips_Out_LineWidth
 *
 * ---------------------------------------------------------------- */
bool	pdLips_Out_LineWidth( pd_PrinterInfo *pri )
{
	pd_GraphicsState	*gs = &pri->gstateList.gstate;
	char	obuf[128];
	char	prm1[16];
	int		i;

	i = Fixtoi( gs->lineWidth );
	if ( i < 1 )	i = 1;
	pdLips_Param( i, prm1 );
	sprintf( obuf, "F1%s%c", prm1, C_IS2 );
	if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );
	return( TRUE );
}

/* ----------------------------------------------------------------
 *
 *	pdLips_Out_LineStyle
 *
 * ---------------------------------------------------------------- */
bool	pdLips_Out_LineStyle( pd_PrinterInfo *pri )
{
	pd_GraphicsState	*gs = &pri->gstateList.gstate;
	char	obuf[256];
	char	prm1[16], prm2[128];
	int		dn, n;

	if( gs->lineStyle == pdlineStyleDash && gs->dashNum > 0 ) {
		pdLips_Param( -12, prm1 );
		pdLips_Param( Fixtoi( gs->dashOffset ), prm2 );
		dn = gs->dashNum;
		if( gs->dashNum % 2 )	dn *= 2;
		for( n = 0; n < dn; n++ ) {
			pdLips_ParamAdd( Fixtoi( gs->lineDash[n] ), prm2 );
		}
		sprintf( obuf, "}d%s1%s%c", prm1, prm2, C_IS2 );
		if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );

		sprintf( obuf, "E1%s1%c", prm1, C_IS2 );
		if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );
	} else {
		sprintf( obuf, "E101%c", C_IS2 );
		if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );
	}
	return( TRUE );
}

/* ----------------------------------------------------------------
 *
 *	pdLips_Out_LineCap
 *
 * ---------------------------------------------------------------- */
bool	pdLips_Out_LineCap( pd_PrinterInfo *pri )
{
	pd_GraphicsState	*gs = &pri->gstateList.gstate;
	char	obuf[128];

	if( gs->lineCap == pdlineCapRound ) {
		sprintf( obuf, "}E1%c", C_IS2 );
	} else if( gs->lineCap == pdlineCapSquare ) {
		sprintf( obuf, "}E2%c", C_IS2 );
	} else {
		sprintf( obuf, "}E0%c", C_IS2 );
	}
	if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );
	return( TRUE );
}

/* ----------------------------------------------------------------
 *
 *	pdLips_Out_LineJoin
 *
 * ---------------------------------------------------------------- */
bool	pdLips_Out_LineJoin( pd_PrinterInfo *pri )
{
	pd_GraphicsState	*gs = &pri->gstateList.gstate;
	char	obuf[128];

	if( gs->lineJoin == pdlineJoinMiter ) {
		sprintf( obuf, "}F2%c", C_IS2 );
	} else if( gs->lineJoin == pdlineJoinRound ) {
		sprintf( obuf, "}F1%c", C_IS2 );
	} else if( gs->lineJoin == pdlineJoinBevel ) {
		sprintf( obuf, "}F3%c", C_IS2 );
	} else {
		sprintf( obuf, "}F0%c", C_IS2 );
	}
	if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );
	return( TRUE );
}

/* ----------------------------------------------------------------
 *
 *	pdLips_Out_MiterLimit
 *
 * ---------------------------------------------------------------- */
bool	pdLips_Out_MiterLimit( pd_PrinterInfo *pri )
{
	pd_GraphicsState	*gs = &pri->gstateList.gstate;
	char	obuf[128];
	char	prm1[16];
	double	w, l;
	int		i;

	w = Fixtof( gs->lineWidth );
	l = Fixtof( gs->miterLimit );
	if( l <= 0 || w <= 0 ) {
		i = 0;
	} else if( w >= l * 2 ) {
		i = 2 * 16383;
	} else {
		i = w * 16383 / l;
	}
	pdLips_Param( i, prm1 );
	sprintf( obuf, "}M%s%c", prm1, C_IS2 );
	if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );
	return( TRUE );
}

/* ----------------------------------------------------------------
 *
 *	pdLips_Out_PaintMode
 *
 * ---------------------------------------------------------------- */
bool	pdLips_Out_PaintMode( pd_PrinterInfo *pri )
{
	pd_GraphicsState	*gs = &pri->gstateList.gstate;
	char	obuf[128];

	if( gs->paintMode == pdpaintModeTransparent ) {
		sprintf( obuf, "}H1%c", C_IS2 );
	} else {
		sprintf( obuf, "}H0%c", C_IS2 );
	}
	if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );
	return( TRUE );
}

/*==================================================================*/
/* ----------------------------------------------------------------
 *
 *	pdLips_Out_PaintMode
 *
 * ---------------------------------------------------------------- */
void	pdLips_Prm_Color( int *col, char *str )
{
	int		i;
	i = (double)(unsigned char)col[2] * 1000 / 255;
	pdLips_Param( i, str );
	i = (double)(unsigned char)col[1] * 1000 / 255;
	pdLips_ParamAdd( i, str );
	i = (double)(unsigned char)col[0] * 1000 / 255;
	pdLips_ParamAdd( i, str );
}

/* ----------------------------------------------------------------
 *
 *	out_PtnBrush
 *
 * ---------------------------------------------------------------- */
static bool	out_PtnBrush( pd_PrinterInfo *pri, pdBrush *brs, int id )
{
	char	obuf[128];
	char	prm1[16];
	char	ptn[ 4 * 32 ];
	int		sx, sy, dx, dy;

	pdLips_Param( id, prm1 );
	sprintf( obuf, "}O1%s%c", prm1, C_IS2 );	// binary format
	if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );

	memset( ptn, 0, 4 * 32 );
	for( sy = brs->yorg, dy = 0; dy < 32; dy++, sy++ ) {
		if( sy >= brs->pbrush->height )	sy = 0;
		for( sx = brs->xorg, dx = 0; dx < 32; dx++, sx++ ) {
			if( sx >= brs->pbrush->width )	sx = 0;
			if( brs->pbrush->data[ ( brs->pbrush->pitch * sy ) + ( sx / 8 ) ] & ( 0x80 >> ( sx % 8 ) ) ) {
				ptn[ ( 4 * dy ) + ( dx / 8 ) ] |= 0x80 >> ( dx % 8 );
			}
		}
	}
	if( !pd_DvWrite( pri, ptn, 4 * 32 ) )	return( FALSE );
	return( TRUE );
}

/* ----------------------------------------------------------------
 *
 *	pdLips_Out_StrokeBrush
 *
 * ---------------------------------------------------------------- */
bool	pdLips_Out_StrokeBrush( pd_PrinterInfo *pri )
{
	pd_GraphicsState	*gs = &pri->gstateList.gstate;
	char	obuf[128];
	char	prm1[16];

	pdLips_Prm_Color( gs->strokeBrush.color, prm1 );
	sprintf( obuf, "G%s%c", prm1, C_IS2 );
	if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );

	if( gs->strokeBrush.pbrush ) {
		if( !out_PtnBrush( pri, &gs->strokeBrush, -10 ) )	return( FALSE );
		pdLips_Param( -10, prm1 );
		sprintf( obuf, "}G3%s%c", prm1, C_IS2 );
		if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );
	} else {
		sprintf( obuf, "}G0%c", C_IS2 );
		if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );
	}
	return( TRUE );
}

/* ----------------------------------------------------------------
 *
 *	pdLips_Out_FillBrush
 *
 * ---------------------------------------------------------------- */
bool	pdLips_Out_FillBrush( pd_PrinterInfo *pri )
{
	pd_GraphicsState	*gs = &pri->gstateList.gstate;
	char	obuf[128];
	char	prm1[16];

	pdLips_Prm_Color( gs->fillBrush.color, prm1 );
	sprintf( obuf, "J%s%c", prm1, C_IS2 );
	if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );

	if( gs->fillBrush.pbrush ) {
		if( !out_PtnBrush( pri, &gs->fillBrush, -11 ) )	return( FALSE );
		pdLips_Param( -11, prm1 );
		sprintf( obuf, "I%s%c", prm1, C_IS2 );
		if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );
	} else {
		sprintf( obuf, "I1%c", C_IS2 );
		if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );
	}
	return( TRUE );
}

/*==================================================================*/
/* ----------------------------------------------------------------
 *
 *	pdLips_AttrOutGS
 *
 * ---------------------------------------------------------------- */
bool	pdLips_AttrOutGS( pd_PrinterInfo *pri )
{
	if( !pdLips_Out_FillMode( pri ) )		return( FALSE );
	if( !pdLips_Out_LineWidth( pri ) )		return( FALSE );
	if( !pdLips_Out_LineStyle( pri ) )		return( FALSE );
	if( !pdLips_Out_LineCap( pri ) )		return( FALSE );
	if( !pdLips_Out_LineJoin( pri ) )		return( FALSE );
	if( !pdLips_Out_MiterLimit( pri ) )		return( FALSE );
	if( !pdLips_Out_PaintMode( pri ) )		return( FALSE );
	if( !pdLips_Out_StrokeBrush( pri ) )	return( FALSE );
	if( !pdLips_Out_FillBrush( pri ) )		return( FALSE );
	return( TRUE );
}
