/* Copyright (c) 2004 E-D Inc.  All rights reserved.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
/*******************************************************************

  OpenPrinting Vector Printer Driver API : for Canon

 ==================================================================

  Define : LIPS Control

********************************************************************/
#ifndef	_PD_LIPS_H
#define	_PD_LIPS_H

/* Control Code */
#define	C_BS	(0x08)
#define	C_HT	(0x09)
#define	C_LF	(0x0a)
#define	C_VT	(0x0b)
#define	C_FF	(0x0c)
#define	C_CR	(0x0d)
#define	C_SO	(0x0e)
#define	C_SI	(0x0f)
#define	C_ESC	(0x1b)
#define	C_IS2	(0x1e)
#define	C_IS1	(0x1f)
#define	C_SP	(0x20)

/*******************************************************************
  PDAPI functions
********************************************************************/
/* Job Control Operations */
int LIPS_StartJob( int printerContext, char *jobInfo );
int LIPS_EndJob( int printerContext );
int LIPS_StartDoc( int printerContext, char *docInfo );
int LIPS_EndDoc( int printerContext );
int LIPS_StartPage( int printerContext, char *pageInfo );
int LIPS_EndPage( int printerContext );

/* Graphics State Object Operations */
int LIPS_ResetCTM( int printerContext );
int LIPS_SetCTM( int printerContext, pdCTM *pCTM );
int LIPS_GetCTM( int printerContext, pdCTM *pCTM );
int LIPS_InitGS( int printerContext );
int LIPS_SaveGS( int printerContext );
int LIPS_RestoreGS( int printerContext );
int LIPS_QueryColorSpace( int printerContext, pdColorSpace *pcspace, int *pnum );
int LIPS_SetColorSpace( int printerContext, pdColorSpace cspace );
int LIPS_GetColorSpace( int printerContext, pdColorSpace *pcspace );
int LIPS_QueryROP( int printerContext, int *pnum, int *prop );
int LIPS_SetROP( int printerContext, int rop );
int LIPS_GetROP( int printerContext, int *prop );
int LIPS_SetFillMode( int printerContext, pdFillMode fillmode );
int LIPS_GetFillMode( int printerContext, pdFillMode *pfillmode );
int LIPS_SetAlphaConstant( int printerContext, float alpha );
int LIPS_GetAlphaConstant( int printerContext, float *palpha );
int LIPS_SetLineWidth( int printerContext, pdFix width );
int LIPS_GetLineWidth( int printerContext, pdFix *pwidth );
int LIPS_SetLineDash( int printerContext, pdFix pdash[], int num );
int LIPS_GetLineDash( int printerContext, pdFix pdash[], int *pnum );
int LIPS_SetLineDashOffset( int printerContext, pdFix offset );
int LIPS_GetLineDashOffset( int printerContext, pdFix *poffset );
int LIPS_SetLineStyle( int printerContext, pdLineStyle linestyle );
int LIPS_GetLineStyle( int printerContext, pdLineStyle *plinestyle );
int LIPS_SetLineCap( int printerContext, pdLineCap linecap );
int LIPS_GetLineCap( int printerContext, pdLineCap *plinecap );
int LIPS_SetLineJoin( int printerContext, pdLineJoin linejoin );
int LIPS_GetLineJoin( int printerContext, pdLineJoin *plinejoin );
int LIPS_SetMiterLimit( int printerContext, pdFix miterlimit );
int LIPS_GetMiterLimit( int printerContext, pdFix *pmiterlimit );
int LIPS_SetPaintMode( int printerContext, pdPaintMode paintmode );
int LIPS_GetPaintMode( int printerContext, pdPaintMode *ppaintmode );
int LIPS_SetStrokeColor( int printerContext, pdBrush *brush );
int LIPS_SetFillColor( int printerContext, pdBrush *brush );
int LIPS_SetBgColor( int printerContext, pdBrush *brush );

/* Path Operations */
int LIPS_NewPath( int printerContext );
int LIPS_EndPath( int printerContext );
int LIPS_StrokePath( int printerContext );
int LIPS_FillPath( int printerContext );
int LIPS_StrokeFillPath( int printerContext );
int LIPS_SetClipPath( int printerContext, pdClipRule clipRule );
int LIPS_ResetClipPath( int printerContext );
int LIPS_SetCurrentPoint( int printerContext, pdFix x, pdFix y );
int LIPS_LinePath( int printerContext, int flag, int npoints, pdPoint *points );
int LIPS_PolygonPath( int printerContext, int npolygons, int *nvertexes, pdPoint *points );
int LIPS_RectanglePath( int printerContext, int nrectangles, pdRectangle *rectangles );
int LIPS_RoundRectanglePath( int printerContext, int nrectangles, pdRoundRectangle *rectangles );
int LIPS_BezierPath( int printerContext, int npoints, pdPoint *points );
int LIPS_ArchPath( int printerContext, int kind, int dir, pdFix bbx0, pdFix bby0, pdFix bbx1, pdFix bby1, pdFix x0, pdFix y0, pdFix x1, pdFix y1 );

/* Text Operations */
int LIPS_DrawBitmapText( int printerContext, int width, int height, int pitch, void *fontdata );

/* Bitmap Image Operations */
int LIPS_DrawImage( int printerContext, int sourceWidth, int sourceHeight, int colorDepth, pdImageFormat imageFormat, pdRectangle destinationSize, int count, void *imageData );
int LIPS_StartDrawImage( int printerContext, int sourceWidth, int sourceHeight, int colorDepth, pdImageFormat imageFormat, pdRectangle destinationSize );
int LIPS_TransferDrawImage( int printerContext, int count, void *imageData );
int LIPS_EndDrawImage( int printerContext );

/* Scan Line Operations */
int LIPS_StartScanline( int printerContext, int yposition );
int LIPS_Scanline( int printerContext, int nscanpairs, int *scanpairs );
int LIPS_EndScanline( int printerContext );

/* Raster Image Operations */
int LIPS_StartRaster( int printerContext, int rasterWidth );
int LIPS_TransferRasterData( int printerContext, int count, unsigned char *data );
int LIPS_SkipRaster( int printerContext, int count );
int LIPS_EndRaster( int printerContext );

/* Stream Data Operations */
int LIPS_StartStream( int printerContext );
int LIPS_TransferStreamData( int printerContext, int count, void *data );
int LIPS_EndStream( int printerContext );

/*******************************************************************
  common functions
********************************************************************/
/* for Data Convert */
void	pdLips_Param( int, char * );
void	pdLips_ParamAdd( int, char * );

/* for Attribute Out */
bool	pdLips_Out_FillMode( pd_PrinterInfo * );
bool	pdLips_Out_LineWidth( pd_PrinterInfo * );
bool	pdLips_Out_LineStyle( pd_PrinterInfo * );
bool	pdLips_Out_LineCap( pd_PrinterInfo * );
bool	pdLips_Out_LineJoin( pd_PrinterInfo * );
bool	pdLips_Out_MiterLimit( pd_PrinterInfo * );
bool	pdLips_Out_PaintMode( pd_PrinterInfo * );

void	pdLips_Prm_Color( int *, char * );
bool	pdLips_Out_StrokeBrush( pd_PrinterInfo * );
bool	pdLips_Out_FillBrush( pd_PrinterInfo * );

bool	pdLips_AttrOutGS( pd_PrinterInfo * );

#endif
